/* bx3DtoOrg.cpp */

#include "bx3DtoOrg.h"
#include "../../Math/bxMath.h"


bool  Cbx3DtoOrg::
HandleVec( const double axisVec[ 2 ][ 3 ], const double radialVec[ 3 ], const double tan[ 3 ],
	double handleVec[ 2 ] )
{
	bool    errResult, minusSin;
	double  axisVec2[ 2 ][ 3 ], tanE[ 3 ], cosP, lenRadialVec, lenTan, normal[ 3 ];

	CbxMath::RegularAxis( axisVec, axisVec2, &errResult );
	if( errResult == true )
		return( true );

	CbxMath::Normalize3( tan, CbxMath::zero3, tanE );
	cosP = CbxMath::Dot3( axisVec2[ 0 ], tanE );
	CbxMath::Cross3( axisVec2[ 0 ], axisVec2[ 1 ], normal );
	minusSin = CbxMath::IsMinusSin( normal, axisVec2[ 0 ], tanE );
	handleVec[ 0 ] = CbxMath::CosToDeg( cosP, minusSin, &errResult );
	if( errResult == true )
		return( true );

	lenRadialVec = CbxMath::Length3( radialVec );
	lenTan       = CbxMath::Length3( tan );
	if( lenRadialVec < CbxMath::nearZero ) {
		handleVec[ 1 ] = 0.0;
		return( false );
	}

	handleVec[ 1 ] = lenTan / lenRadialVec;

	return( false );
}


bool  Cbx3DtoOrg::
CtlPlaneDeg( const double axisInCtlPlaneDeg[ 3 ], const double defaultNormalInCtlPlaneDeg[ 3 ],
	const double normalInCtlPlaneDeg[ 3 ], double* ctlPlaneDeg )
{
	double  defaultNormalInCtlPlaneDeg2[ 3 ], normalInCtlPlaneDeg2[ 3 ], cosP;
	bool    minusSin, errResult;

	CbxMath::Normalize3( defaultNormalInCtlPlaneDeg, CbxMath::zero3, defaultNormalInCtlPlaneDeg2 );
	CbxMath::Normalize3( normalInCtlPlaneDeg,        CbxMath::zero3, normalInCtlPlaneDeg2        );

	cosP = CbxMath::Dot3( defaultNormalInCtlPlaneDeg, normalInCtlPlaneDeg );
	minusSin = CbxMath::IsMinusSin( axisInCtlPlaneDeg, defaultNormalInCtlPlaneDeg, normalInCtlPlaneDeg );
	*ctlPlaneDeg = CbxMath::CosToDeg( cosP, minusSin, &errResult );
	if( errResult == true )
		return( true );

	return( false );
}


bool  Cbx3DtoOrg::
CtlVec( const double axisVec[ 2 ][ 3 ], const double lenBaseCtlVec, const double iCtlVec[ 3 ], double oCtlVec[ 2 ] )
{
	bool    errResult, minusSin;
	double  axisVec2[ 2 ][ 3 ], iCtlVecE[ 3 ], cosP, normal[ 3 ];

	CbxMath::RegularAxis( axisVec, axisVec2, &errResult );
	if( errResult == true )
		return( true );

	CbxMath::Normalize3( iCtlVec, CbxMath::zero3, iCtlVecE );
	cosP = CbxMath::Dot3( axisVec2[ 0 ], iCtlVecE );
	CbxMath::Cross3( axisVec2[ 0 ], axisVec2[ 1 ], normal );
	minusSin = CbxMath::IsMinusSin( normal, axisVec2[ 0 ], iCtlVecE );
	oCtlVec[ 0 ] = CbxMath::CosToDeg( cosP, minusSin, &errResult );
	if( errResult == true )
		return( true );

	oCtlVec[ 1 ] = CbxMath::Length3( iCtlVec ) / lenBaseCtlVec;

	return( false );
}
