/* bxCoreSub.cpp */

#include "bxCoreSub.h"
#include "bxErrChk.h"
#include "../../Math/bxMath.h"
#include "../3DtoOrg/bx3DtoOrg.h"
#include "../OrgTo3D/bxOrgTo3D.h"
#include "../Element/bxElement.h"
#include "../Vector/bxVector.h"



bool  CbxCoreSub::
NormalFromAxisForDefault( const CbxCore::AxisForDefaultIn* in, double normal[ 3 ] )
{
	ushort  no0, no1, num;
	double  normalOne[ 3 ];
	bool    errResult;

	CbxMath::Copy3( CbxMath::zero3, normal );

	num = in->numRadial;
	if( in->loopFlag == false )
		num--;

	for( no0=0; no0<num; no0++ ) {
		no1 = ( no0 + 1 ) % in->numRadial;
		NormalFromAxisForDefaultOne( no0, no1, in, normalOne );
		CbxMath::Add3( normal, normalOne, normal );
	}

	CbxMath::Normalize3( normal, 0, normal, &errResult );
	if( errResult == true )
		return( true );

	return( false );
}


void  CbxCoreSub::
NormalFromAxisForDefaultOne( const ulong radialNo0, const ulong radialNo1, const CbxCore::AxisForDefaultIn* in,
	double normal[ 3 ] )
{
	double  vec[ 2 ][ 3 ];

	CbxMath::Subtract3( in->radial[ radialNo0 ].vertexPos, in->vertexPos, vec[ 0 ] );
	CbxMath::Subtract3( in->radial[ radialNo1 ].vertexPos, in->vertexPos, vec[ 1 ] );
	
	CbxMath::Cross3( vec[ 0 ], vec[ 1 ], normal );
	CbxMath::Normalize3( normal, CbxMath::zero3, normal );
}


bool  CbxCoreSub::
GetDefaultAxisMain( const double normal[ 3 ], CbxCore::AxisForDefaultOut* out )
{
	double  tmp[ 2 ][ 3 ];
	bool    errResult;

	CbxMath::MakeSubAxis( normal, tmp, &errResult );
	if( errResult == true )
		return( true );

	CbxMath::Copy3( tmp[ 1 ], out->axisVec[ 0 ] );
	CbxMath::Cross3( normal, out->axisVec[ 0 ], out->axisVec[ 1 ] );

	CbxMath::Normalize3( out->axisVec[ 1 ], 0, out->axisVec[ 1 ], &errResult );
	if( errResult == true )
		return( true );

	return( false );
}


bool  CbxCoreSub::
NormalFromHandleForDefault( const CbxCore::HandleForTransIn* in, double normal[ 3 ] )
{
	bool  errResult;

	CbxMath::Cross3( in->axisVec[ 0 ], in->axisVec[ 1 ], normal );

	CbxMath::Normalize3( normal, 0, normal, &errResult );
	if( errResult == true )
		return( true );

	return( false );
}


void  CbxCoreSub::
TanEfromHandleForDefault( const double radialVec[ 3 ], const double normal[ 3 ], double tanE[ 3 ] )
{
	double  tmp[ 3 ];

	CbxMath::Cross3( radialVec, normal, tmp );
	CbxMath::Cross3( normal, tmp, tanE );
	CbxMath::Normalize3( tanE, CbxMath::zero3, tanE );
}


bool  CbxCoreSub::
GetDefaultHandleOut( const CbxCore::HandleForTransIn* in, const double radialVec[ 3 ], const double tan[ 3 ],
	CbxCore::HandleForTransOut* out )
{
	return( Cbx3DtoOrg::HandleVec( in->axisVec, radialVec, tan, out->handleVec ) );
}


bool  CbxCoreSub::
NormalInCtlPlaneDeg( const CbxCore::CtlPlaneDegForRotSub sub[ 2 ], const double ctlPlaneDeg, double normalVec[ 3 ] )
{
	bool    result;
	double  handleVec[ 2 ][ 3 ], axisInCtlPlaneDeg[ 3 ];

	result = ElementInCtlPlaneDeg( sub, ctlPlaneDeg, handleVec, axisInCtlPlaneDeg, normalVec );
	if( result == true )
		return( true );

	return( false );
}


bool  CbxCoreSub::
ElementInCtlPlaneDeg( const CbxCore::CtlPlaneDegForRotSub sub[ 2 ], const double ctlPlaneDeg, double handleVec[ 2 ][ 3 ],
	double axisInCtlPlaneDeg[ 3 ], double normalInCtlPlaneDeg[ 3 ] )
{
	bool    result;
	short   i;
	double  len, normalOnVertex[ 2 ][ 3 ], defaultNormalInCtlPlaneDeg[ 3 ];

	len = CbxMath::Distance3( sub[ 0 ].vertexPos, sub[ 1 ].vertexPos );

	for( i=0; i<2; i++ ) {
		result = CbxOrgTo3D::HandleVec( sub[ i ].axisVec, len, sub[ i ].handleVec, handleVec[ i ] );
		if( result == true )
			return( true );

		CbxMath::Cross3( sub[ i ].axisVec[ 0 ], sub[ i ].axisVec[ 1 ], normalOnVertex[ i ] );
		CbxMath::Normalize3( normalOnVertex[ i ], CbxMath::zero3, normalOnVertex[ i ] );
	}

	CbxElement::AxisInCtlPlaneDeg( sub[ 0 ].vertexPos, handleVec[ 0 ], handleVec[ 1 ],
		sub[ 1 ].vertexPos, axisInCtlPlaneDeg );
	CbxElement::DefaultNormalInCtlPlaneDeg( normalOnVertex[ 0 ], axisInCtlPlaneDeg, normalOnVertex[ 1 ],
		defaultNormalInCtlPlaneDeg );
	
	result = CbxOrgTo3D::NormalInCtlPlaneDeg( axisInCtlPlaneDeg, defaultNormalInCtlPlaneDeg, ctlPlaneDeg,
				 normalInCtlPlaneDeg );
	if( result == true )
		return( true );

	return( false );
}


bool  CbxCoreSub::
SetHandleInfo( const double newHandleVec[ 3 ], const CbxCore::HandleForTransPlusIn* in, CbxCore::HandleForTransPlusOut* out )
{
	bool    result;
	double  axisForNormal[ 3 ], radialVec[ 3 ], cross[ 3 ], len;

	if( CbxMath::Length3( newHandleVec ) < CbxMath::nearZero ) {
		out->handleVec[ 1 ] = 0.0;
		return( false );
	}

	result = GetCrossBoundOnVertex( in, out, axisForNormal );
	if( result == true || CbxMath::Length3( axisForNormal ) < CbxMath::nearZero )
		return( true );

	CbxMath::Cross3( newHandleVec, axisForNormal, cross );
	if( CbxMath::Length3( cross ) < CbxMath::nearZero ) {
		len = CbxMath::Distance3( in->vertexPos, in->radialVertexPos );
		out->handleVec[ 1 ] = CbxMath::Length3( newHandleVec ) / len;
		return( false );
	}

	result = GetNewAxisVec( newHandleVec, axisForNormal, out->axisVec );
	if( result == true )
		return( true );

	CbxMath::Subtract3( in->radialVertexPos, in->vertexPos, radialVec );
	result = Cbx3DtoOrg::HandleVec( out->axisVec, radialVec, newHandleVec, out->handleVec );
	if( result == true )
		return( true );

	return( false );
}


bool  CbxCoreSub::
GetCrossBoundOnVertex( const CbxCore::HandleForTransPlusIn* in, const CbxCore::HandleForTransPlusOut* out,
	double axisForNormal[ 3 ] )
{
	bool    result;
	short   i;
	double  wingHandle[ 2 ][ 3 ], len, handleVec[ 3 ], normal[ 3 ];

	for( i=0; i<2; i++ ) {
		result = CbxOrgTo3D::HandleVec( out->axisVec, in->lenRadialVec[ i ], in->wingHandle[ i ], wingHandle[ i ] );
		if( result == true )
			return( true );
	}
		
	CbxElement::CrossBoundOnVertex( wingHandle, axisForNormal );
	if( CbxMath::Length3( axisForNormal ) >= CbxMath::nearZero )
		return( false );
	
	len = CbxMath::Distance3( in->vertexPos, in->radialVertexPos );
	result = CbxOrgTo3D::HandleVec( out->axisVec, len, out->handleVec, handleVec );
	if( result == true )
		return( true );

	CbxMath::Cross3( out->axisVec[ 0 ], out->axisVec[ 1 ], normal );
	CbxMath::Cross3( handleVec, normal, axisForNormal );
	CbxMath::Normalize3( axisForNormal, CbxMath::zero3, axisForNormal );

	return( false );
}


bool  CbxCoreSub::
GetNewAxisVec( const double newHandleVec[ 3 ], const double axisForNormal[ 3 ], double axisVec[ 2 ][ 3 ] )
{
	bool    errResult, minusSin;
	short   i;
	double  oldNormal[ 3 ], newNormal[ 3 ], cosP, deg, mat[ 4 ][ 4 ];

	CbxMath::Cross3( axisVec[ 0 ], axisVec[ 1 ], oldNormal );
	CbxMath::Normalize3( oldNormal, CbxMath::zero3, oldNormal );

	CbxMath::Cross3( axisForNormal, newHandleVec, newNormal );
	CbxMath::Normalize3( newNormal, CbxMath::zero3, newNormal );

	cosP     = CbxMath::Dot3( oldNormal, newNormal );
	minusSin = CbxMath::IsMinusSin( axisForNormal, oldNormal, newNormal );
	deg      = CbxMath::CosToDeg( cosP, minusSin, &errResult );
	if( errResult == true )
		return( true );
	CbxMath::RotateAxis4x4( deg, axisForNormal, mat, &errResult );
	if( errResult == true )
		return( true );

	for( i=0; i<2; i++ )
		CbxMath::Transform3( axisVec[ i ], mat, axisVec[ i ] );

	CbxMath::RegularAxis( axisVec, axisVec, &errResult );
	if( errResult == true )
		return( true );

	return( false );
}


bool  CbxCoreSub::
SetCtlInfo( const double newCtlVec[ 3 ], const CbxCore::CtlVecForTransPlusIn* in, CbxCore::CtlVecForTransPlusOut* out )
{
	bool    result;
	double  axisInCtlPlaneDeg[ 3 ], defaultNormalInCtlPlaneDeg[ 3 ], normalInCtlPlaneDeg[ 3 ];
	double  len, axis[ 2 ][ 3 ], lenBaseCtlVec;

	if( CbxMath::Length3( newCtlVec ) < CbxMath::nearZero ) {
		out->ctlVec[ 1 ] = 0.0;
		return( false );
	}

	result = ElementInCtlVecPlus( in->sub, axisInCtlPlaneDeg, defaultNormalInCtlPlaneDeg );
	if( result == true )
		return( true );

	CbxMath::Cross3( newCtlVec, axisInCtlPlaneDeg, normalInCtlPlaneDeg );
	len = CbxMath::Length3( normalInCtlPlaneDeg );
	if( len < CbxMath::nearZero ) {
		out->ctlVec[ 1 ] = CbxMath::Length3( newCtlVec ) / CbxMath::Length3( defaultNormalInCtlPlaneDeg );
		return( false );
	}
	CbxMath::Scale3( normalInCtlPlaneDeg, 1/len, normalInCtlPlaneDeg );

	result = Cbx3DtoOrg::CtlPlaneDeg( axisInCtlPlaneDeg, defaultNormalInCtlPlaneDeg, normalInCtlPlaneDeg, &out->ctlPlaneDeg );
	if( result == true )
		return( true );
	result = CbxVector::ElementAxisInCtlVecPlus( in->sub, out->ctlPlaneDeg, axis, &lenBaseCtlVec );
	if( result == true )
		return( true );
	result = Cbx3DtoOrg::CtlVec( axis, lenBaseCtlVec, newCtlVec, out->ctlVec );
	if( result == true )
		return( true );

	return( false );
}


bool  CbxCoreSub::
ElementInCtlVecPlus( const CbxCore::CtlVecForTransSub sub[ 2 ], double axisInCtlPlaneDeg[ 3 ],
	double defaultNormalInCtlPlaneDeg[ 3 ] )
{
	bool    result;
	short   i, j;
	double  len, wingHandle[ 2 ][ 3 ], crossBoundOnVertex[ 2 ][ 3 ], handleVec[ 2 ][ 3 ], normalOnVertex[ 2 ][ 3 ];

	len = CbxMath::Distance3( sub[ 0 ].vertexPos, sub[ 1 ].vertexPos );

	for( i=0; i<2; i++ ) {
		for( j=0; j<2; j++ ) {
			result = CbxOrgTo3D::HandleVec( sub[ i ].axisVec, sub[ i ].lenRadialVec[ j ],
						 sub[ i ].wingHandle[ j ], wingHandle[ j ] );
			if( result == true )
				return( true );
		}
		
		CbxElement::CrossBoundOnVertex( wingHandle, crossBoundOnVertex[ i ] );
		result = CbxOrgTo3D::HandleVec( sub[ i ].axisVec, len, sub[ i ].handleVec, handleVec[ i ] );
		if( result == true )
			return( true );

		CbxMath::Cross3( sub[ i ].axisVec[ 0 ], sub[ i ].axisVec[ 1 ], normalOnVertex[ i ] );
		CbxMath::Normalize3( normalOnVertex[ i ], CbxMath::zero3, normalOnVertex[ i ] );
	}

	CbxElement::AxisInCtlPlaneDeg( sub[ 0 ].vertexPos, handleVec[ 0 ], handleVec[ 1 ],
		sub[ 1 ].vertexPos, axisInCtlPlaneDeg );
	CbxElement::DefaultNormalInCtlPlaneDeg( normalOnVertex[ 0 ], axisInCtlPlaneDeg, normalOnVertex[ 1 ],
		defaultNormalInCtlPlaneDeg );

	return( false );
}


bool  CbxCoreSub::
DispChkOnEdgeSurface( const CbxCore::SurfaceForPick* data, bool* dispChk )
{
	CbxErrChk::ChkOnEdgeSurface  data2;

	CbxErrChk::ConvSurfaceForPickToChkOnEdgeSurface( data, &data2 );

	return( CbxErrChk::DispChkOnEdgeSurface( &data2, dispChk ) );
}
