/* bxErrChk.cpp */

#include <math.h>
#include "bxErrChk.h"
#include "../../Math/bxMath.h"
#include "../OrgTo3D/bxOrgTo3D.h"


const double  CbxErrChk::minHandleVec1          =   0.05;
const double  CbxErrChk::maxHandleVec1          =   1.0;
const double  CbxErrChk::minHandleVec0to0       =   5.0;
const double  CbxErrChk::limCtlPlaneDeg         =  45.0;
const double  CbxErrChk::limCtlVec0             =  45.0;
const double  CbxErrChk::minCtlVec1             =   0.15;
const double  CbxErrChk::maxCtlVec1             =   3.0;
const double  CbxErrChk::maxRadialToHandle		=  90.0;



ushort  CbxErrChk::
MaxNumRadial( void )
{
	return( ( ushort )( ( 360.0 - CbxMath::nearZero ) / minHandleVec0to0 ) );
}


double  CbxErrChk::
MaxHandleVec0to0( void )
{
	return( 180.0 - minHandleVec0to0 );
}


bool  CbxErrChk::
ErrChkAxisVec( const double axisVec[ 2 ][ 3 ] )
{
	double  normal[ 3 ], len;

	CbxMath::Cross3( axisVec[ 0 ], axisVec[ 1 ], normal );
	len = CbxMath::Length3( normal );
	if( len < CbxMath::nearZero )
		return( true );

	return( false );
}


bool  CbxErrChk::
ErrChkHandleVec1( const double handleVec1 /* handleVec[ 1 ] */ )
{
	if( minHandleVec1 > handleVec1 || handleVec1 > maxHandleVec1 )
		return( true );

	return( false );
}


bool  CbxErrChk::
ErrChkCtlPlaneDeg( const double ctlPlaneDeg )
{
	double  ctlPlaneDeg2;

	ctlPlaneDeg2 = CbxMath::RegularDeg( ctlPlaneDeg );
	if( -limCtlPlaneDeg > ctlPlaneDeg2 || ctlPlaneDeg2 > limCtlPlaneDeg )
		return( true );

	return( false );
}


bool  CbxErrChk::
ErrChkCtlVec( const double ctlVec[ 2 ][ 2 ] )
{
	double  ctlVec0[ 2 ];
	short   i;

	for( i=0; i<2; i++ )
		ctlVec0[ i ] = CbxMath::RegularDeg( ctlVec[ i ][ 0 ] );

	if( -limCtlVec0 > ctlVec0[ 0 ] || ctlVec0[ 0 ] > limCtlVec0 )
		return( true );
	if( -( 180.0 - limCtlVec0 ) < ctlVec0[ 1 ] && ctlVec0[ 1 ] < ( 180.0 - limCtlVec0 ) )
		return( true );

	for( i=0; i<2; i++ ) {
		if( minCtlVec1 > ctlVec[ i ][ 1 ] || ctlVec[ i ][ 1 ] > maxCtlVec1 )
			return( true );
	}

	return( false );
}


void  CbxErrChk::
CorrectHandleVec1( double* handleVec1 /* handleVec[ 1 ] */ )
{
	if( minHandleVec1 > *handleVec1 )
		*handleVec1 = minHandleVec1;
	if( *handleVec1 > maxHandleVec1 )
		*handleVec1 = maxHandleVec1;
}


void  CbxErrChk::
CorrectCtlPlaneDeg( double* ctlPlaneDeg )
{
	double  ctlPlaneDeg2;

	ctlPlaneDeg2 = CbxMath::RegularDeg( *ctlPlaneDeg );
	if( -limCtlPlaneDeg > ctlPlaneDeg2 )
		*ctlPlaneDeg = -limCtlPlaneDeg;
	if( ctlPlaneDeg2 > limCtlPlaneDeg )
		*ctlPlaneDeg =  limCtlPlaneDeg;
}


void  CbxErrChk::
CorrectCtlVec( const short minusPlusNo, double ctlVec[ 2 ] )
{
	double  ctlVec0;

	ctlVec0 = CbxMath::RegularDeg( ctlVec[ 0 ] );

	if( minusPlusNo == 0 ) {
		if( -limCtlVec0 > ctlVec0 )
			ctlVec[ 0 ] = -limCtlVec0;
		if( ctlVec0 > limCtlVec0 )
			ctlVec[ 0 ] =  limCtlVec0;
	}
	else {
		if( -( 180.0 - limCtlVec0 ) < ctlVec0 && ctlVec0 < ( 180.0 - limCtlVec0 ) ) {
			if( ctlVec0 < 0.0 )
				ctlVec[ 0 ] = -( 180.0 - limCtlVec0 );
			else
				ctlVec[ 0 ] =  ( 180.0 - limCtlVec0 );
		}
	}

	if( minCtlVec1 > ctlVec[ 1 ] )
		ctlVec[ 1 ] = minCtlVec1;
	if( ctlVec[ 1 ] > maxCtlVec1 )
		ctlVec[ 1 ] = maxCtlVec1;
}


void  CbxErrChk::
DispChkOnEdgeVertex( const ChkOnEdgeVertex* data, bool* dispChk )
{
	double  len;

	*dispChk = true;

	len = CbxMath::Distance3( data->sub[ 0 ].vertexPos, data->sub[ 1 ].vertexPos );
	if( len < CbxMath::nearZero )
		*dispChk = false;
}


bool  CbxErrChk::
DispChkOnEdgeRadial( const ChkOnEdgeRadial* data, bool* dispChk )
{
	bool   result;
	short  i;
	ChkOnEdgeVertex  data2;

	ConvRadialToVertex( data, &data2 );
	DispChkOnEdgeVertex( &data2, dispChk );
	if( *dispChk == false )
		return( false );

	for( i=0; i<2; i++ ) {
		result = DispChkOnEdgeRadialMain( data, i, dispChk );
		if( result == true )
			return( true );
		else if( *dispChk == false )
			return( false );
	}

	return( false );
}


bool  CbxErrChk::
DispChkOnEdgeWing( const ChkOnEdgeWing* data, bool* dispChk )
{
	bool   result;
	short  i;
	ChkOnEdgeVertex  data2;
	ChkOnEdgeRadial  data3;

	ConvWingToVertex( data, &data2 );
	ConvWingToRadial( data, &data3 );

	DispChkOnEdgeVertex( &data2, dispChk );
	if( *dispChk == false )
		return( false );

	for( i=0; i<2; i++ ) {
		result = DispChkOnEdgeRadialMain( &data3, i, dispChk );
		if( result == true )
			return( true );
		else if( *dispChk == false )
			return( false );

		DispChkOnEdgeWingMain( data, i, dispChk );
		if( *dispChk == false )
			return( false );
  }

	return( false );
}


bool  CbxErrChk::
DispChkOnEdgeSurface( const ChkOnEdgeSurface* data, bool* dispChk )
{
	short  i;
	bool   result;

	for( i=0; i<data->numEdge; i++ ) {
		result = DispChkOnEdgeWing( &data->edge[ i ], dispChk );
		if( result == true )
			return( true );
		else if( *dispChk == false )
			return( false );
	}

	return( false );
}


void  CbxErrChk::
ConvRadialToVertex( const ChkOnEdgeRadial* radial, ChkOnEdgeVertex* vertex )
{
	short  i;

	for( i=0; i<2; i++ )
		CbxMath::Copy3( radial->sub[ i ].vertexPos, vertex->sub[ i ].vertexPos );
}


void  CbxErrChk::
ConvWingToVertex( const ChkOnEdgeWing* wing, ChkOnEdgeVertex* vertex )
{
	short  i;

	for( i=0; i<2; i++ )
		CbxMath::Copy3( wing->sub[ i ].vertexPos, vertex->sub[ i ].vertexPos );
}


void  CbxErrChk::
ConvWingToRadial( const ChkOnEdgeWing* wing, ChkOnEdgeRadial* radial )
{
	short  i;

	for( i=0; i<2; i++ ) {
		CbxMath::Copy3( wing->sub[ i ].vertexPos, radial->sub[ i ].vertexPos );
		CbxMath::CopyAxis( wing->sub[ i ].axisVec, radial->sub[ i ].axisVec );
		radial->sub[ i ].handleVec0 = wing->sub[ i ].handleVec0;
	}
}


bool  CbxErrChk::
DispChkOnEdgeRadialMain( const ChkOnEdgeRadial* data, const short beginEndNo, bool* dispChk )
{
	double  radialVec[ 3 ], iHandleVec[ 2 ], handleVec[ 3 ], cosP, deg;
	bool    result, errResult;

	*dispChk = true;
	
	CbxMath::Subtract3( data->sub[ 1-beginEndNo ].vertexPos, data->sub[ beginEndNo ].vertexPos, radialVec );
	CbxMath::Normalize3( radialVec, CbxMath::zero3, radialVec, &errResult );
	if( errResult == true )
		return( true );

	iHandleVec[ 0 ] = data->sub[ beginEndNo ].handleVec0;
	iHandleVec[ 1 ] = 1.0 / 3.0;

	result = CbxOrgTo3D::HandleVec( data->sub[ beginEndNo ].axisVec, 1.0, iHandleVec, handleVec );
	if( result == true )
		return( true );
	CbxMath::Normalize3( handleVec, CbxMath::zero3, handleVec, &errResult );
	if( errResult == true )
		return( true );

	cosP = CbxMath::Dot3( radialVec, handleVec );
	deg  = fabs( CbxMath::CosToDeg( cosP, false, &errResult ) );
	if( errResult == true )
		return( true );

	if( deg > maxRadialToHandle )
		*dispChk = false;

	return( false );
}


void  CbxErrChk::
DispChkOnEdgeWingMain( const ChkOnEdgeWing* data, const short beginEndNo, bool* dispChk )
{
	double  degOfs[ 2 ];
	short   i;

	*dispChk = true;
	
	HandleVec0toDegOfs( data->sub[ beginEndNo ].wingHandle0[ 0 ], data->sub[ beginEndNo ].handleVec0, &degOfs[ 0 ] );
	HandleVec0toDegOfs( data->sub[ beginEndNo ].handleVec0, data->sub[ beginEndNo ].wingHandle0[ 1 ], &degOfs[ 1 ] );

	for( i=0; i<2; i++ ) {
		if( minHandleVec0to0 > degOfs[ i ] || degOfs[ i ] > MaxHandleVec0to0() ) {
			*dispChk = false;
			return;
		}
	}
}


void  CbxErrChk::
HandleVec0toDegOfs( const double handleDeg0, const double handleDeg1, double* degOfs )
{
	double  deg0, deg1, deg;

	deg0 = CbxMath::UnsignedRegularDeg( handleDeg0 );
	deg1 = CbxMath::UnsignedRegularDeg( handleDeg1 );
	if( deg0 <= deg1 )
		deg = deg1 - deg0;
	else
		deg = ( deg1 + 360.0 ) - deg0;

	*degOfs = CbxMath::RegularDeg( deg );
}


void  CbxErrChk::
ConvHandleForTransToChkOnEdgeVertex( const CbxCore::HandleForTrans* in, ChkOnEdgeVertex* out )
{
	CbxMath::Copy3( in->in.vertexPos,       out->sub[ 0 ].vertexPos );
	CbxMath::Copy3( in->in.radialVertexPos, out->sub[ 1 ].vertexPos );
}


void  CbxErrChk::
ConvHandleForTransPlusInToChkOnEdgeVertex( const CbxCore::HandleForTransPlusIn* in, ChkOnEdgeVertex* out )
{
	CbxMath::Copy3( in->vertexPos,       out->sub[ 0 ].vertexPos );
	CbxMath::Copy3( in->radialVertexPos, out->sub[ 1 ].vertexPos );
}


void  CbxErrChk::
ConvCtlPlaneDegForRotToChkOnEdgeRadial( const CbxCore::CtlPlaneDegForRot* in, ChkOnEdgeRadial* out )
{
	short  i;

	for( i=0; i<2; i++ ) {
		CbxMath::Copy3( in->in.sub[ i ].vertexPos, out->sub[ i ].vertexPos );
		CbxMath::CopyAxis( in->in.sub[ i ].axisVec, out->sub[ i ].axisVec );
		out->sub[ i ].handleVec0 = in->in.sub[ i ].handleVec[ 0 ];
	}
}


void  CbxErrChk::
ConvCtlVecForTransSubToChkOnEdgeWing( const CbxCore::CtlVecForTransSub in[ 2 ], ChkOnEdgeWing* out )
{
	short  i, j;

	for( i=0; i<2; i++ ) {
		CbxMath::Copy3( in[ i ].vertexPos, out->sub[ i ].vertexPos );
		CbxMath::CopyAxis( in[ i ].axisVec, out->sub[ i ].axisVec );
		out->sub[ i ].handleVec0 = in[ i ].handleVec[ 0 ];
		for( j=0; j<2; j++ )
			out->sub[ i ].wingHandle0[ j ] = in[ i ].wingHandle[ j ][ 0 ];
	}
}


void  CbxErrChk::
ConvSurfaceForPickToChkOnEdgeSurface( const CbxCore::SurfaceForPick* in, ChkOnEdgeSurface* out )
{
	short  i, j;

	out->numEdge = in->numVertex;

	for( i=0; i<in->numVertex; i++ ) {
		j = ( i + 1 ) % in->numVertex;

		CbxMath::Copy3( in->vertex[ i ].vertexPos, out->edge[ i ].sub[ 0 ].vertexPos );
		CbxMath::CopyAxis( in->vertex[ i ].axisVec, out->edge[ i ].sub[ 0 ].axisVec );
		out->edge[ i ].sub[ 0 ].handleVec0       = in->vertex[ i ].edge[ 1 ].handleVec[ 0 ];
		out->edge[ i ].sub[ 0 ].wingHandle0[ 0 ] = in->vertex[ i ].edge[ 1 ].wingHandle[ 0 ];
		out->edge[ i ].sub[ 0 ].wingHandle0[ 1 ] = in->vertex[ i ].edge[ 0 ].handleVec[ 0 ];

		CbxMath::Copy3( in->vertex[ j ].vertexPos, out->edge[ i ].sub[ 1 ].vertexPos );
		CbxMath::CopyAxis( in->vertex[ j ].axisVec, out->edge[ i ].sub[ 1 ].axisVec );
		out->edge[ i ].sub[ 1 ].handleVec0       = in->vertex[ j ].edge[ 0 ].handleVec[ 0 ];
		out->edge[ i ].sub[ 1 ].wingHandle0[ 0 ] = in->vertex[ j ].edge[ 1 ].handleVec[ 0 ];
		out->edge[ i ].sub[ 1 ].wingHandle0[ 1 ] = in->vertex[ j ].edge[ 0 ].wingHandle[ 0 ];
	}
}


void  CbxErrChk::
ConvCtlVecForDefaultToChkOnEdgeSurface( const CbxCore::CtlVecForDefaultIn* in, ChkOnEdgeSurface* out )
{
	short  i, j;

	out->numEdge = in->numVertex;

	for( i=0; i<in->numVertex; i++ ) {
		j = ( i + 1 ) % in->numVertex;

		CbxMath::Copy3( in->vertex[ i ].vertexPos, out->edge[ i ].sub[ 0 ].vertexPos );
		CbxMath::CopyAxis( in->vertex[ i ].axisVec, out->edge[ i ].sub[ 0 ].axisVec );
		out->edge[ i ].sub[ 0 ].handleVec0       = in->vertex[ i ].edge[ 1 ].handleVec[ 0 ];
		out->edge[ i ].sub[ 0 ].wingHandle0[ 0 ] = in->vertex[ i ].edge[ 1 ].wingHandle[ 0 ];
		out->edge[ i ].sub[ 0 ].wingHandle0[ 1 ] = in->vertex[ i ].edge[ 0 ].handleVec[ 0 ];

		CbxMath::Copy3( in->vertex[ j ].vertexPos, out->edge[ i ].sub[ 1 ].vertexPos );
		CbxMath::CopyAxis( in->vertex[ j ].axisVec, out->edge[ i ].sub[ 1 ].axisVec );
		out->edge[ i ].sub[ 1 ].handleVec0       = in->vertex[ j ].edge[ 0 ].handleVec[ 0 ];
		out->edge[ i ].sub[ 1 ].wingHandle0[ 0 ] = in->vertex[ j ].edge[ 1 ].handleVec[ 0 ];
		out->edge[ i ].sub[ 1 ].wingHandle0[ 1 ] = in->vertex[ j ].edge[ 0 ].wingHandle[ 0 ];
	}
}
