/* bxErrChk.h */

#include "../bxCore.h"
#include "bxCoreSub.h"


class CbxErrChk {
		friend  CbxCore;
		friend  CbxCoreSub;

	private:		// for CbxCore, CbxCoreSub:
		struct ChkOnEdgeVertexSub {
			double  vertexPos[ 3 ];
		};
		struct ChkOnEdgeVertex {
			ChkOnEdgeVertexSub  sub[ 2 ];
		};

		struct ChkOnEdgeRadialSub {
			double  vertexPos[ 3 ];
			double  axisVec[ 2 ][ 3 ];
			double  handleVec0;				/* handleVec[ 0 ] */
		};
		struct ChkOnEdgeRadial {
			ChkOnEdgeRadialSub  sub[ 2 ];
		};

		struct ChkOnEdgeWingSub {
			double  vertexPos[ 3 ];
			double  axisVec[ 2 ][ 3 ];
			double  handleVec0;				/* handleVec[ 0 ] */
			double  wingHandle0[ 2 ];		/* wingHandle[ i ][ 0 ], i=0,1 */
		};
		struct ChkOnEdgeWing {
			ChkOnEdgeWingSub  sub[ 2 ];
		};

		struct ChkOnEdgeSurface {
			short          numEdge;
			ChkOnEdgeWing  edge[ 4 ];
		};

	private:		// for CbxCore:
		static ushort
		MaxNumRadial( void );

		static bool
		ErrChkAxisVec( const double axisVec[ 2 ][ 3 ] );
		static bool
		ErrChkHandleVec1( const double handleVec1 /* handleVec[ 1 ] */ );
		static bool
		ErrChkCtlPlaneDeg( const double ctlPlaneDeg );
		static bool
		ErrChkCtlVec( const double ctlVec[ 2 ][ 2 ] );
		static void
		DispChkOnEdgeVertex( const ChkOnEdgeVertex* data, bool* dispChk );
		static bool
		DispChkOnEdgeRadial( const ChkOnEdgeRadial* data, bool* dispChk );
		static bool
		DispChkOnEdgeWing( const ChkOnEdgeWing* data, bool* dispChk );
		static bool
		DispChkOnEdgeSurface( const ChkOnEdgeSurface* data, bool* dispChk );

		static void
		ConvHandleForTransToChkOnEdgeVertex( const CbxCore::HandleForTrans* in, ChkOnEdgeVertex* out );
		static void
		ConvHandleForTransPlusInToChkOnEdgeVertex( const CbxCore::HandleForTransPlusIn* in, ChkOnEdgeVertex* out );
		static void
		ConvCtlPlaneDegForRotToChkOnEdgeRadial( const CbxCore::CtlPlaneDegForRot* in, ChkOnEdgeRadial* out );
		static void
		ConvCtlVecForTransSubToChkOnEdgeWing( const CbxCore::CtlVecForTransSub in[ 2 ], ChkOnEdgeWing* out );
		static void
		ConvSurfaceForPickToChkOnEdgeSurface( const CbxCore::SurfaceForPick* in, ChkOnEdgeSurface* out );
		static void
		ConvCtlVecForDefaultToChkOnEdgeSurface( const CbxCore::CtlVecForDefaultIn* in, ChkOnEdgeSurface* out );

	private:		// for CbxCoreSub:
		static void
		CorrectHandleVec1( double* handleVec1 /* handleVec[ 1 ] */ );
		static void
		CorrectCtlPlaneDeg( double* ctlPlaneDeg );
		static void
		CorrectCtlVec( const short minusPlusNo, double ctlVec[ 2 ] );

	private:
		static double
		MaxHandleVec0to0( void );

		static void
		ConvRadialToVertex( const ChkOnEdgeRadial* radial, ChkOnEdgeVertex* vertex );
		static void
		ConvWingToVertex( const ChkOnEdgeWing* wing, ChkOnEdgeVertex* vertex );
		static void
		ConvWingToRadial( const ChkOnEdgeWing* wing, ChkOnEdgeRadial* radial );
		static bool
		DispChkOnEdgeRadialMain( const ChkOnEdgeRadial* data, const short beginEndNo, bool* dispChk );
		static void
		DispChkOnEdgeWingMain( const ChkOnEdgeWing* data, const short beginEndNo, bool* dispChk );
		static void
		HandleVec0toDegOfs( const double handleDeg0, const double handleDeg1, double* degOfs );

	private:
		static const double  minHandleVec1;
		static const double  maxHandleVec1;
		static const double  minHandleVec0to0;
		static const double  limCtlPlaneDeg;
		static const double  limCtlVec0;
		static const double  minCtlVec1;
		static const double  maxCtlVec1;
		static const double  maxRadialToHandle;
};
