/* bxElement.cpp */

#include "bxElement.h"
#include "../../Math/bxMath.h"
#include "../OrgTo3D/bxOrgTo3D.h"



void  CbxElement::
AxisInCtlPlaneDeg( const double vertexPosOrg[ 3 ], const double handleVecOrg[ 3 ], const double handleVecRadial[ 3 ],
	const double vertexPosRadial[ 3 ], double axisOrg[ 3 ] )
{
	double  handlePosOrg[ 3 ], handlePosRadial[ 3 ], a[ 3 ], b[ 3 ];

	CbxMath::Add3( vertexPosOrg, handleVecOrg, handlePosOrg );
	CbxMath::Add3( vertexPosRadial, handleVecRadial, handlePosRadial );

	CbxMath::Ratio3( vertexPosOrg, handlePosOrg, 3.0/4.0, a );
	CbxMath::Ratio3( handlePosOrg, handlePosRadial, 2.0/4.0, b );

	CbxMath::Subtract3( b, a, axisOrg );
	CbxMath::Normalize3( axisOrg, CbxMath::zero3, axisOrg );
}


void  CbxElement::
CrossBoundOnVertex( const double wingHandleVec[ 2 ][ 3 ], double crossBoundOnVertex[ 3 ] )
{
	CbxMath::Subtract3( wingHandleVec[ 0 ], wingHandleVec[ 1 ], crossBoundOnVertex );
	CbxMath::Scale3( crossBoundOnVertex, 1.0/2.0, crossBoundOnVertex );
}


void  CbxElement::
DefaultNormalInCtlPlaneDeg( const double normalOrgOnVertex[ 3 ], const double axisOrgInCtlPlaneDeg[ 3 ],
	const double normalRadialOnVertex[ 3 ], double normalInCtlPlaneDeg[ 3 ] )
{
	double  normal0[ 3 ], tmp[ 3 ];

	CbxMath::Ratio3( normalOrgOnVertex, normalRadialOnVertex, 1.0/3.0, normal0 );

	CbxMath::Cross3( axisOrgInCtlPlaneDeg, normal0, tmp );
	CbxMath::Cross3( tmp, axisOrgInCtlPlaneDeg, normalInCtlPlaneDeg );
	CbxMath::Normalize3( normalInCtlPlaneDeg, CbxMath::zero3, normalInCtlPlaneDeg );
}


void  CbxElement::
SubAxisInCtlPlaneDeg( const double axisInCtlPlaneDeg[ 3 ], const double defaultNormalInCtlPlaneDeg[ 3 ],
	double subAxisInCtlPlaneDeg[ 3 ] )
{
	CbxMath::Cross3( axisInCtlPlaneDeg, defaultNormalInCtlPlaneDeg, subAxisInCtlPlaneDeg );
}


bool  CbxElement::
BaseCtlVec( const double handleVec[ 2 ][ 3 ], const double crossBoundOnVertex[ 2 ][ 3 ],
	const double normalInCtlPlaneDeg[ 3 ], const double axisInCtlPlaneDeg[ 3 ], double ctlVec[ 3 ] )
{
	double  dot[ 2 ], cosP, deg, mat[ 4 ][ 4 ], len0[ 2 ], len;
	bool    errResult;
	short   i;

	for( i=0; i<2; i++ ) {
		dot[ i ]  = CbxMath::Dot3( handleVec[ i ], crossBoundOnVertex[ i ] );
		len0[ i ] = CbxMath::Length3( crossBoundOnVertex[ i ] );
	}

	cosP = ( 1.0 - 1.0/3.0 ) * dot[ 0 ] + 1.0/3.0 * dot[ 1 ];
	len  = ( len0[ 0 ] + len0[ 1 ] ) / 2.0;

	deg = CbxMath::CosToDeg( cosP, true, &errResult );
	if( errResult == true )
		return( true );

	CbxMath::RotateAxis4x4( deg, normalInCtlPlaneDeg, mat, &errResult );
	if( errResult == true )
		return( true );
	CbxMath::Transform3( axisInCtlPlaneDeg, mat, ctlVec );

	CbxMath::Normalize3( ctlVec, CbxMath::zero3, ctlVec );
	CbxMath::Scale3( ctlVec, len, ctlVec );

	return( false );
}


bool  CbxElementEx::
ElementOnCtlVec( const CbxCore::CtlVecForTransSub sub[ 2 ], const double ctlPlaneDeg, double handleVec[ 2 ][ 3 ],
	double crossBoundOnVertex[ 2 ][ 3 ], double axisInCtlPlaneDeg[ 3 ], double normalInCtlPlaneDeg[ 3 ] )
{
	bool    result;
	short   i, j;
	double  len, wingHandle[ 2 ][ 3 ], normalOnVertex[ 2 ][ 3 ], defaultNormalInCtlPlaneDeg[ 3 ];

	len = CbxMath::Distance3( sub[ 0 ].vertexPos, sub[ 1 ].vertexPos );

	for( i=0; i<2; i++ ) {
		for( j=0; j<2; j++ ) {
			result = CbxOrgTo3D::HandleVec( sub[ i ].axisVec, sub[ i ].lenRadialVec[ j ],
						 sub[ i ].wingHandle[ j ], wingHandle[ j ] );
			if( result == true )
				return( true );
		}
		
		CbxElement::CrossBoundOnVertex( wingHandle, crossBoundOnVertex[ i ] );
		result = CbxOrgTo3D::HandleVec( sub[ i ].axisVec, len, sub[ i ].handleVec, handleVec[ i ] );
		if( result == true )
			return( true );

		CbxMath::Cross3( sub[ i ].axisVec[ 0 ], sub[ i ].axisVec[ 1 ], normalOnVertex[ i ] );
		CbxMath::Normalize3( normalOnVertex[ i ], CbxMath::zero3, normalOnVertex[ i ] );
	}

	CbxElement::AxisInCtlPlaneDeg( sub[ 0 ].vertexPos, handleVec[ 0 ], handleVec[ 1 ],
		sub[ 1 ].vertexPos, axisInCtlPlaneDeg );
	CbxElement::DefaultNormalInCtlPlaneDeg( normalOnVertex[ 0 ], axisInCtlPlaneDeg, normalOnVertex[ 1 ],
		defaultNormalInCtlPlaneDeg );
	
	result = CbxOrgTo3D::NormalInCtlPlaneDeg( axisInCtlPlaneDeg, defaultNormalInCtlPlaneDeg, ctlPlaneDeg,
				 normalInCtlPlaneDeg );
	if( result == true )
		return( true );

	return( false );
}


bool  CbxElementEx::
ElementAxisInCtlVec( const CbxCore::CtlVecForTrans* ctl, double axis[ 2 ][ 3 ], double* lenBaseCtlVec )
{
	bool    result, errResult;
	double  handleVec[ 2 ][ 3 ], crossBoundOnVertex[ 2 ][ 3 ];
	double  axisInCtlPlaneDeg[ 3 ], normalInCtlPlaneDeg[ 3 ], baseCtlVec[ 3 ];

	result = ElementOnCtlVec( ctl->in.sub, ctl->in.ctlPlaneDeg, handleVec, crossBoundOnVertex,
				 axisInCtlPlaneDeg, normalInCtlPlaneDeg );
	if( result == true )
		return( true );

	result = CbxElement::BaseCtlVec( handleVec, crossBoundOnVertex, normalInCtlPlaneDeg, axisInCtlPlaneDeg, baseCtlVec );
	if( result == true )
		return( true );

	*lenBaseCtlVec = CbxMath::Length3( baseCtlVec );

	CbxMath::Copy3( baseCtlVec, axis[ 0 ] );
	CbxMath::Copy3( axisInCtlPlaneDeg, axis[ 1 ] );
	CbxMath::RegularAxis( axis, axis, &errResult );
	if( errResult == true )
		return( true );

	return( false );
}
