/* bxOrgTo3D.cpp */

#include "bxOrgTo3D.h"
#include "../../Math/bxMath.h"
#include "../Element/bxElement.h"


bool  CbxOrgTo3D::
HandleVec( const double axisVec[ 2 ][ 3 ], const double lenRadialVec, const double iHandleVec[ 2 ],
	double oHandleVec[ 3 ] )
{
	bool    errResult;
	double  axisVec2[ 2 ][ 3 ], normal[ 3 ], mat[ 4 ][ 4 ], len;

	CbxMath::RegularAxis( axisVec, axisVec2, &errResult );
	if( errResult == true )
		return( true );

	CbxMath::Cross3( axisVec2[ 0 ], axisVec2[ 1 ], normal );
	CbxMath::RotateAxis4x4( iHandleVec[ 0 ], normal, mat, &errResult );
	if( errResult == true )
		return( true );
	CbxMath::Transform3( axisVec2[ 0 ], mat, oHandleVec );

	len = lenRadialVec * iHandleVec[ 1 ];
	CbxMath::Scale3( oHandleVec, len, oHandleVec );

	return( false );
}


bool  CbxOrgTo3D::
NormalInCtlPlaneDeg( const double axisInCtlPlaneDeg[ 3 ], const double defaultNormalInCtlPlaneDeg[ 3 ],
	const double ctlPlaneDeg, double normalInCtlPlaneDeg[ 3 ] )
{
	bool    errResult;
	double  mat[ 4 ][ 4 ];

	CbxMath::RotateAxis4x4( ctlPlaneDeg, axisInCtlPlaneDeg, mat, &errResult );
	if( errResult == true )
		return( true );

	CbxMath::Transform3( defaultNormalInCtlPlaneDeg, mat, normalInCtlPlaneDeg );

	return( false );
}


bool  CbxOrgTo3D::
CtlVec( const double axisVec[ 2 ][ 3 ], const double lenBaseCtlVec, const double iCtlVec[ 2 ], double oCtlVec[ 3 ] )
{
	bool    errResult;
	double  axisVec2[ 2 ][ 3 ], normal[ 3 ], mat[ 4 ][ 4 ], len;

	CbxMath::RegularAxis( axisVec, axisVec2, &errResult );
	if( errResult == true )
		return( true );

	CbxMath::Cross3( axisVec2[ 0 ], axisVec2[ 1 ], normal );
	CbxMath::RotateAxis4x4( iCtlVec[ 0 ], normal, mat, &errResult );
	if( errResult == true )
		return( true );
	CbxMath::Transform3( axisVec2[ 0 ], mat, oCtlVec );

	len = lenBaseCtlVec * iCtlVec[ 1 ];
	CbxMath::Scale3( oCtlVec, len, oCtlVec );

	return( false );
}


bool  CbxOrgTo3Dex::
GetHandle( const CbxCore::HandleForTrans* handle, double handleVec[ 3 ] )
{
	bool    result;
	double  lenRadialVec;

	lenRadialVec = CbxMath::Distance3( handle->in.radialVertexPos, handle->in.vertexPos );

	result = CbxOrgTo3D::HandleVec( handle->in.axisVec, lenRadialVec, handle->out.handleVec, handleVec );
	if( result == true )
		return( true );

	return( false );
}


bool  CbxOrgTo3Dex::
GetCtlInCtlVec( const CbxCore::CtlVecForTrans* ctl, double ctlVec[ 3 ] )
{
	bool    result;
	double  axisVec[ 2 ][ 3 ], lenBaseCtlVec;

	result = CbxElementEx::ElementAxisInCtlVec( ctl, axisVec, &lenBaseCtlVec );
	if( result == true )
		return( true );

	result = CbxOrgTo3D::CtlVec( axisVec, lenBaseCtlVec, ctl->out.ctlVec, ctlVec );
	if( result == true )
		return( true );

	return( false );
}
