/* bxRect.cpp */

#include "bxRect.h"
#include "../Surface/bxSurface.h"
#include "../Vector/bxVector.h"
#include "../../Math/bxMath.h"


bool    CbxRect::enableForSaveTime		= false;
double  CbxRect::puForSaveTime[ 3 ];
double  CbxRect::pvForSaveTime[ 3 ];



bool  CbxRect::
ConvSurface( const CbxCore::SurfaceForPickVertex vertex[ 4 ], CbxCore::RectPatch* rect )
{
	bool    result;
	double  f[ 4 ][ 4 ][ 3 ], a[ 4 ][ 5 ][ 3 ];

	result = CbxSurface::GetElement( 4, vertex, f, a );
	if( result == true )
		return( true );

	ConvSurfaceMain( vertex, ( const CbxSurface::Double4x3* )f, ( const CbxSurface::Double5x3* )a, rect );

	return( false );
}


void  CbxRect::
ConvSurfaceMain( const CbxCore::SurfaceForPickVertex vertex[ 4 ], const double f[ 4 ][ 4 ][ 3 ],
	const double a[ 4 ][ 5 ][ 3 ], CbxCore::RectPatch* rect )
{
	double  f2[ 4 ][ 5 ][ 3 ];
	double  xu0[ 4 ][ 5 ][ 3 ], xu1[ 4 ][ 5 ][ 3 ], xv0[ 5 ][ 4 ][ 3 ], xv1[ 5 ][ 4 ][ 3 ];
	double  xu[ 5 ][ 5 ][ 3 ], xv[ 5 ][ 5 ][ 3 ];
	short   i;

	for( i=0; i<4; i++ )
		CbxSurface::OneDimUpVecCross( 3, f[ i ], f2[ i ] );

	ConvSurfaceXu0( vertex, ( const CbxSurface::Double5x3* )f2, a, xu0 );
	ConvSurfaceXu1( vertex, ( const CbxSurface::Double5x3* )f2, a, xu1 );
	ConvSurfaceXv0( vertex, ( const CbxSurface::Double5x3* )f2, a, xv0 );
	ConvSurfaceXv1( vertex, ( const CbxSurface::Double5x3* )f2, a, xv1 );

	ConvSurfaceXu( ( const CbxSurface::Double5x3* )xu0, ( const CbxSurface::Double5x3* )xu1, xu );
	ConvSurfaceXv( ( const CbxSurface::Double4x3* )xv0, ( const CbxSurface::Double4x3* )xv1, xv );

	ConvSurfaceSub( ( const CbxSurface::Double5x3* )xu, ( const CbxSurface::Double5x3* )xv, rect );
}


void  CbxRect::
ConvSurfaceXu0( const CbxCore::SurfaceForPickVertex vertex[ 4 ], const double f[ 4 ][ 5 ][ 3 ],
	const double a[ 4 ][ 5 ][ 3 ], double xu0[ 4 ][ 5 ][ 3 ] )
{
	double  p[ 4 ][ 3 ], p2[ 5 ][ 3 ];
	short   i, no;

	no = 3;
	CbxVector::Pos( no, 4, vertex, p );
	CbxSurface::OneDimUpPos( 3, p, p2 );
	for( i=0; i<5; i++ ) {
		CbxMath::Copy3( p2[ i ], xu0[ 0 ][ 4-i ] );
		CbxMath::Add3( p2[ i ], a[ no ][ i ], xu0[ 1 ][ 4-i ] );
	}

	no = 1;
	CbxVector::Pos( no, 4, vertex, p );
	CbxSurface::OneDimUpPos( 3, p, p2 );
	for( i=0; i<5; i++ ) {
		CbxMath::Copy3( p2[ i ], xu0[ 3 ][ i ] );
		CbxMath::Add3( p2[ i ], f[ no ][ i ], xu0[ 2 ][ i ] );
	}
}


void  CbxRect::
ConvSurfaceXu1( const CbxCore::SurfaceForPickVertex vertex[ 4 ], const double f[ 4 ][ 5 ][ 3 ],
	const double a[ 4 ][ 5 ][ 3 ], double xu1[ 4 ][ 5 ][ 3 ] )
{
	double  p[ 4 ][ 3 ], p2[ 5 ][ 3 ];
	short   i, no;

	no = 3;
	CbxVector::Pos( no, 4, vertex, p );
	CbxSurface::OneDimUpPos( 3, p, p2 );
	for( i=0; i<5; i++ ) {
		CbxMath::Copy3( p2[ i ], xu1[ 0 ][ 4-i ] );
		CbxMath::Add3( p2[ i ], f[ no ][ i ], xu1[ 1 ][ 4-i ] );
	}

	no = 1;
	CbxVector::Pos( no, 4, vertex, p );
	CbxSurface::OneDimUpPos( 3, p, p2 );
	for( i=0; i<5; i++ ) {
		CbxMath::Copy3( p2[ i ], xu1[ 3 ][ i ] );
		CbxMath::Add3( p2[ i ], a[ no ][ i ], xu1[ 2 ][ i ] );
	}
}


void  CbxRect::
ConvSurfaceXv0( const CbxCore::SurfaceForPickVertex vertex[ 4 ], const double f[ 4 ][ 5 ][ 3 ],
	const double a[ 4 ][ 5 ][ 3 ], double xv0[ 5 ][ 4 ][ 3 ] )
{
	double  p[ 4 ][ 3 ], p2[ 5 ][ 3 ];
	short   i, no;

	no = 0;
	CbxVector::Pos( no, 4, vertex, p );
	CbxSurface::OneDimUpPos( 3, p, p2 );
	for( i=0; i<5; i++ ) {
		CbxMath::Copy3( p2[ i ], xv0[ i ][ 0 ] );
		CbxMath::Add3( p2[ i ], a[ no ][ i ], xv0[ i ][ 1 ] );
	}

	no = 2;
	CbxVector::Pos( no, 4, vertex, p );
	CbxSurface::OneDimUpPos( 3, p, p2 );
	for( i=0; i<5; i++ ) {
		CbxMath::Copy3( p2[ i ], xv0[ 4-i ][ 3 ] );
		CbxMath::Add3( p2[ i ], f[ no ][ i ], xv0[ 4-i ][ 2 ] );
	}
}


void  CbxRect::
ConvSurfaceXv1( const CbxCore::SurfaceForPickVertex vertex[ 4 ], const double f[ 4 ][ 5 ][ 3 ],
	const double a[ 4 ][ 5 ][ 3 ], double xv1[ 5 ][ 4 ][ 3 ] )
{
	double  p[ 4 ][ 3 ], p2[ 5 ][ 3 ];
	short   i, no;

	no = 0;
	CbxVector::Pos( no, 4, vertex, p );
	CbxSurface::OneDimUpPos( 3, p, p2 );
	for( i=0; i<5; i++ ) {
		CbxMath::Copy3( p2[ i ], xv1[ i ][ 0 ] );
		CbxMath::Add3( p2[ i ], f[ no ][ i ], xv1[ i ][ 1 ] );
	}

	no = 2;
	CbxVector::Pos( no, 4, vertex, p );
	CbxSurface::OneDimUpPos( 3, p, p2 );
	for( i=0; i<5; i++ ) {
		CbxMath::Copy3( p2[ i ], xv1[ 4-i ][ 3 ] );
		CbxMath::Add3( p2[ i ], a[ no ][ i ], xv1[ 4-i ][ 2 ] );
	}
}


void  CbxRect::
ConvSurfaceXu( const double xu0[ 4 ][ 5 ][ 3 ], const double xu1[ 4 ][ 5 ][ 3 ], double xu[ 5 ][ 5 ][ 3 ] )
{
	short  i;

	for( i=0; i<5; i++ ) {
		CbxMath::Copy3( xu0[ 0 ][ i ], xu[ 0 ][ i ] );
		CbxMath::Ratio3( xu0[ 1 ][ i ], xu1[ 0 ][ i ], 1.0/4.0, xu[ 1 ][ i ] );
		CbxMath::Ratio3( xu0[ 2 ][ i ], xu1[ 1 ][ i ], 2.0/4.0, xu[ 2 ][ i ] );
		CbxMath::Ratio3( xu0[ 3 ][ i ], xu1[ 2 ][ i ], 3.0/4.0, xu[ 3 ][ i ] );
		CbxMath::Copy3( xu1[ 3 ][ i ], xu[ 4 ][ i ] );
	}
}


void  CbxRect::
ConvSurfaceXv( const double xv0[ 5 ][ 4 ][ 3 ], const double xv1[ 5 ][ 4 ][ 3 ], double xv[ 5 ][ 5 ][ 3 ] )
{
	short  i;

	for( i=0; i<5; i++ ) {
		CbxMath::Copy3( xv0[ i ][ 0 ], xv[ i ][ 0 ] );
		CbxMath::Ratio3( xv0[ i ][ 1 ], xv1[ i ][ 0 ], 1.0/4.0, xv[ i ][ 1 ] );
		CbxMath::Ratio3( xv0[ i ][ 2 ], xv1[ i ][ 1 ], 2.0/4.0, xv[ i ][ 2 ] );
		CbxMath::Ratio3( xv0[ i ][ 3 ], xv1[ i ][ 2 ], 3.0/4.0, xv[ i ][ 3 ] );
		CbxMath::Copy3( xv1[ i ][ 3 ], xv[ i ][ 4 ] );
	}
}


void  CbxRect::
ConvSurfaceSub( const double xu[ 5 ][ 5 ][ 3 ], const double xv[ 5 ][ 5 ][ 3 ], CbxCore::RectPatch* rect )
{
	short  u, v;

	for( v=0; v<5; v++ ) {
		for( u=0; u<5; u++ ) {
			CbxMath::Copy3( xu[ u ][ v ], rect->data[ 0 ].p[ u ][ v ].pos );
			CbxMath::Copy3( xv[ u ][ v ], rect->data[ 1 ].p[ u ][ v ].pos );
		}
	}
}


void  CbxRect::
PosNormalSurface( const CbxCore::RectPatch* rect, const double uv[ 2 ], double pos[ 3 ], double normalVec[ 3 ] )
{
	SetDisableForSaveTime();

	PosSurfaceCom( rect, uv, pos );
	NormalSurfaceCom( rect, uv, normalVec );
}


void  CbxRect::
PosSurface( const CbxCore::RectPatch* rect, const double uv[ 2 ], double pos[ 3 ] )
{
	SetDisableForSaveTime();
	PosSurfaceCom( rect, uv, pos );
}


void  CbxRect::
PosSurfaceCom( const CbxCore::RectPatch* rect, const double uv[ 2 ], double pos[ 3 ] )
{
	if( uv[ 1 ] < CbxMath::nearZero ) {
		if( uv[ 0 ] < CbxMath::nearZero ) {
			CbxMath::Copy3( rect->data[ 0 ].p[ 0 ][ 0 ].pos, pos );
			return;
		}
		else if( uv[ 0 ] > ( 1.0 - CbxMath::nearZero ) ) {
			CbxMath::Copy3( rect->data[ 0 ].p[ 4 ][ 0 ].pos, pos );
			return;
		}
	}
	else if( uv[ 1 ] > ( 1.0 - CbxMath::nearZero ) ) {
		if( uv[ 0 ] < CbxMath::nearZero ) {
			CbxMath::Copy3( rect->data[ 0 ].p[ 0 ][ 4 ].pos, pos );
			return;
		}
		else if( uv[ 0 ] > ( 1.0 - CbxMath::nearZero ) ) {
			CbxMath::Copy3( rect->data[ 0 ].p[ 4 ][ 4 ].pos, pos );
			return;
		}
	}

	PosSurfaceMain( rect, uv, pos );
}


void  CbxRect::
PosSurfaceMain( const CbxCore::RectPatch* rect, const double uv[ 2 ], double pos[ 3 ] )
{
	double  patch[ 5 ][ 5 ][ 3 ], pu[ 3 ], pv[ 3 ], uParam, vParam;
	short   u, v, i;

	if( ChkEnableForSaveTime( pu, pv ) == false ) {
		for( v=0; v<5; v++ ) {
			for( u=0; u<5; u++ )
				CbxMath::Copy3( rect->data[ 0 ].p[ u ][ v ].pos, patch[ u ][ v ] );
		}
		CbxSurface::PosSurface4( ( const CbxSurface::Double5x3* )patch, uv, pu );

		for( v=0; v<5; v++ ) {
			for( u=0; u<5; u++ )
				CbxMath::Copy3( rect->data[ 1 ].p[ u ][ v ].pos, patch[ u ][ v ] );
		}
		CbxSurface::PosSurface4( ( const CbxSurface::Double5x3* )patch, uv, pv );

		SetEnableForSaveTime( pu, pv );
	}

	uParam = uv[ 0 ] * ( 1.0 - uv[ 0 ] );
	vParam = uv[ 1 ] * ( 1.0 - uv[ 1 ] );
	for( i=0; i<3; i++ )
		pos[ i ] = 1.0 / ( uParam + vParam ) * ( vParam * pu[ i ] + uParam * pv[ i ] );
}


void  CbxRect::
NormalSurface( const CbxCore::RectPatch* rect, const double uv[ 2 ], double normalVec[ 3 ] )
{
	SetDisableForSaveTime();
	NormalSurfaceCom( rect, uv, normalVec );
}


void  CbxRect::
NormalSurfaceCom( const CbxCore::RectPatch* rect, const double uv[ 2 ], double normalVec[ 3 ] )
{
	double  du[ 3 ], dv[ 3 ];

	DiffUsurface( rect, uv, du );
	DiffVsurface( rect, uv, dv );
	CbxMath::Cross3( du, dv, normalVec );
	CbxMath::Normalize3( normalVec, CbxMath::zero3, normalVec );
}


void  CbxRect::
DiffUsurface( const CbxCore::RectPatch* rect, const double uv[ 2 ], double vec[ 3 ] )
{
	if( uv[ 1 ] < CbxMath::nearZero ) {
		if( uv[ 0 ] < CbxMath::nearZero ) {
			DiffUsurface00( rect, vec );
			return;
		}
		else if( uv[ 0 ] > ( 1.0 - CbxMath::nearZero ) ) {
			DiffUsurface10( rect, vec );
			return;
		}
	}
	else if( uv[ 1 ] > ( 1.0 - CbxMath::nearZero ) ) {
		if( uv[ 0 ] < CbxMath::nearZero ) {
			DiffUsurface01( rect, vec );
			return;
		}
		else if( uv[ 0 ] > ( 1.0 - CbxMath::nearZero ) ) {
			DiffUsurface11( rect, vec );
			return;
		}
	}

	DiffUsurfaceMain( rect, uv, vec );
}


void  CbxRect::
DiffUsurface00( const CbxCore::RectPatch* rect, double vec[ 3 ] )
{
	CbxMath::Subtract3( rect->data[ 0 ].p[ 1 ][ 0 ].pos, rect->data[ 0 ].p[ 0 ][ 0 ].pos, vec );
	CbxMath::Scale3( vec, 4.0, vec );
}


void  CbxRect::
DiffUsurface10( const CbxCore::RectPatch* rect, double vec[ 3 ] )
{
	CbxMath::Subtract3( rect->data[ 0 ].p[ 4 ][ 0 ].pos, rect->data[ 0 ].p[ 3 ][ 0 ].pos, vec );
	CbxMath::Scale3( vec, 4.0, vec );
}


void  CbxRect::
DiffUsurface01( const CbxCore::RectPatch* rect, double vec[ 3 ] )
{
	CbxMath::Subtract3( rect->data[ 0 ].p[ 1 ][ 4 ].pos, rect->data[ 0 ].p[ 0 ][ 4 ].pos, vec );
	CbxMath::Scale3( vec, 4.0, vec );
}


void  CbxRect::
DiffUsurface11( const CbxCore::RectPatch* rect, double vec[ 3 ] )
{
	CbxMath::Subtract3( rect->data[ 0 ].p[ 4 ][ 4 ].pos, rect->data[ 0 ].p[ 3 ][ 4 ].pos, vec );
	CbxMath::Scale3( vec, 4.0, vec );
}


void  CbxRect::
DiffUsurfaceMain( const CbxCore::RectPatch* rect, const double uv[ 2 ], double vec[ 3 ] )
{
	double  patch[ 5 ][ 5 ][ 3 ], pu[ 3 ], pv[ 3 ], vu[ 3 ], vv[ 3 ], uParam, vParam, param0, param1;
	short   u, v, i;
	bool    chkEnableForSaveTime;

	chkEnableForSaveTime = ChkEnableForSaveTime( pu, pv );

	for( v=0; v<5; v++ ) {
		for( u=0; u<5; u++ )
			CbxMath::Copy3( rect->data[ 0 ].p[ u ][ v ].pos, patch[ u ][ v ] );
	}
	if( chkEnableForSaveTime == false )
		CbxSurface::PosSurface4( ( const CbxSurface::Double5x3* )patch, uv, pu );
	CbxSurface::DiffUsurface4( ( const CbxSurface::Double5x3* )patch, uv, vu );

	for( v=0; v<5; v++ ) {
		for( u=0; u<5; u++ )
			CbxMath::Copy3( rect->data[ 1 ].p[ u ][ v ].pos, patch[ u ][ v ] );
	}
	if( chkEnableForSaveTime == false )
		CbxSurface::PosSurface4( ( const CbxSurface::Double5x3* )patch, uv, pv );
	CbxSurface::DiffUsurface4( ( const CbxSurface::Double5x3* )patch, uv, vv );

	if( chkEnableForSaveTime == false )
		SetEnableForSaveTime( pu, pv );

	uParam = uv[ 0 ] * ( 1.0 - uv[ 0 ] );
	vParam = uv[ 1 ] * ( 1.0 - uv[ 1 ] );
	param0 = ( ( 1.0 - 2.0 * uv[ 0 ] ) * uv[ 1 ] * ( 1 - uv[ 1 ] ) ) / ( ( uParam + vParam ) * ( uParam + vParam ) );
	param1 = 1.0 / ( uParam + vParam );
	for( i=0; i<3; i++ )
		vec[ i ] = param0 * ( -pu[ i ] + pv[ i ] ) + param1 * ( vParam * vu[ i ] + uParam * vv[ i ] );
}


void  CbxRect::
DiffVsurface( const CbxCore::RectPatch* rect, const double uv[ 2 ], double vec[ 3 ] )
{
	if( uv[ 1 ] < CbxMath::nearZero ) {
		if( uv[ 0 ] < CbxMath::nearZero ) {
			DiffVsurface00( rect, vec );
			return;
		}
		else if( uv[ 0 ] > ( 1.0 - CbxMath::nearZero ) ) {
			DiffVsurface10( rect, vec );
			return;
		}
	}
	else if( uv[ 1 ] > ( 1.0 - CbxMath::nearZero ) ) {
		if( uv[ 0 ] < CbxMath::nearZero ) {
			DiffVsurface01( rect, vec );
			return;
		}
		else if( uv[ 0 ] > ( 1.0 - CbxMath::nearZero ) ) {
			DiffVsurface11( rect, vec );
			return;
		}
	}

	DiffVsurfaceMain( rect, uv, vec );
}


void  CbxRect::
DiffVsurface00( const CbxCore::RectPatch* rect, double vec[ 3 ] )
{
	CbxMath::Subtract3( rect->data[ 0 ].p[ 0 ][ 1 ].pos, rect->data[ 0 ].p[ 0 ][ 0 ].pos, vec );
	CbxMath::Scale3( vec, 4.0, vec );
}


void  CbxRect::
DiffVsurface10( const CbxCore::RectPatch* rect, double vec[ 3 ] )
{
	CbxMath::Subtract3( rect->data[ 0 ].p[ 4 ][ 1 ].pos, rect->data[ 0 ].p[ 4 ][ 0 ].pos, vec );
	CbxMath::Scale3( vec, 4.0, vec );
}


void  CbxRect::
DiffVsurface01( const CbxCore::RectPatch* rect, double vec[ 3 ] )
{
	CbxMath::Subtract3( rect->data[ 0 ].p[ 0 ][ 4 ].pos, rect->data[ 0 ].p[ 0 ][ 3 ].pos, vec );
	CbxMath::Scale3( vec, 4.0, vec );
}


void  CbxRect::
DiffVsurface11( const CbxCore::RectPatch* rect, double vec[ 3 ] )
{
	CbxMath::Subtract3( rect->data[ 0 ].p[ 4 ][ 4 ].pos, rect->data[ 0 ].p[ 4 ][ 3 ].pos, vec );
	CbxMath::Scale3( vec, 4.0, vec );
}


void  CbxRect::
DiffVsurfaceMain( const CbxCore::RectPatch* rect, const double uv[ 2 ], double vec[ 3 ] )
{
	double  patch[ 5 ][ 5 ][ 3 ], pu[ 3 ], pv[ 3 ], vu[ 3 ], vv[ 3 ], uParam, vParam, param0, param1;
	short   u, v, i;
	bool    chkEnableForSaveTime;

	chkEnableForSaveTime = ChkEnableForSaveTime( pu, pv );

	for( v=0; v<5; v++ ) {
		for( u=0; u<5; u++ )
			CbxMath::Copy3( rect->data[ 0 ].p[ u ][ v ].pos, patch[ u ][ v ] );
	}
	if( chkEnableForSaveTime == false )
		CbxSurface::PosSurface4( ( const CbxSurface::Double5x3* )patch, uv, pu );
	CbxSurface::DiffVsurface4( ( const CbxSurface::Double5x3* )patch, uv, vu );

	for( v=0; v<5; v++ ) {
		for( u=0; u<5; u++ )
			CbxMath::Copy3( rect->data[ 1 ].p[ u ][ v ].pos, patch[ u ][ v ] );
	}
	if( chkEnableForSaveTime == false )
		CbxSurface::PosSurface4( ( const CbxSurface::Double5x3* )patch, uv, pv );
	CbxSurface::DiffVsurface4( ( const CbxSurface::Double5x3* )patch, uv, vv );

	if( chkEnableForSaveTime == false )
		SetEnableForSaveTime( pu, pv );

	uParam = uv[ 0 ] * ( 1.0 - uv[ 0 ] );
	vParam = uv[ 1 ] * ( 1.0 - uv[ 1 ] );
	param0 = ( uv[ 0 ] * ( 1 - uv[ 0 ] ) * ( 1.0 - 2.0 * uv[ 1 ] ) ) / ( ( uParam + vParam ) * ( uParam + vParam ) );
	param1 = 1.0 / ( uParam + vParam );
	for( i=0; i<3; i++ )
		vec[ i ] = param0 * ( pu[ i ] - pv[ i ] ) + param1 * ( vParam * vu[ i ] + uParam * vv[ i ] );
}


void  CbxRect::
SetDisableForSaveTime( void )
{
	enableForSaveTime = false;
}


bool  CbxRect::
ChkEnableForSaveTime( double pu[ 3 ], double pv[ 3 ] )
{
	if( enableForSaveTime == false )
		return( false );

	CbxMath::Copy3( puForSaveTime, pu );
	CbxMath::Copy3( pvForSaveTime, pv );

	return( true );
}


void  CbxRect::
SetEnableForSaveTime( const double pu[ 3 ], const double pv[ 3 ] )
{
	enableForSaveTime = true;

	CbxMath::Copy3( pu, puForSaveTime );
	CbxMath::Copy3( pv, pvForSaveTime );
}


bool  CbxRect::
GetDefaultCtlVec( const CbxCore::CtlVecForDefaultInVertex in[ 4 ], CbxCore::CtlVecForDefaultOutVertex out[ 4 ] )
{
	bool    result;
	double  posVec[ 4 ][ 2 ][ 3 ], srcCtlVec[ 4 ][ 2 ][ 3 ], dstCtlVec[ 4 ][ 2 ][ 3 ];

	result = GetPosToPos( in, posVec );
	if( result == true )
		return( true );
	result = CbxSurface::CtlVecOrgTo3D( 4, in, out, srcCtlVec );
	if( result == true )
		return( true );

	result = CbxSurface::GetDefaultCtlVec( 4, ( const CbxSurface::Double2x3* )posVec,
				 ( const CbxSurface::Double2x3* )srcCtlVec, dstCtlVec );
	if( result == true )
		return( true );

	result = CbxSurface::CtlVec3DtoOrg( 4, in, ( const CbxSurface::Double2x3* )dstCtlVec, out );
	if( result == true )
		return( true );

	return( false );
}


bool  CbxRect::
GetPosToPos( const CbxCore::CtlVecForDefaultInVertex vertex[], double posVec[ 4 ][ 2 ][ 3 ] )
{
	bool    result;
	short   i;

	for( i=0; i<4; i++ ) {
		result = GetPosToPosMain( i, vertex, posVec[ i ] );
		if( result == true )
			return( true );
	}

	return( false );
}


bool  CbxRect::
GetPosToPosMain( const short no, const CbxCore::CtlVecForDefaultInVertex vertex[], double posVec[ 2 ][ 3 ] )
{
	double  pos0[ 3 ], pos1[ 3 ];
	short   n2;

	n2 = ( no + 2 ) % 4;

	if( CbxVector::Pos1ForCtlVecForDefault( no, 4, vertex, pos0 ) == true )
		return( true );
	if( CbxVector::Pos2ForCtlVecForDefault( n2, 4, vertex, pos1 ) == true )
		return( true );
	CbxMath::Subtract3( pos1, pos0, posVec[ 0 ] );

	if( CbxVector::Pos2ForCtlVecForDefault( no, 4, vertex, pos0 ) == true )
		return( true );
	if( CbxVector::Pos1ForCtlVecForDefault( n2, 4, vertex, pos1 ) == true )
		return( true );
	CbxMath::Subtract3( pos1, pos0, posVec[ 1 ] );

	return( false );
}
