/* bxSurface.cpp */

#include <math.h>

#include "bxSurface.h"
#include "../../Math/bxMath.h"
#include "../Vector/bxVector.h"



bool  CbxSurface::
GetElement( const short numVertex, const CbxCore::SurfaceForPickVertex vertex[], double f[][ 4 ][ 3 ], double a[][ 5 ][ 3 ] )
{
	bool   result;
	short  i;

	for( i=0; i<numVertex; i++ ) {
		result = GetFormCtlVec( i, numVertex, vertex, f[ i ] );
		if( result == true )
			return( true );

		result = GetJoinCtlVec( i, numVertex, vertex, a[ i ] );
		if( result == true )
			return( true );
	}

	return( false );
}


bool  CbxSurface::
GetFormCtlVec( const short no, const short numVertex, const CbxCore::SurfaceForPickVertex vertex[], double f[ 4 ][ 3 ] )
{
	return( CbxVector::Handle( no, numVertex, vertex, f ) );
}


bool  CbxSurface::
GetJoinCtlVec( const short no, const short numVertex, const CbxCore::SurfaceForPickVertex vertex[], double a[ 5 ][ 3 ] )
{
	bool    result;
	double  d[ 4 ][ 3 ];

	result = GetBaseCtlVec( no, numVertex, vertex, d );
	if( result == true )
		return( true );

	result = GetJoinCtlVecMain( no, numVertex, vertex, d, a );
	if( result == true )
		return( true );

	return( false );
}


bool  CbxSurface::
GetBaseCtlVec( const short no, const short numVertex, const CbxCore::SurfaceForPickVertex vertex[], double d[ 4 ][ 3 ] )
{
	bool    result;
	double  f[ 4 ][ 3 ], g[ 4 ][ 3 ];
	short   i;

	result = CbxVector::Handle( no, numVertex, vertex, f );
	if( result == true )
		return( true );
	result = CbxVector::AntiHandle( no, numVertex, vertex, g );
	if( result == true )
		return( true );
	
	for( i=0; i<4; i++ ) {
		CbxMath::Subtract3( f[ i ], g[ i ], d[ i ] );
		CbxMath::Scale3( d[ i ], 1.0/2.0, d[ i ] );
	}

	return( false );
}


bool  CbxSurface::
GetJoinCtlVecMain( const short no, const short numVertex, const CbxCore::SurfaceForPickVertex vertex[],
	const double d[ 4 ][ 3 ], double a[ 5 ][ 3 ] )
{
	bool    result;
	double  e[ 4 ][ 3 ], ad[ 2 ][ 3 ], dd[ 2 ][ 3 ], ed[ 2 ][ 3 ];
	double  ae[ 2 ][ 3 ], ee[ 2 ][ 3 ], k[ 2 ], h[ 2 ];


	result = GetTanVecDim4( no, numVertex, vertex, e );
	if( result == true )
		return( true );
	result = GetBaseJoinCtlVecOnVertex( no, numVertex, vertex, ad );
	if( result == true )
		return( true );
	GetElementForJoinCtlVec( d, e, ad, dd, ed, a[ 0 ], a[ 4 ] );

	result = NormalizeCtlVec( ad, ed, ae, ee );
	if( result == true )
		return( true );
	result = GetRatioForJoinCtlVec( ee, ae, dd, ad, ed, k, h );
	if( result == true )
		return( true );

	GetJoinCtlVecSub( k, h, d, e, a[ 1 ], a[ 2 ], a[ 3 ] );

	return( false );
}


bool  CbxSurface::
GetTanVecDim4( const short no, const short numVertex, const CbxCore::SurfaceForPickVertex vertex[],
	double e[ 4 ][ 3 ] )
{
	bool    result;
	double  tan[ 3 ][ 3 ];

	result = CbxVector::Tan( no, numVertex, vertex, tan );
	if( result == true )
		return( true );

	OneDimUpVecTan( 3, tan, e );

	return( false );
}


bool  CbxSurface::
GetBaseJoinCtlVecOnVertex( const short no, const short numVertex, const CbxCore::SurfaceForPickVertex vertex[],
	double ad[ 2 ][ 3 ] )
{
	bool  result;

	result = CbxVector::Handle0( no, numVertex, vertex, ad[ 0 ] );
	if( result == true )
		return( true );
	result = CbxVector::Handle3( no, numVertex, vertex, ad[ 1 ] );
	if( result == true )
		return( true );

	return( false );
}


void  CbxSurface::
GetElementForJoinCtlVec( const double d[ 4 ][ 3 ], const double e[ 4 ][ 3 ], const double ad[ 2 ][ 3 ],
	double dd[ 2 ][ 3 ], double ed[ 2 ][ 3 ], double a0[ 3 ], double a4[ 3 ] )
{
	CbxMath::Copy3( d[ 0 ], dd[ 0 ] );
	CbxMath::Copy3( d[ 3 ], dd[ 1 ] );
	CbxMath::Copy3( e[ 0 ], ed[ 0 ] );
	CbxMath::Copy3( e[ 3 ], ed[ 1 ] );

	CbxMath::Copy3( ad[ 0 ], a0 );
	CbxMath::Copy3( ad[ 1 ], a4 );
}


bool  CbxSurface::
NormalizeCtlVec( const double ad[ 2 ][ 3 ], const double ed[ 2 ][ 3 ], double ae[ 2 ][ 3 ], double ee[ 2 ][ 3 ] )
{
	bool   errResult;
	short  i;

	for( i=0; i<2; i++ ) {
		CbxMath::Normalize3( ad[ i ], CbxMath::zero3, ae[ i ], &errResult );
		if( errResult == true )
			return( true );
		CbxMath::Normalize3( ed[ i ], CbxMath::zero3, ee[ i ], &errResult );
		if( errResult == true )
			return( true );
	}

	return( false );
}


bool  CbxSurface::
GetRatioForJoinCtlVec( const double ee[ 2 ][ 3 ], const double ae[ 2 ][ 3 ], const double dd[ 2 ][ 3 ],
	const double ad[ 2 ][ 3 ], const double ed[ 2 ][ 3 ], double k[ 2 ], double h[ 2 ] )
{
	double  tmp[ 3 ], f[ 3 ], fe[ 3 ], g[ 3 ], ge[ 3 ], dot;
	short   i;
	bool    errResult;

	for( i=0; i<2; i++ ) {
		CbxMath::Cross3( ee[ i ], dd[ i ], tmp );
		if( CbxMath::Length3( tmp ) < CbxMath::nearZero )
			return( true );
		CbxMath::Cross3( tmp, ee[ i ], f );
		CbxMath::Normalize3( f, CbxMath::zero3, fe, &errResult );
		if( errResult == true )
			return( true );
		dot = CbxMath::Dot3( ad[ i ], fe );
		if( fabs( dot ) < CbxMath::nearZero )
			return( true );
		k[ i ] = CbxMath::Length3( f ) / dot;

		CbxMath::Cross3( ae[ i ], dd[ i ], tmp );
		if( CbxMath::Length3( tmp ) < CbxMath::nearZero )
			h[ i ] = 0.0;
		else {
			CbxMath::Cross3( ae[ i ], tmp, g );
			CbxMath::Normalize3( g, CbxMath::zero3, ge, &errResult );
			if( errResult == true )
				return( true );
			dot = CbxMath::Dot3( ed[ i ], ge );
			if( fabs( dot ) < CbxMath::nearZero )
				return( true );
			h[ i ] = CbxMath::Length3( g ) / dot;
		}
	}

	return( false );
}


void  CbxSurface::
GetJoinCtlVecSub( const double k[ 2 ], const double h[ 2 ], const double d[ 4 ][ 3 ], const double e[ 4 ][ 3 ],
	double a1[ 3 ], double a2[ 3 ], double a3[ 3 ] )
{
	short   i, j;
	double  a[ 3 ][ 3 ];

	for( i=1; i<=3; i++ ) {
		for( j=0; j<3; j++ ) {
			a[ i-1 ][ j ] = ( 1.0 - i/4.0 ) * ( k[ 0 ] * d[ i   ][ j ] + h[ 0 ] * e[ i   ][ j ] )
								    + i/4.0 * ( k[ 1 ] * d[ i-1 ][ j ] + h[ 1 ] * e[ i-1 ][ j ] );
		}
	}

	CbxMath::Copy3( a[ 0 ], a1 );
	CbxMath::Copy3( a[ 1 ], a2 );
	CbxMath::Copy3( a[ 2 ], a3 );
}


void  CbxSurface::
OneDimUpPos( const short srcDim, const double src[/*srcDim+1*/][ 3 ], double dst[/*srcDim+2*/][ 3 ] )
{
	short  i, dstDim;

	dstDim = srcDim + 1;

	CbxMath::Copy3( src[ 0 ],      dst[ 0 ]      );
	CbxMath::Copy3( src[ srcDim ], dst[ dstDim ] );

	for( i=0; i<srcDim; i++ )
		CbxMath::Ratio3( src[ i ], src[ i+1 ], ( double )( srcDim-i )/( double )dstDim, dst[ i+1 ] );
}


void  CbxSurface::
OneDimUpVecTan( const short srcDim, const double src[/*srcDim*/][ 3 ], double dst[/*srcDim+1*/][ 3 ] )
{
	short   i, dstDim;
	double  s0[ 3 ], s1[ 3 ], d0[ 3 ], d1[ 3 ];

	dstDim = srcDim + 1;

	CbxMath::Copy3( CbxMath::zero3, s0 );
	CbxMath::Copy3( CbxMath::zero3, d0 );

	for( i=0; i<srcDim; i++ ) {
		CbxMath::Add3( s0, src[ i ], s1 );
		CbxMath::Ratio3( s0, s1, ( double )( srcDim-i )/( double )dstDim, d1 );
		CbxMath::Subtract3( d1, d0, dst[ i ] );
		CbxMath::Copy3( s1, s0 );
		CbxMath::Copy3( d1, d0 );
	}

	CbxMath::Copy3( s0, d1 );
	CbxMath::Subtract3( d1, d0, dst[ dstDim-1 ] );
}


void  CbxSurface::
OneDimUpVecCross( const short srcDim, const double src[/*srcDim+1*/][ 3 ], double dst[/*srcDim+2*/][ 3 ] )
{
	OneDimUpPos( srcDim, src, dst );
}


void  CbxSurface::
PosSurface4( const double patch[ 5 ][ 5 ][ 3 ], const double uv[ 2 ], double pos[ 3 ] )
{
	double  uv2[ 2 ], uv3[ 2 ], uv4[ 2 ], muv[ 2 ], muv2[ 2 ], muv3[ 2 ], muv4[ 2 ], param[ 2 ][ 5 ], p;
	short   i, u, v;

	for( i=0; i<2; i++ ) {
		uv2[ i ] = uv[ i ]  * uv[ i ];
		uv3[ i ] = uv2[ i ] * uv[ i ];
		uv4[ i ] = uv3[ i ] * uv[ i ];

		muv[ i ]  = 1.0 - uv[ i ];
		muv2[ i ] = muv[ i ]  * muv[ i ];
		muv3[ i ] = muv2[ i ] * muv[ i ];
		muv4[ i ] = muv3[ i ] * muv[ i ];

		param[ i ][ 0 ] = muv4[ i ];
		param[ i ][ 1 ] = 4.0 * uv[ i ]  * muv3[ i ];
		param[ i ][ 2 ] = 6.0 * uv2[ i ] * muv2[ i ];
		param[ i ][ 3 ] = 4.0 * uv3[ i ] * muv[ i ];
		param[ i ][ 4 ] = uv4[ i ];
	}

	CbxMath::Copy3( CbxMath::zero3, pos );
	for( i=0; i<3; i++ ) {
		for( v=0; v<5; v++ ) {
			p = 0.0;
			for( u=0; u<5; u++ )
				p += param[ 0 ][ u ] * patch[ u ][ v ][ i ];

			pos[ i ] += p * param[ 1 ][ v ];
		}
	}
}


void  CbxSurface::
DiffUsurface4( const double patch[ 5 ][ 5 ][ 3 ], const double uv[ 2 ], double vec[ 3 ] )
{
	double  uv2[ 2 ], uv3[ 2 ], uv4[ 2 ], muv[ 2 ], muv2[ 2 ], muv3[ 2 ], muv4[ 2 ], param[ 2 ][ 5 ], p;
	short   i, u, v, n;

	for( i=0; i<2; i++ ) {
		uv2[ i ] = uv[ i ]  * uv[ i ];
		uv3[ i ] = uv2[ i ] * uv[ i ];
		uv4[ i ] = uv3[ i ] * uv[ i ];

		muv[ i ]  = 1.0 - uv[ i ];
		muv2[ i ] = muv[ i ]  * muv[ i ];
		muv3[ i ] = muv2[ i ] * muv[ i ];
		muv4[ i ] = muv3[ i ] * muv[ i ];
	}

	n = 0;
	param[ n ][ 0 ] = -4.0 * muv3[ n ];
	param[ n ][ 1 ] = -4.0 * muv2[ n ] * ( 4.0 * uv[ n ] - 1.0 );
	param[ n ][ 2 ] = 12.0 *  uv[ n ]  * ( 2.0 * uv[ n ] - 1.0 ) * ( uv[ n ] - 1.0 );
	param[ n ][ 3 ] =  4.0 * uv2[ n ]  * ( 3.0 - 4.0 * uv[ n ] );
	param[ n ][ 4 ] =  4.0 * uv3[ n ];

	n = 1;
	param[ n ][ 0 ] = muv4[ n ];
	param[ n ][ 1 ] = 4.0 * uv[ n ]  * muv3[ n ];
	param[ n ][ 2 ] = 6.0 * uv2[ n ] * muv2[ n ];
	param[ n ][ 3 ] = 4.0 * uv3[ n ] * muv[ n ];
	param[ n ][ 4 ] = uv4[ n ];

	CbxMath::Copy3( CbxMath::zero3, vec );
	for( i=0; i<3; i++ ) {
		for( v=0; v<5; v++ ) {
			p = 0.0;
			for( u=0; u<5; u++ )
				p += param[ 0 ][ u ] * patch[ u ][ v ][ i ];

			vec[ i ] += p * param[ 1 ][ v ];
		}
	}
}


void  CbxSurface::
DiffVsurface4( const double patch[ 5 ][ 5 ][ 3 ], const double uv[ 2 ], double vec[ 3 ] )
{
	double  uv2[ 2 ], uv3[ 2 ], uv4[ 2 ], muv[ 2 ], muv2[ 2 ], muv3[ 2 ], muv4[ 2 ], param[ 2 ][ 5 ], p;
	short   i, u, v, n;

	for( i=0; i<2; i++ ) {
		uv2[ i ] = uv[ i ]  * uv[ i ];
		uv3[ i ] = uv2[ i ] * uv[ i ];
		uv4[ i ] = uv3[ i ] * uv[ i ];

		muv[ i ]  = 1.0 - uv[ i ];
		muv2[ i ] = muv[ i ]  * muv[ i ];
		muv3[ i ] = muv2[ i ] * muv[ i ];
		muv4[ i ] = muv3[ i ] * muv[ i ];
	}

	n = 0;
	param[ n ][ 0 ] = muv4[ n ];
	param[ n ][ 1 ] = 4.0 * uv[ n ]  * muv3[ n ];
	param[ n ][ 2 ] = 6.0 * uv2[ n ] * muv2[ n ];
	param[ n ][ 3 ] = 4.0 * uv3[ n ] * muv[ n ];
	param[ n ][ 4 ] = uv4[ n ];

	n = 1;
	param[ n ][ 0 ] = -4.0 * muv3[ n ];
	param[ n ][ 1 ] = -4.0 * muv2[ n ] * ( 4.0 * uv[ n ] - 1.0 );
	param[ n ][ 2 ] = 12.0 *  uv[ n ]  * ( 2.0 * uv[ n ] - 1.0 ) * ( uv[ n ] - 1.0 );
	param[ n ][ 3 ] =  4.0 * uv2[ n ]  * ( 3.0 - 4.0 * uv[ n ] );
	param[ n ][ 4 ] =  4.0 * uv3[ n ];

	CbxMath::Copy3( CbxMath::zero3, vec );
	for( i=0; i<3; i++ ) {
		for( v=0; v<5; v++ ) {
			p = 0.0;
			for( u=0; u<5; u++ )
				p += param[ 0 ][ u ] * patch[ u ][ v ][ i ];

			vec[ i ] += p * param[ 1 ][ v ];
		}
	}
}


void  CbxSurface::
PosSurface5( const double patch[ 6 ][ 6 ][ 3 ], const double uv[ 2 ], double pos[ 3 ] )
{
	double  uv2[ 2 ], uv3[ 2 ], uv4[ 2 ], uv5[ 2 ], muv[ 2 ], muv2[ 2 ], muv3[ 2 ], muv4[ 2 ], muv5[ 2 ];
	double  param[ 2 ][ 6 ], p;
	short   i, u, v;

	for( i=0; i<2; i++ ) {
		uv2[ i ] = uv[ i ]  * uv[ i ];
		uv3[ i ] = uv2[ i ] * uv[ i ];
		uv4[ i ] = uv3[ i ] * uv[ i ];
		uv5[ i ] = uv4[ i ] * uv[ i ];

		muv[ i ]  = 1.0 - uv[ i ];
		muv2[ i ] = muv[ i ]  * muv[ i ];
		muv3[ i ] = muv2[ i ] * muv[ i ];
		muv4[ i ] = muv3[ i ] * muv[ i ];
		muv5[ i ] = muv4[ i ] * muv[ i ];

		param[ i ][ 0 ] =  muv5[ i ];
		param[ i ][ 1 ] =  5.0 * uv[ i ]  * muv4[ i ];
		param[ i ][ 2 ] = 10.0 * uv2[ i ] * muv3[ i ];
		param[ i ][ 3 ] = 10.0 * uv3[ i ] * muv2[ i ];
		param[ i ][ 4 ] =  5.0 * uv4[ i ] * muv[ i ];
		param[ i ][ 5 ] =  uv5[ i ];
	}

	CbxMath::Copy3( CbxMath::zero3, pos );
	for( i=0; i<3; i++ ) {
		for( v=0; v<6; v++ ) {
			p = 0.0;
			for( u=0; u<6; u++ )
				p += param[ 0 ][ u ] * patch[ u ][ v ][ i ];

			pos[ i ] += p * param[ 1 ][ v ];
		}
	}
}


void  CbxSurface::
DiffUsurface5( const double patch[ 6 ][ 6 ][ 3 ], const double uv[ 2 ], double vec[ 3 ] )
{
	double  uv2[ 2 ], uv3[ 2 ], uv4[ 2 ], uv5[ 2 ], muv[ 2 ], muv2[ 2 ], muv3[ 2 ], muv4[ 2 ], muv5[ 2 ];
	double  param[ 2 ][ 6 ], p;
	short   i, u, v, n;

	for( i=0; i<2; i++ ) {
		uv2[ i ] = uv[ i ]  * uv[ i ];
		uv3[ i ] = uv2[ i ] * uv[ i ];
		uv4[ i ] = uv3[ i ] * uv[ i ];
		uv5[ i ] = uv4[ i ] * uv[ i ];

		muv[ i ]  = 1.0 - uv[ i ];
		muv2[ i ] = muv[ i ]  * muv[ i ];
		muv3[ i ] = muv2[ i ] * muv[ i ];
		muv4[ i ] = muv3[ i ] * muv[ i ];
		muv5[ i ] = muv4[ i ] * muv[ i ];
	}

	n = 0;
	param[ n ][ 0 ] = -5.0 * muv4[ n ];
	param[ n ][ 1 ] = -5.0 * muv3[ n ] * ( 5.0 * uv[ n ] - 1.0 );
	param[ n ][ 2 ] = 10.0 * muv2[ n ] * uv[ n ] * ( 2.0 - 5.0 * uv[ n ] );
	param[ n ][ 3 ] = 10.0 * uv2[ n ]  * ( 5.0 * uv[ n ] - 3.0 ) * ( uv[ n ] - 1.0 );
	param[ n ][ 4 ] =  5.0 * uv3[ n ]  * ( 4.0 - 5.0 * uv[ n ] );
	param[ n ][ 5 ] =  5.0 * uv4[ n ];

	n = 1;
	param[ n ][ 0 ] =  muv5[ n ];
	param[ n ][ 1 ] =  5.0 * uv[ n ]  * muv4[ n ];
	param[ n ][ 2 ] = 10.0 * uv2[ n ] * muv3[ n ];
	param[ n ][ 3 ] = 10.0 * uv3[ n ] * muv2[ n ];
	param[ n ][ 4 ] =  5.0 * uv4[ n ] * muv[ n ];
	param[ n ][ 5 ] =  uv5[ n ];

	CbxMath::Copy3( CbxMath::zero3, vec );
	for( i=0; i<3; i++ ) {
		for( v=0; v<6; v++ ) {
			p = 0.0;
			for( u=0; u<6; u++ )
				p += param[ 0 ][ u ] * patch[ u ][ v ][ i ];

			vec[ i ] += p * param[ 1 ][ v ];
		}
	}
}


void  CbxSurface::
DiffVsurface5( const double patch[ 6 ][ 6 ][ 3 ], const double uv[ 2 ], double vec[ 3 ] )
{
	double  uv2[ 2 ], uv3[ 2 ], uv4[ 2 ], uv5[ 2 ], muv[ 2 ], muv2[ 2 ], muv3[ 2 ], muv4[ 2 ], muv5[ 2 ];
	double  param[ 2 ][ 6 ], p;
	short   i, u, v, n;

	for( i=0; i<2; i++ ) {
		uv2[ i ] = uv[ i ]  * uv[ i ];
		uv3[ i ] = uv2[ i ] * uv[ i ];
		uv4[ i ] = uv3[ i ] * uv[ i ];
		uv5[ i ] = uv4[ i ] * uv[ i ];

		muv[ i ]  = 1.0 - uv[ i ];
		muv2[ i ] = muv[ i ]  * muv[ i ];
		muv3[ i ] = muv2[ i ] * muv[ i ];
		muv4[ i ] = muv3[ i ] * muv[ i ];
		muv5[ i ] = muv4[ i ] * muv[ i ];
	}

	n = 0;
	param[ n ][ 0 ] =  muv5[ n ];
	param[ n ][ 1 ] =  5.0 * uv[ n ]  * muv4[ n ];
	param[ n ][ 2 ] = 10.0 * uv2[ n ] * muv3[ n ];
	param[ n ][ 3 ] = 10.0 * uv3[ n ] * muv2[ n ];
	param[ n ][ 4 ] =  5.0 * uv4[ n ] * muv[ n ];
	param[ n ][ 5 ] =  uv5[ n ];

	n = 1;
	param[ n ][ 0 ] = -5.0 * muv4[ n ];
	param[ n ][ 1 ] = -5.0 * muv3[ n ] * ( 5.0 * uv[ n ] - 1.0 );
	param[ n ][ 2 ] = 10.0 * muv2[ n ] * uv[ n ] * ( 2.0 - 5.0 * uv[ n ] );
	param[ n ][ 3 ] = 10.0 * uv2[ n ]  * ( 5.0 * uv[ n ] - 3.0 ) * ( uv[ n ] - 1.0 );
	param[ n ][ 4 ] =  5.0 * uv3[ n ]  * ( 4.0 - 5.0 * uv[ n ] );
	param[ n ][ 5 ] =  5.0 * uv4[ n ];

	CbxMath::Copy3( CbxMath::zero3, vec );
	for( i=0; i<3; i++ ) {
		for( v=0; v<6; v++ ) {
			p = 0.0;
			for( u=0; u<6; u++ )
				p += param[ 0 ][ u ] * patch[ u ][ v ][ i ];

			vec[ i ] += p * param[ 1 ][ v ];
		}
	}
}


bool  CbxSurface::
CtlVecOrgTo3D( const short numVertex, const CbxCore::CtlVecForDefaultInVertex in[],
	const CbxCore::CtlVecForDefaultOutVertex out[], double ctlVec[][ 2 ][ 3 ] )
{
	bool   result;
	short  i;

	for( i=0; i<numVertex; i++ ) {
		result = CtlVecOrgTo3Dmain( i, numVertex, in, out, ctlVec[ i ] );
		if( result == true )
			return( true );
	}

	return( false );
}


bool  CbxSurface::
CtlVecOrgTo3Dmain( const short no, const short numVertex, const CbxCore::CtlVecForDefaultInVertex in[],
	const CbxCore::CtlVecForDefaultOutVertex out[], double ctlVec[ 2 ][ 3 ] )
{
	if( CbxVector::Handle1ForCtlVecForDefault( no, numVertex, in, out, ctlVec[ 0 ] ) == true )
		return( true );
	if( CbxVector::Handle2ForCtlVecForDefault( no, numVertex, in, out, ctlVec[ 1 ] ) == true )
		return( true );

	return( false );
}


bool  CbxSurface::
GetDefaultCtlVec( const short numVertex, const double posVec[][ 2 ][ 3 ], const double srcCtlVec[][ 2 ][ 3 ],
	double dstCtlVec[][ 2 ][ 3 ] )
{
	bool   result;
	short  i, j;

	for( i=0; i<numVertex; i++ ) {
		for( j=0; j<2; j++ ) {
			result = CbxVector::GetDefaultCtlVec( posVec[ i ][ j ], srcCtlVec[ i ][ j ], dstCtlVec[ i ][ j ] );
			if( result == true )
				return( true );
		}
	}

	return( false );
}


bool  CbxSurface::
CtlVec3DtoOrg( const short numVertex, const CbxCore::CtlVecForDefaultInVertex in[], const double ctlVec[][ 2 ][ 3 ],
	CbxCore::CtlVecForDefaultOutVertex out[] )
{
	bool   result;
	short  i;

	for( i=0; i<numVertex; i++ ) {
		result = CtlVec3DtoOrgMain( i, numVertex, in, ctlVec[ i ], out );
		if( result == true )
			return( true );
	}

	return( false );
}


bool  CbxSurface::
CtlVec3DtoOrgMain( const short no, const short numVertex, const CbxCore::CtlVecForDefaultInVertex in[],
	const double ctlVec[ 2 ][ 3 ], CbxCore::CtlVecForDefaultOutVertex out[] )
{
	if( CbxVector::SetHandle1( no, numVertex, in, ctlVec[ 0 ], out ) == true )
		return( true );
	if( CbxVector::SetHandle2( no, numVertex, in, ctlVec[ 1 ], out ) == true )
		return( true );

	return( false );
}
