/* bxTri.cpp */

#include "bxTri.h"
#include "../Surface/bxSurface.h"
#include "../Vector/bxVector.h"
#include "../../Math/bxMath.h"


bool    CbxTri::enableForSaveTime[ 3 ]				= { false, false, false };
double  CbxTri::p0ForSaveTime[ 3 ][ 3 ];
double  CbxTri::p1ForSaveTime[ 3 ][ 3 ];

const double  CbxTri::ratioForCorrectDefaultCtlVec	= 2.0/3.0;



bool  CbxTri::
ConvSurface( const CbxCore::SurfaceForPickVertex vertex[ 3 ], CbxCore::TriPatch tri[ 3 ] )
{
	bool  result;
	CbxCore::SurfaceForPickVertex  vertex2[ 3 ];

	vertex2[ 0 ] = vertex[ 1 ];		vertex2[ 1 ] = vertex[ 2 ];		vertex2[ 2 ] = vertex[ 0 ];
	result = ConvSurfaceMain( vertex2, &tri[ triCalcTypeR ] );
	if( result == true )
		return( true );

	vertex2[ 0 ] = vertex[ 2 ];		vertex2[ 1 ] = vertex[ 0 ];		vertex2[ 2 ] = vertex[ 1 ];
	result = ConvSurfaceMain( vertex2, &tri[ triCalcTypeS ] );
	if( result == true )
		return( true );

	vertex2[ 0 ] = vertex[ 0 ];		vertex2[ 1 ] = vertex[ 1 ];		vertex2[ 2 ] = vertex[ 2 ];
	result = ConvSurfaceMain( vertex2, &tri[ triCalcTypeT ] );
	if( result == true )
		return( true );

	return( false );
}


bool  CbxTri::
ConvSurfaceMain( const CbxCore::SurfaceForPickVertex vertex[ 3 ], CbxCore::TriPatch* tri )
{
	bool    result;
	double  f[ 3 ][ 4 ][ 3 ], a[ 3 ][ 5 ][ 3 ];

	result = CbxSurface::GetElement( 3, vertex, f, a );
	if( result == true )
		return( true );

	ConvSurfaceSub( vertex, ( const CbxSurface::Double4x3* )f, ( const CbxSurface::Double5x3* )a, tri );

	return( false );
}



void  CbxTri::
ConvSurfaceSub( const CbxCore::SurfaceForPickVertex vertex[ 3 ], const double f[ 3 ][ 4 ][ 3 ],
	const double a[ 3 ][ 5 ][ 3 ], CbxCore::TriPatch* tri )
{
	CbxCore::Patch5  triForm[ 3 ];
	CbxCore::Patch6  triJoin[ 3 ];

	ConvSurfaceTriForm( vertex, f, triForm );
	ConvFormTriToRect( triForm, tri->form );

	ConvSurfaceTriJoin( vertex, a, triJoin );
	ConvJoinTriToRect( triJoin, tri->join );
}



void  CbxTri::
ConvSurfaceTriForm( const CbxCore::SurfaceForPickVertex vertex[ 3 ], const double f[ 3 ][ 4 ][ 3 ],
	CbxCore::Patch5 triForm[ 3 ] )
{
	double  edge[ 3 ][ 4 ][ 3 ], in[ 3 ][ 2 ][ 3 ];

	ConvSurfaceTriFormEdge( vertex, edge );
	ConvSurfaceTriFormIn( vertex, f, in );

	ConvSurfaceTriFormMain( ( const CbxSurface::Double4x3* )edge, ( const CbxSurface::Double2x3* )in, triForm );
}


void  CbxTri::
ConvSurfaceTriFormEdge( const CbxCore::SurfaceForPickVertex vertex[ 3 ], double edge[ 3 ][ 4 ][ 3 ] )
{
	double  p[ 4 ][ 3 ], p2[ 5 ][ 3 ];
	short   i, j, k;

	for( i=0; i<3; i++ ) {
		k = ( i + 1 ) % 3;
		CbxVector::Pos( i, 3, vertex, p );
		CbxSurface::OneDimUpPos( 3, p, p2 );
		for( j=0; j<4; j++ )
			CbxMath::Copy3( p2[ j ], edge[ k ][ j ] );
	}
}


void  CbxTri::
ConvSurfaceTriFormIn( const CbxCore::SurfaceForPickVertex vertex[ 3 ], const double f[ 3 ][ 4 ][ 3 ],
	double in[ 3 ][ 2 ][ 3 ] )
{
	double  f2[ 3 ][ 4 ][ 3 ], p[ 2 ][ 3 ];
	short   i, j, k;

	for( i=0; i<3; i++ ) {
		for( j=0; j<4; j++ )
			CbxMath::Scale3( f[ i ][ j ], 3.0/4.0, f2[ i ][ j ] );
	}

	for( i=0; i<3; i++ ) {
		k = ( i + 1 ) % 3;
		CbxVector::Pos1( i, 3, vertex, p[ 0 ] );
		CbxVector::Pos2( i, 3, vertex, p[ 1 ] );
		for( j=0; j<2; j++ )
			CbxMath::Add3( p[ j ], f2[ i ][ j+1 ], in[ k ][ j ] );
	}
}


void  CbxTri::
ConvSurfaceTriFormMain( const double edge[ 3 ][ 4 ][ 3 ], const double in[ 3 ][ 2 ][ 3 ],
	CbxCore::Patch5 triForm[ 3 ] )
{
	ConvSurfaceTriFormMainEdge( edge, triForm );
	ConvSurfaceTriFormMainInBase( in, triForm );
}


void  CbxTri::
ConvSurfaceTriFormMainEdge( const double edge[ 3 ][ 4 ][ 3 ], CbxCore::Patch5 triForm[ 3 ] )
{
	short  i, j;

	for( i=0; i<3; i++ ) {
		for( j=0; j<4; j++ )
			CbxMath::Copy3( edge[ 0 ][ j ], triForm[ i ].p[ 0 ][ 4-j ].pos );
		for( j=0; j<4; j++ )
			CbxMath::Copy3( edge[ 1 ][ j ], triForm[ i ].p[ j ][ 0 ].pos );
		for( j=0; j<4; j++ )
			CbxMath::Copy3( edge[ 2 ][ j ], triForm[ i ].p[ 4-j ][ j ].pos );
	}
}


void  CbxTri::
ConvSurfaceTriFormMainInBase( const double in[ 3 ][ 2 ][ 3 ], CbxCore::Patch5 triForm[ 3 ] )
{
	CbxMath::Copy3( in[ 0 ][ 0 ], triForm[ 0 ].p[ 1 ][ 2 ].pos );
	CbxMath::Copy3( in[ 0 ][ 1 ], triForm[ 0 ].p[ 1 ][ 1 ].pos );

	CbxMath::Copy3( in[ 1 ][ 0 ], triForm[ 1 ].p[ 1 ][ 1 ].pos );
	CbxMath::Copy3( in[ 1 ][ 1 ], triForm[ 1 ].p[ 2 ][ 1 ].pos );

	CbxMath::Copy3( in[ 2 ][ 0 ], triForm[ 2 ].p[ 2 ][ 1 ].pos );
	CbxMath::Copy3( in[ 2 ][ 1 ], triForm[ 2 ].p[ 1 ][ 2 ].pos );

	CbxMath::Ratio3( triForm[ 0 ].p[ 1 ][ 2 ].pos, triForm[ 2 ].p[ 1 ][ 2 ].pos, 0.5, triForm[ 1 ].p[ 1 ][ 2 ].pos );
	CbxMath::Ratio3( triForm[ 0 ].p[ 1 ][ 1 ].pos, triForm[ 1 ].p[ 1 ][ 1 ].pos, 0.5, triForm[ 2 ].p[ 1 ][ 1 ].pos );
	CbxMath::Ratio3( triForm[ 1 ].p[ 2 ][ 1 ].pos, triForm[ 2 ].p[ 2 ][ 1 ].pos, 0.5, triForm[ 0 ].p[ 2 ][ 1 ].pos );
}



void  CbxTri::
ConvFormTriToRect( const CbxCore::Patch5 triForm[ 3 ], CbxCore::Patch5 rectForm[ 3 ] )
{
	short  i, u, v;
	CbxCore::PatchPnt  *p0[ 5 ], *p1[ 5 ];
	CbxCore::PatchPnt   buf0[ 5 ][ 5 ], buf1[ 5 ][ 5 ];

	for( i=0; i<5; i++ ) {
		p0[ i ] = buf0[ i ];
		p1[ i ] = buf1[ i ];
	}

	for( i=0; i<3; i++ ) {
		for( v=0; v<5; v++ ) {
			for( u=0; u<5; u++ )
				p0[ u ][ v ] = triForm[ i ].p[ u ][ v ];
		}

		DimUpAndCopy( 4, 0, ( const CbxCore::PatchPnt** )p0, p1 );
		DimUpAndCopy( 4, 1, ( const CbxCore::PatchPnt** )p1, p0 );
		DimUpAndCopy( 4, 2, ( const CbxCore::PatchPnt** )p0, p1 );
		DimUpAndCopy( 4, 3, ( const CbxCore::PatchPnt** )p1, p0 );

		for( v=0; v<5; v++ ) {
			for( u=0; u<5; u++ )
				rectForm[ i ].p[ u ][ v ] = p0[ u ][ v ];
		}
	}
}


void  CbxTri::
DimUpAndCopy( const short numDim, const short srcDim, const CbxCore::PatchPnt** src, CbxCore::PatchPnt** dst )
{
	short   u, v, dstDim;
	double  p0[ 5 ][ 3 ], p1[ 6 ][ 3 ];

	dstDim = srcDim + 1;

	for( v=0; v<( numDim-srcDim ); v++ ) {
		for( u=0; u<( numDim+1 ); u++ )
			CbxMath::Copy3( src[ u ][ v ].pos, dst[ u ][ v ].pos );
	}

	for( v=0; v<dstDim; v++ ) {
		for( u=0; u<dstDim; u++ )
			CbxMath::Copy3( src[ u ][ numDim-v ].pos, p0[ u ] );

		CbxSurface::OneDimUpPos( srcDim, p0, p1 );

		for( u=0; u<( dstDim+1 ); u++ )
			CbxMath::Copy3( p1[ u ], dst[ u ][ numDim-v ].pos );
	}
}


void  CbxTri::
ConvSurfaceTriJoin( const CbxCore::SurfaceForPickVertex vertex[ 3 ], const double a[ 3 ][ 5 ][ 3 ],
	CbxCore::Patch6 triJoin[ 3 ] )
{
	double  edge[ 3 ][ 5 ][ 3 ], in[ 3 ][ 3 ][ 3 ];

	ConvSurfaceTriJoinEdge( vertex, edge );
	ConvSurfaceTriJoinIn( vertex, a, in );

	ConvSurfaceTriJoinMain( ( const CbxSurface::Double5x3* )edge, ( const CbxSurface::Double3x3* )in, triJoin );
}


void  CbxTri::
ConvSurfaceTriJoinEdge( const CbxCore::SurfaceForPickVertex vertex[ 3 ], double edge[ 3 ][ 5 ][ 3 ] )
{
	double  p[ 4 ][ 3 ], p2[ 5 ][ 3 ], p3[ 6 ][ 3 ];
	short   i, j, k;

	for( i=0; i<3; i++ ) {
		k = ( i + 1 ) % 3;
		CbxVector::Pos( i, 3, vertex, p );
		CbxSurface::OneDimUpPos( 3, p,  p2 );
		CbxSurface::OneDimUpPos( 4, p2, p3 );
		for( j=0; j<5; j++ )
			CbxMath::Copy3( p3[ j ], edge[ k ][ j ] );
	}
}


void  CbxTri::
ConvSurfaceTriJoinIn( const CbxCore::SurfaceForPickVertex vertex[ 3 ], const double a[ 3 ][ 5 ][ 3 ],
	double in[ 3 ][ 3 ][ 3 ] )
{
	double  a2[ 3 ][ 5 ][ 3 ], p[ 4 ][ 3 ], p2[ 5 ][ 3 ];
	short   i, j, k;

	for( i=0; i<3; i++ ) {
		for( j=0; j<5; j++ )
			CbxMath::Scale3( a[ i ][ j ], 3.0/5.0, a2[ i ][ j ] );
	}

	for( i=0; i<3; i++ ) {
		k = ( i + 1 ) % 3;
		CbxVector::Pos( i, 3, vertex, p );
		CbxSurface::OneDimUpPos( 3, p, p2 );
		for( j=0; j<3; j++ )
			CbxMath::Add3( p2[ j+1 ], a2[ i ][ j+1 ], in[ k ][ j ] );
	}
}


void  CbxTri::
ConvSurfaceTriJoinMain( const double edge[ 3 ][ 5 ][ 3 ], const double in[ 3 ][ 3 ][ 3 ],
	CbxCore::Patch6 triJoin[ 3 ] )
{
	ConvSurfaceTriJoinMainEdge( edge, triJoin );
	ConvSurfaceTriJoinMainInBase( in, triJoin );

	ConvSurfaceTriJoinMainInCenterU0( triJoin[ 0 ].p );
	ConvSurfaceTriJoinMainInCenterV0( triJoin[ 1 ].p );
	ConvSurfaceTriJoinMainInCenterU1( triJoin[ 2 ].p );
}


void  CbxTri::
ConvSurfaceTriJoinMainEdge( const double edge[ 3 ][ 5 ][ 3 ], CbxCore::Patch6 triJoin[ 3 ] )
{
	short  i, j;

	for( i=0; i<3; i++ ) {
		for( j=0; j<5; j++ )
			CbxMath::Copy3( edge[ 0 ][ j ], triJoin[ i ].p[ 0 ][ 5-j ].pos );
		for( j=0; j<5; j++ )
			CbxMath::Copy3( edge[ 1 ][ j ], triJoin[ i ].p[ j ][ 0 ].pos );
		for( j=0; j<5; j++ )
			CbxMath::Copy3( edge[ 2 ][ j ], triJoin[ i ].p[ 5-j ][ j ].pos );
	}
}


void  CbxTri::
ConvSurfaceTriJoinMainInBase( const double in[ 3 ][ 3 ][ 3 ], CbxCore::Patch6 triJoin[ 3 ] )
{
	CbxMath::Copy3( in[ 0 ][ 0 ], triJoin[ 0 ].p[ 1 ][ 3 ].pos );
	CbxMath::Copy3( in[ 0 ][ 1 ], triJoin[ 0 ].p[ 1 ][ 2 ].pos );
	CbxMath::Copy3( in[ 0 ][ 2 ], triJoin[ 0 ].p[ 1 ][ 1 ].pos );

	CbxMath::Copy3( in[ 1 ][ 0 ], triJoin[ 1 ].p[ 1 ][ 1 ].pos );
	CbxMath::Copy3( in[ 1 ][ 1 ], triJoin[ 1 ].p[ 2 ][ 1 ].pos );
	CbxMath::Copy3( in[ 1 ][ 2 ], triJoin[ 1 ].p[ 3 ][ 1 ].pos );

	CbxMath::Copy3( in[ 2 ][ 0 ], triJoin[ 2 ].p[ 3 ][ 1 ].pos );
	CbxMath::Copy3( in[ 2 ][ 1 ], triJoin[ 2 ].p[ 2 ][ 2 ].pos );
	CbxMath::Copy3( in[ 2 ][ 2 ], triJoin[ 2 ].p[ 1 ][ 3 ].pos );

	CbxMath::Ratio3( triJoin[ 0 ].p[ 1 ][ 3 ].pos, triJoin[ 2 ].p[ 1 ][ 3 ].pos, 0.5, triJoin[ 1 ].p[ 1 ][ 3 ].pos );
	CbxMath::Ratio3( triJoin[ 0 ].p[ 1 ][ 1 ].pos, triJoin[ 1 ].p[ 1 ][ 1 ].pos, 0.5, triJoin[ 2 ].p[ 1 ][ 1 ].pos );
	CbxMath::Ratio3( triJoin[ 1 ].p[ 3 ][ 1 ].pos, triJoin[ 2 ].p[ 3 ][ 1 ].pos, 0.5, triJoin[ 0 ].p[ 3 ][ 1 ].pos );
}


void  CbxTri::
ConvSurfaceTriJoinMainInCenterU0( CbxCore::PatchPnt triJoin[ 6 ][ 6 ] )
{
	double  p0[ 4 ][ 3 ], p1[ 5 ][ 3 ];

	CbxMath::Copy3( triJoin[ 0 ][ 4 ].pos, p0[ 0 ] );
	CbxMath::Copy3( triJoin[ 4 ][ 0 ].pos, p0[ 3 ] );
	CbxMath::Ratio3( p0[ 0 ], triJoin[ 1 ][ 3 ].pos, 4.0/3.0, p0[ 1 ] );
	CbxMath::Ratio3( p0[ 3 ], triJoin[ 3 ][ 1 ].pos, 4.0/3.0, p0[ 2 ] );
	CbxSurface::OneDimUpPos( 3, p0, p1 );
	CbxMath::Copy3( p1[ 2 ], triJoin[ 2 ][ 2 ].pos );

	CbxMath::Copy3( triJoin[ 0 ][ 1 ].pos, p0[ 0 ] );
	CbxMath::Copy3( triJoin[ 4 ][ 1 ].pos, p0[ 3 ] );
	CbxMath::Ratio3( p0[ 0 ], triJoin[ 1 ][ 1 ].pos, 4.0/3.0, p0[ 1 ] );
	CbxMath::Ratio3( p0[ 3 ], triJoin[ 3 ][ 1 ].pos, 4.0/3.0, p0[ 2 ] );
	CbxSurface::OneDimUpPos( 3, p0, p1 );
	CbxMath::Copy3( p1[ 2 ], triJoin[ 2 ][ 1 ].pos );
}


void  CbxTri::
ConvSurfaceTriJoinMainInCenterV0( CbxCore::PatchPnt triJoin[ 6 ][ 6 ] )
{
	double  p0[ 4 ][ 3 ], p1[ 5 ][ 3 ];

	CbxMath::Copy3( triJoin[ 1 ][ 0 ].pos, p0[ 0 ] );
	CbxMath::Copy3( triJoin[ 1 ][ 4 ].pos, p0[ 3 ] );
	CbxMath::Ratio3( p0[ 0 ], triJoin[ 1 ][ 1 ].pos, 4.0/3.0, p0[ 1 ] );
	CbxMath::Ratio3( p0[ 3 ], triJoin[ 1 ][ 3 ].pos, 4.0/3.0, p0[ 2 ] );
	CbxSurface::OneDimUpPos( 3, p0, p1 );
	CbxMath::Copy3( p1[ 2 ], triJoin[ 1 ][ 2 ].pos );

	CbxMath::Copy3( triJoin[ 4 ][ 0 ].pos, p0[ 0 ] );
	CbxMath::Copy3( triJoin[ 0 ][ 4 ].pos, p0[ 3 ] );
	CbxMath::Ratio3( p0[ 0 ], triJoin[ 3 ][ 1 ].pos, 4.0/3.0, p0[ 1 ] );
	CbxMath::Ratio3( p0[ 3 ], triJoin[ 1 ][ 3 ].pos, 4.0/3.0, p0[ 2 ] );
	CbxSurface::OneDimUpPos( 3, p0, p1 );
	CbxMath::Copy3( p1[ 2 ], triJoin[ 2 ][ 2 ].pos );
}


void  CbxTri::
ConvSurfaceTriJoinMainInCenterU1( CbxCore::PatchPnt triJoin[ 6 ][ 6 ] )
{
	double  p0[ 4 ][ 3 ], p1[ 5 ][ 3 ];

	CbxMath::Copy3( triJoin[ 4 ][ 1 ].pos, p0[ 0 ] );
	CbxMath::Copy3( triJoin[ 0 ][ 1 ].pos, p0[ 3 ] );
	CbxMath::Ratio3( p0[ 0 ], triJoin[ 3 ][ 1 ].pos, 4.0/3.0, p0[ 1 ] );
	CbxMath::Ratio3( p0[ 3 ], triJoin[ 1 ][ 1 ].pos, 4.0/3.0, p0[ 2 ] );
	CbxSurface::OneDimUpPos( 3, p0, p1 );
	CbxMath::Copy3( p1[ 2 ], triJoin[ 2 ][ 1 ].pos );

	CbxMath::Copy3( triJoin[ 1 ][ 4 ].pos, p0[ 0 ] );
	CbxMath::Copy3( triJoin[ 1 ][ 0 ].pos, p0[ 3 ] );
	CbxMath::Ratio3( p0[ 0 ], triJoin[ 1 ][ 3 ].pos, 4.0/3.0, p0[ 1 ] );
	CbxMath::Ratio3( p0[ 3 ], triJoin[ 1 ][ 1 ].pos, 4.0/3.0, p0[ 2 ] );
	CbxSurface::OneDimUpPos( 3, p0, p1 );
	CbxMath::Copy3( p1[ 2 ], triJoin[ 1 ][ 2 ].pos );
}



void  CbxTri::
ConvJoinTriToRect( const CbxCore::Patch6 triJoin[ 3 ], CbxCore::Patch6 rectJoin[ 3 ] )
{
	short  i, u, v;
	CbxCore::PatchPnt  *p0[ 6 ], *p1[ 6 ];
	CbxCore::PatchPnt   buf0[ 6 ][ 6 ], buf1[ 6 ][ 6 ];

	for( i=0; i<6; i++ ) {
		p0[ i ] = buf0[ i ];
		p1[ i ] = buf1[ i ];
	}

	for( i=0; i<3; i++ ) {
		for( v=0; v<6; v++ ) {
			for( u=0; u<6; u++ )
				p0[ u ][ v ] = triJoin[ i ].p[ u ][ v ];
		}

		DimUpAndCopy( 5, 0, ( const CbxCore::PatchPnt** )p0, p1 );
		DimUpAndCopy( 5, 1, ( const CbxCore::PatchPnt** )p1, p0 );
		DimUpAndCopy( 5, 2, ( const CbxCore::PatchPnt** )p0, p1 );
		DimUpAndCopy( 5, 3, ( const CbxCore::PatchPnt** )p1, p0 );
		DimUpAndCopy( 5, 4, ( const CbxCore::PatchPnt** )p0, p1 );

		for( v=0; v<6; v++ ) {
			for( u=0; u<6; u++ )
				rectJoin[ i ].p[ u ][ v ] = p1[ u ][ v ];
		}
	}
}


void  CbxTri::
PosNormalSurface( const CbxCore::TriPatch tri[ 3 ], const double uv[ 2 ], double pos[ 3 ], double normalVec[ 3 ] )
{
	SetDisableForSaveTime();

	PosSurfaceCom( tri, uv, pos );
	NormalSurfaceCom( tri, uv, normalVec );
}


void  CbxTri::
PosSurface( const CbxCore::TriPatch tri[ 3 ], const double uv[ 2 ], double pos[ 3 ], short* triCalcType/*= 0*/ )
{
	SetDisableForSaveTime();
	PosSurfaceCom( tri, uv, pos, triCalcType );
}


void  CbxTri::
PosSurfaceCom( const CbxCore::TriPatch tri[ 3 ], const double uv[ 2 ], double pos[ 3 ], short* triCalcType/*= 0*/ )
{
	short   type;
	double  uv2[ 2 ];

	type = GetTriCalcType( uv );
	if( triCalcType != 0 )
		*triCalcType = type;

	if( uv[ 1 ] < CbxMath::nearZero ) {
		if( uv[ 0 ] < CbxMath::nearZero ) {
			CbxMath::Copy3( tri[ triCalcTypeT ].form[ 0 ].p[ 0 ][ 0 ].pos, pos );
			return;
		}
		else if( uv[ 0 ] > ( 1.0 - CbxMath::nearZero ) ) {
			CbxMath::Copy3( tri[ triCalcTypeR ].form[ 0 ].p[ 0 ][ 0 ].pos, pos );
			return;
		}
	}
	else if( uv[ 1 ] > ( 1.0 - CbxMath::nearZero ) ) {
		CbxMath::Copy3( tri[ triCalcTypeS ].form[ 0 ].p[ 0 ][ 0 ].pos, pos );
		return;
	}

	ConvUV( type, uv, uv2 );
	PosSurfaceMain( &tri[ type ], uv2, pos );
}


short  CbxTri::
GetTriCalcType( const double uv[ 2 ] )
{
	double  r, s, t;

	r = uv[ 0 ] * ( 1.0 - uv[ 1 ] );
	s = uv[ 1 ];
	t = ( 1.0 - uv[ 0 ] ) * ( 1.0 - uv[ 1 ] );

	if( r >= s ) {
		if( r >= t )
			return( triCalcTypeR );
		else
			return( triCalcTypeT );
	}
	else {
		if( s >= t )
			return( triCalcTypeS );
		else
			return( triCalcTypeT );
	}
}


void  CbxTri::
PosSurfaceMain( const CbxCore::TriPatch* tri, const double uv[ 2 ], double pos[ 3 ] )
{
	double  param[ 2 ], yr[ 3 ], ys[ 3 ], yt[ 3 ], rParam, sParam, tParam;
	short   i;

	param[ 0 ] = 1.0 - uv[ 0 ] + uv[ 0 ] * uv[ 1 ];
	param[ 1 ] = uv[ 0 ] * ( 1.0 - uv[ 1 ] );
	PosSurfaceSub( triCalcTypeR, tri->form[ 0 ].p, tri->join[ 0 ].p, param, uv, yr );

	param[ 0 ] = 1.0 - uv[ 1 ];
	param[ 1 ] = uv[ 1 ];
	PosSurfaceSub( triCalcTypeS, tri->form[ 1 ].p, tri->join[ 1 ].p, param, uv, ys );

	param[ 0 ] = uv[ 0 ] + uv[ 1 ] - uv[ 0 ] * uv[ 1 ];
	param[ 1 ] = ( 1.0 - uv[ 0 ] ) * ( 1.0 - uv[ 1 ] );
	PosSurfaceSub( triCalcTypeT, tri->form[ 2 ].p, tri->join[ 2 ].p, param, uv, yt );

	rParam = ( 1.0 - uv[ 0 ] ) * uv[ 1 ];
	sParam = uv[ 0 ] * ( 1.0 - uv[ 0 ] ) * ( 1.0 - uv[ 1 ] );
	tParam = uv[ 0 ] * uv[ 1 ];
	for( i=0; i<3; i++ )
		pos[ i ] = 1.0 / ( rParam + sParam + tParam ) * ( rParam * yr[ i ] + sParam * ys[ i ] + tParam * yt[ i ] );
}


void  CbxTri::
ConvUV( const short triCalcType, const double srcUV[ 2 ], double dstUV[ 2 ] )
{
	double  srcRST[ 3 ], dstRST[ 2 ];

	srcRST[ 0 ] = srcUV[ 0 ] * ( 1.0 - srcUV[ 1 ] );
	srcRST[ 1 ] = srcUV[ 1 ];
	srcRST[ 2 ] = ( 1.0 - srcUV[ 0 ] ) * ( 1.0 - srcUV[ 1 ] );

	switch( triCalcType ) {
		case triCalcTypeR:
			dstRST[ 0 ] = srcRST[ 1 ];		dstRST[ 1 ] = srcRST[ 2 ];
			break;
		case triCalcTypeS:
			dstRST[ 0 ] = srcRST[ 2 ];		dstRST[ 1 ] = srcRST[ 0 ];
			break;
		case triCalcTypeT:
			dstRST[ 0 ] = srcRST[ 0 ];		dstRST[ 1 ] = srcRST[ 1 ];
			break;
	}

	dstUV[ 0 ] = dstRST[ 0 ] / ( 1.0 - dstRST[ 1 ] );
	dstUV[ 1 ] = dstRST[ 1 ];
}


void  CbxTri::
PosSurfaceSub( const short triCalcType, const CbxCore::PatchPnt form[ 5 ][ 5 ], const CbxCore::PatchPnt join[ 6 ][ 6 ],
	const double param[ 2 ], const double uv[ 2 ], double pos[ 3 ] )
{
	double  patch0[ 6 ][ 6 ][ 3 ], patch1[ 5 ][ 5 ][ 3 ], p0[ 3 ], p1[ 3 ];
	short   u, v, i;

	if( ChkEnableForSaveTime( triCalcType, p0, p1 ) == false ) {
		for( v=0; v<6; v++ ) {
			for( u=0; u<6; u++ )
				CbxMath::Copy3( join[ u ][ v ].pos, patch0[ u ][ v ] );
		}
		CbxSurface::PosSurface5( ( const CbxSurface::Double6x3* )patch0, uv, p0 );

		for( v=0; v<5; v++ ) {
			for( u=0; u<5; u++ )
				CbxMath::Copy3( form[ u ][ v ].pos, patch1[ u ][ v ] );
		}
		CbxSurface::PosSurface4( ( const CbxSurface::Double5x3* )patch1, uv, p1 );

		SetEnableForSaveTime( triCalcType, p0, p1 );
	}

	for( i=0; i<3; i++ )
		pos[ i ] = param[ 0 ] * p0[ i ] + param[ 1 ] * p1[ i ];
}


void  CbxTri::
NormalSurface( const CbxCore::TriPatch tri[ 3 ], const double uv[ 2 ], double normalVec[ 3 ] )
{
	SetDisableForSaveTime();
	NormalSurfaceCom( tri, uv, normalVec );
}


void  CbxTri::
NormalSurfaceCom( const CbxCore::TriPatch tri[ 3 ], const double uv[ 2 ], double normalVec[ 3 ] )
{
	double  du[ 3 ], dv[ 3 ];

	DiffUsurface( tri, uv, du );
	DiffVsurface( tri, uv, dv );
	CbxMath::Cross3( du, dv, normalVec );
	CbxMath::Normalize3( normalVec, CbxMath::zero3, normalVec );
}


void  CbxTri::
DiffUsurface( const CbxCore::TriPatch tri[ 3 ], const double uv[ 2 ], double vec[ 3 ], short* triCalcType/*= 0*/ )
{
	short   type;
	double  uv2[ 2 ];

	type = GetTriCalcType( uv );
	if( triCalcType != 0 )
		*triCalcType = type;

	if( uv[ 1 ] < CbxMath::nearZero ) {
		if( uv[ 0 ] < CbxMath::nearZero ) {
			DiffUsurfaceVertex( &tri[ triCalcTypeT ], vec );
			return;
		}
		else if( uv[ 0 ] > ( 1.0 - CbxMath::nearZero ) ) {
			DiffUsurfaceVertex( &tri[ triCalcTypeR ], vec );
			return;
		}
	}
	else if( uv[ 1 ] > ( 1.0 - CbxMath::nearZero ) ) {
		DiffUsurfaceVertex( &tri[ triCalcTypeS ], vec );
		return;
	}

	ConvUV( type, uv, uv2 );
	DiffUsurfaceMain( &tri[ type ], uv2, vec );
}


void  CbxTri::
DiffUsurfaceVertex( const CbxCore::TriPatch* tri, double vec[ 3 ] )
{
	CbxMath::Subtract3( tri->form[ 0 ].p[ 1 ][ 0 ].pos, tri->form[ 0 ].p[ 0 ][ 0 ].pos, vec );
	CbxMath::Scale3( vec, 4.0, vec );
}


void  CbxTri::
DiffUsurfaceMain( const CbxCore::TriPatch* tri, const double uv[ 2 ], double vec[ 3 ] )
{
	double  param[ 3 ], y[ 3 ], yr[ 3 ], ys[ 3 ], yt[ 3 ], dyr[ 3 ], dys[ 3 ], dyt[ 3 ];
	double  rParam0, sParam0, tParam0, rParam1, sParam1, tParam1, element0, element1, element2;
	short   i;

	param[ 0 ] = 1.0 - uv[ 0 ] + uv[ 0 ] * uv[ 1 ];
	param[ 1 ] = uv[ 0 ] * ( 1.0 - uv[ 1 ] );
	param[ 2 ] = 1.0 - uv[ 1 ];
	PosSurfaceSub( triCalcTypeR, tri->form[ 0 ].p, tri->join[ 0 ].p, param, uv, yr );
	DiffUsurfaceSub( triCalcTypeR, tri->form[ 0 ].p, tri->join[ 0 ].p, param, uv, dyr );

	param[ 0 ] = 1.0 - uv[ 1 ];
	param[ 1 ] = uv[ 1 ];
	param[ 2 ] = 0.0;
	PosSurfaceSub( triCalcTypeS, tri->form[ 1 ].p, tri->join[ 1 ].p, param, uv, ys );
	DiffUsurfaceSub( triCalcTypeS, tri->form[ 1 ].p, tri->join[ 1 ].p, param, uv, dys );

	param[ 0 ] = uv[ 0 ] + uv[ 1 ] - uv[ 0 ] * uv[ 1 ];
	param[ 1 ] = ( 1.0 - uv[ 0 ] ) * ( 1.0 - uv[ 1 ] );
	param[ 2 ] = -( 1.0 - uv[ 1 ] );
	PosSurfaceSub( triCalcTypeT, tri->form[ 2 ].p, tri->join[ 2 ].p, param, uv, yt );
	DiffUsurfaceSub( triCalcTypeT, tri->form[ 2 ].p, tri->join[ 2 ].p, param, uv, dyt );

	rParam0 = ( 1.0 - uv[ 0 ] ) * uv[ 1 ];
	sParam0 = uv[ 0 ] * ( 1.0 - uv[ 0 ] ) * ( 1.0 - uv[ 1 ] );
	tParam0 = uv[ 0 ] * uv[ 1 ];
	rParam1 = -uv[ 1 ];
	sParam1 = ( 1.0 - 2.0 * uv[ 0 ] ) * ( 1.0 - uv[ 1 ] );
	tParam1 =  uv[ 1 ];
	for( i=0; i<3; i++ ) {
		y[ i ] = 1.0 / ( rParam0 + sParam0 + tParam0 ) * ( rParam0 * yr[ i ] + sParam0 * ys[ i ] + tParam0 * yt[ i ] );
		element0 = ( 2.0 * uv[ 0 ] - 1.0 ) * ( 1.0 - uv[ 1 ] ) * y[ i ];
		element1 = rParam1 *  yr[ i ] + sParam1 *  ys[ i ] + tParam1 *  yt[ i ];
		element2 = rParam0 * dyr[ i ] + sParam0 * dys[ i ] + tParam0 * dyt[ i ];

		vec[ i ] = 1.0 / ( rParam0 + sParam0 + tParam0 ) * ( element0 + element1 + element2 );
	}
}


void  CbxTri::
DiffUsurfaceSub( const short triCalcType, const CbxCore::PatchPnt form[ 5 ][ 5 ], const CbxCore::PatchPnt join[ 6 ][ 6 ],
	const double param[ 3 ], const double uv[ 2 ], double vec[ 3 ] )
{
	double  patch0[ 6 ][ 6 ][ 3 ], patch1[ 5 ][ 5 ][ 3 ], p0[ 3 ], p1[ 3 ], v0[ 3 ], v1[ 3 ];
	short   u, v, i;
	bool    chkEnableForSaveTime;

	chkEnableForSaveTime = ChkEnableForSaveTime( triCalcType, p0, p1 );

	for( v=0; v<6; v++ ) {
		for( u=0; u<6; u++ )
			CbxMath::Copy3( join[ u ][ v ].pos, patch0[ u ][ v ] );
	}
	if( chkEnableForSaveTime == false )
		CbxSurface::PosSurface5( ( const CbxSurface::Double6x3* )patch0, uv, p0 );
	CbxSurface::DiffUsurface5( ( const CbxSurface::Double6x3* )patch0, uv, v0 );

	for( v=0; v<5; v++ ) {
		for( u=0; u<5; u++ )
			CbxMath::Copy3( form[ u ][ v ].pos, patch1[ u ][ v ] );
	}
	if( chkEnableForSaveTime == false )
		CbxSurface::PosSurface4( ( const CbxSurface::Double5x3* )patch1, uv, p1 );
	CbxSurface::DiffUsurface4( ( const CbxSurface::Double5x3* )patch1, uv, v1 );

	if( chkEnableForSaveTime == false )
		SetEnableForSaveTime( triCalcType, p0, p1 );

	for( i=0; i<3; i++ )
		vec[ i ] = param[ 2 ] * ( p1[ i ] - p0[ i ] ) + param[ 0 ] * v0[ i ] + param[ 1 ] * v1[ i ];
}


void  CbxTri::
DiffVsurface( const CbxCore::TriPatch tri[ 3 ], const double uv[ 2 ], double vec[ 3 ], short* triCalcType/*= 0*/ )
{
	short   type;
	double  uv2[ 2 ];

	type = GetTriCalcType( uv );
	if( triCalcType != 0 )
		*triCalcType = type;

	if( uv[ 1 ] < CbxMath::nearZero ) {
		if( uv[ 0 ] < CbxMath::nearZero ) {
			DiffVsurfaceVertex( &tri[ triCalcTypeT ], vec );
			return;
		}
		else if( uv[ 0 ] > ( 1.0 - CbxMath::nearZero ) ) {
			DiffVsurfaceVertex( &tri[ triCalcTypeR ], vec );
			return;
		}
	}
	else if( uv[ 1 ] > ( 1.0 - CbxMath::nearZero ) ) {
		DiffVsurfaceVertex( &tri[ triCalcTypeS ], vec );
		return;
	}

	ConvUV( type, uv, uv2 );
	DiffVsurfaceMain( &tri[ type ], uv2, vec );
}


void  CbxTri::
DiffVsurfaceVertex( const CbxCore::TriPatch* tri, double vec[ 3 ] )
{
	CbxMath::Subtract3( tri->form[ 0 ].p[ 0 ][ 1 ].pos, tri->form[ 0 ].p[ 0 ][ 0 ].pos, vec );
	CbxMath::Scale3( vec, 4.0, vec );
}


void  CbxTri::
DiffVsurfaceMain( const CbxCore::TriPatch* tri, const double uv[ 2 ], double vec[ 3 ] )
{
	double  param[ 3 ], y[ 3 ], yr[ 3 ], ys[ 3 ], yt[ 3 ], dyr[ 3 ], dys[ 3 ], dyt[ 3 ];
	double  rParam0, sParam0, tParam0, rParam1, sParam1, tParam1, element0, element1, element2;
	short   i;

	param[ 0 ] = 1.0 - uv[ 0 ] + uv[ 0 ] * uv[ 1 ];
	param[ 1 ] = uv[ 0 ] * ( 1.0 - uv[ 1 ] );
	param[ 2 ] = -uv[ 0 ];
	PosSurfaceSub( triCalcTypeR, tri->form[ 0 ].p, tri->join[ 0 ].p, param, uv, yr );
	DiffVsurfaceSub( triCalcTypeR, tri->form[ 0 ].p, tri->join[ 0 ].p, param, uv, dyr );

	param[ 0 ] = 1.0 - uv[ 1 ];
	param[ 1 ] = uv[ 1 ];
	param[ 2 ] = 1.0;
	PosSurfaceSub( triCalcTypeS, tri->form[ 1 ].p, tri->join[ 1 ].p, param, uv, ys );
	DiffVsurfaceSub( triCalcTypeS, tri->form[ 1 ].p, tri->join[ 1 ].p, param, uv, dys );

	param[ 0 ] = uv[ 0 ] + uv[ 1 ] - uv[ 0 ] * uv[ 1 ];
	param[ 1 ] = ( 1.0 - uv[ 0 ] ) * ( 1.0 - uv[ 1 ] );
	param[ 2 ] = -( 1.0 - uv[ 0 ] );
	PosSurfaceSub( triCalcTypeT, tri->form[ 2 ].p, tri->join[ 2 ].p, param, uv, yt );
	DiffVsurfaceSub( triCalcTypeT, tri->form[ 2 ].p, tri->join[ 2 ].p, param, uv, dyt );

	rParam0 = ( 1.0 - uv[ 0 ] ) * uv[ 1 ];
	sParam0 = uv[ 0 ] * ( 1.0 - uv[ 0 ] ) * ( 1.0 - uv[ 1 ] );
	tParam0 = uv[ 0 ] * uv[ 1 ];
	rParam1 = 1.0 - uv[ 0 ];
	sParam1 = uv[ 0 ] * ( uv[ 0 ] - 1.0 );
	tParam1 = uv[ 0 ];
	for( i=0; i<3; i++ ) {
		y[ i ] = 1.0 / ( rParam0 + sParam0 + tParam0 ) * ( rParam0 * yr[ i ] + sParam0 * ys[ i ] + tParam0 * yt[ i ] );
		element0 = -( uv[ 0 ] * uv[ 0 ] - uv[ 0 ] + 1.0 ) * y[ i ];
		element1 = rParam1 *  yr[ i ] + sParam1 *  ys[ i ] + tParam1 *  yt[ i ];
		element2 = rParam0 * dyr[ i ] + sParam0 * dys[ i ] + tParam0 * dyt[ i ];

		vec[ i ] = 1.0 / ( rParam0 + sParam0 + tParam0 ) * ( element0 + element1 + element2 );
	}
}


void  CbxTri::
DiffVsurfaceSub( const short triCalcType, const CbxCore::PatchPnt form[ 5 ][ 5 ], const CbxCore::PatchPnt join[ 6 ][ 6 ],
	const double param[ 3 ], const double uv[ 2 ], double vec[ 3 ] )
{
	double  patch0[ 6 ][ 6 ][ 3 ], patch1[ 5 ][ 5 ][ 3 ], p0[ 3 ], p1[ 3 ], v0[ 3 ], v1[ 3 ];
	short   u, v, i;
	bool    chkEnableForSaveTime;

	chkEnableForSaveTime = ChkEnableForSaveTime( triCalcType, p0, p1 );

	for( v=0; v<6; v++ ) {
		for( u=0; u<6; u++ )
			CbxMath::Copy3( join[ u ][ v ].pos, patch0[ u ][ v ] );
	}
	if( chkEnableForSaveTime == false )
		CbxSurface::PosSurface5( ( const CbxSurface::Double6x3* )patch0, uv, p0 );
	CbxSurface::DiffVsurface5( ( const CbxSurface::Double6x3* )patch0, uv, v0 );

	for( v=0; v<5; v++ ) {
		for( u=0; u<5; u++ )
			CbxMath::Copy3( form[ u ][ v ].pos, patch1[ u ][ v ] );
	}
	if( chkEnableForSaveTime == false )
		CbxSurface::PosSurface4( ( const CbxSurface::Double5x3* )patch1, uv, p1 );
	CbxSurface::DiffVsurface4( ( const CbxSurface::Double5x3* )patch1, uv, v1 );

	if( chkEnableForSaveTime == false )
		SetEnableForSaveTime( triCalcType, p0, p1 );

	for( i=0; i<3; i++ )
		vec[ i ] = param[ 2 ] * ( p1[ i ] - p0[ i ] ) + param[ 0 ] * v0[ i ] + param[ 1 ] * v1[ i ];
}


void  CbxTri::
SetDisableForSaveTime( void )
{
	short  i;

	for( i=0; i<3; i++ )
		enableForSaveTime[ i ] = false;
}


bool  CbxTri::
ChkEnableForSaveTime( const short triCalcType, double p0[ 3 ], double p1[ 3 ] )
{
	if( enableForSaveTime[ triCalcType ] == false )
		return( false );

	CbxMath::Copy3( p0ForSaveTime[ triCalcType ], p0 );
	CbxMath::Copy3( p1ForSaveTime[ triCalcType ], p1 );

	return( true );
}


void  CbxTri::
SetEnableForSaveTime( const short triCalcType, const double p0[ 3 ], const double p1[ 3 ] )
{
	enableForSaveTime[ triCalcType ] = true;

	CbxMath::Copy3( p0, p0ForSaveTime[ triCalcType ] );
	CbxMath::Copy3( p1, p1ForSaveTime[ triCalcType ] );
}


bool  CbxTri::
GetDefaultCtlVec( const CbxCore::CtlVecForDefaultInVertex in[ 3 ], CbxCore::CtlVecForDefaultOutVertex out[ 3 ] )
{
	bool    result;
	double  posVec[ 3 ][ 2 ][ 3 ], srcCtlVec[ 3 ][ 2 ][ 3 ], dstCtlVec[ 3 ][ 2 ][ 3 ];

	result = GetPosToPos( in, posVec );
	if( result == true )
		return( true );
	result = CbxSurface::CtlVecOrgTo3D( 3, in, out, srcCtlVec );
	if( result == true )
		return( true );

	result = CbxSurface::GetDefaultCtlVec( 3, ( const CbxSurface::Double2x3* )posVec,
				 ( const CbxSurface::Double2x3* )srcCtlVec, dstCtlVec );
	if( result == true )
		return( true );

	result = CorrectDefaultCtlVec( in, dstCtlVec );
	if( result == true )
		return( true );

	result = CbxSurface::CtlVec3DtoOrg( 3, in, ( const CbxSurface::Double2x3* )dstCtlVec, out );
	if( result == true )
		return( true );

	return( false );
}


bool  CbxTri::
GetPosToPos( const CbxCore::CtlVecForDefaultInVertex vertex[], double posVec[ 3 ][ 2 ][ 3 ] )
{
	bool    result;
	short   i;

	for( i=0; i<3; i++ ) {
		result = GetPosToPosMain( i, vertex, posVec[ i ] );
		if( result == true )
			return( true );
	}

	return( false );
}


bool  CbxTri::
GetPosToPosMain( const short no, const CbxCore::CtlVecForDefaultInVertex vertex[], double posVec[ 2 ][ 3 ] )
{
	double  pos0[ 3 ], pos1[ 3 ];
	short   n2;

	n2 = ( no + 2 ) % 3;
	CbxVector::Pos0ForCtlVecForDefault( n2, 3, vertex, pos1 );

	if( CbxVector::Pos1ForCtlVecForDefault( no, 3, vertex, pos0 ) == true )
		return( true );
	CbxMath::Subtract3( pos1, pos0, posVec[ 0 ] );

	if( CbxVector::Pos2ForCtlVecForDefault( no, 3, vertex, pos0 ) == true )
		return( true );
	CbxMath::Subtract3( pos1, pos0, posVec[ 1 ] );

	return( false );
}


bool  CbxTri::
CorrectDefaultCtlVec( const CbxCore::CtlVecForDefaultInVertex vertex[], double ctlVec[ 3 ][ 2 ][ 3 ] )
{
	bool    result;
	short   i;

	for( i=0; i<3; i++ ) {
		result = CorrectDefaultCtlVecMain( i, vertex, ctlVec[ i ] );
		if( result == true )
			return( true );
	}

	return( false );
}


bool  CbxTri::
CorrectDefaultCtlVecMain( const short no, const CbxCore::CtlVecForDefaultInVertex vertex[], double ctlVec[ 2 ][ 3 ] )
{
	bool    result;
	double  vec[ 2 ][ 3 ], vec2[ 2 ][ 3 ], ctlVecE[ 3 ], len0, len1, newLen;
	short   n2, i;

	n2 = ( no + 2 ) % 3;
	result = CbxVector::Tan0ForCtlVecForDefault( n2, 3, vertex, vec[ 0 ] );
	if( result == true )
		return( true );

	n2 = ( no + 1 ) % 3;
	result = CbxVector::Tan2ForCtlVecForDefault( n2, 3, vertex, vec[ 1 ] );
	if( result == true )
		return( true );
	CbxMath::Negate3( vec[ 1 ], vec[ 1 ] );

	CbxMath::Ratio3( vec[ 0 ], vec[ 1 ], 1.0/3.0, vec2[ 0 ] );
	CbxMath::Ratio3( vec[ 0 ], vec[ 1 ], 2.0/3.0, vec2[ 1 ] );

	for( i=0; i<2; i++ ) {
		len0 = CbxMath::Length3( ctlVec[ i ] );
		len1 = CbxMath::Length3( vec2[ i ] );
		if( len0 > len1 ) {
			newLen = len0 + ( len0 - len1 ) * ratioForCorrectDefaultCtlVec;
			CbxMath::Normalize3( ctlVec[ i ], CbxMath::zero3, ctlVecE );
			CbxMath::Scale3( ctlVecE, newLen, ctlVec[ i ] );
		}
	}

	return( false );
}
