/* bxVector.cpp */

#include "bxVector.h"
#include "../../Math/bxMath.h"
#include "../OrgTo3D/bxOrgTo3D.h"
#include "../3DtoOrg/bx3DtoOrg.h"
#include "../Element/bxElement.h"



const double  CbxVector::minSinForShorterRadius =   0.1;



bool  CbxVector::
Pos( const short no, const short numVertex, const CbxCore::SurfaceForPickVertex vertex[], double pos[ 4 ][ 3 ] )
{
	Pos0( no, numVertex, vertex, pos[ 0 ] );
	if( Pos1( no, numVertex, vertex, pos[ 1 ] ) == true )
		return( true );
	if( Pos2( no, numVertex, vertex, pos[ 2 ] ) == true )
		return( true );
	Pos3( no, numVertex, vertex, pos[ 3 ] );

	return( false );
}


void  CbxVector::
Pos0( const short no, const short numVertex, const CbxCore::SurfaceForPickVertex vertex[], double pos[ 3 ] )
{
	CbxMath::Copy3( vertex[ no ].vertexPos, pos );
}


bool  CbxVector::
Pos1( const short no, const short numVertex, const CbxCore::SurfaceForPickVertex vertex[], double pos[ 3 ] )
{
	bool    result;
	double  org[ 3 ], tan[ 3 ];

	Pos0( no, numVertex, vertex, org );

	result = Tan0( no, numVertex, vertex, tan );
	if( result == true )
		return( true );
	
	CbxMath::Add3( org, tan, pos );

	return( false );
}


bool  CbxVector::
Pos2( const short no, const short numVertex, const CbxCore::SurfaceForPickVertex vertex[], double pos[ 3 ] )
{
	bool    result;
	double  org[ 3 ], tan[ 3 ];

	Pos3( no, numVertex, vertex, org );

	result = Tan2( no, numVertex, vertex, tan );
	if( result == true )
		return( true );
	
	CbxMath::Subtract3( org, tan, pos );

	return( false );
}


void  CbxVector::
Pos3( const short no, const short numVertex, const CbxCore::SurfaceForPickVertex vertex[], double pos[ 3 ] )
{
	short  p;

	p = ( no + 1 ) % numVertex;

	CbxMath::Copy3( vertex[ p ].vertexPos, pos );
}


bool  CbxVector::
Handle( const short no, const short numVertex, const CbxCore::SurfaceForPickVertex vertex[], double vec[ 4 ][ 3 ] )
{
	if( Handle0( no, numVertex, vertex, vec[ 0 ] ) == true )
		return( true );
	if( Handle1( no, numVertex, vertex, vec[ 1 ] ) == true )
		return( true );
	if( Handle2( no, numVertex, vertex, vec[ 2 ] ) == true )
		return( true );
	if( Handle3( no, numVertex, vertex, vec[ 3 ] ) == true )
		return( true );

	return( false );
}


bool  CbxVector::
Handle0( const short no, const short numVertex, const CbxCore::SurfaceForPickVertex vertex[], double vec[ 3 ] )
{
	bool   result;
	short  n[ 2 ];
	CbxCore::HandleForTrans  handle;
	
	n[ 0 ] = ( no + numVertex - 1 ) % numVertex;
	n[ 1 ] = no;

	CbxMath::Copy3(    vertex[ n[1] ].vertexPos,           handle.in.vertexPos       );
	CbxMath::CopyAxis( vertex[ n[1] ].axisVec,             handle.in.axisVec         );
	CbxMath::Copy3(    vertex[ n[0] ].vertexPos,           handle.in.radialVertexPos );
	CbxMath::Copy2(    vertex[ n[1] ].edge[ 0 ].handleVec, handle.out.handleVec      );

	result = CbxOrgTo3Dex::GetHandle( &handle, vec );
	if( result == true )
		return( true );

	return( false );
}


bool  CbxVector::
Handle1( const short no, const short numVertex, const CbxCore::SurfaceForPickVertex vertex[], double vec[ 3 ] )
{
	bool   result;
	short  n[ 4 ], i;
	CbxCore::CtlVecForTrans  ctl;
	
	for( i=0; i<4; i++ )
		n[ i ] = ( no + i + numVertex - 1 ) % numVertex;

	CbxMath::Copy3(    vertex[ n[1] ].vertexPos,            ctl.in.sub[ 0 ].vertexPos       );
	CbxMath::CopyAxis( vertex[ n[1] ].axisVec,              ctl.in.sub[ 0 ].axisVec         );
	CbxMath::Copy2(    vertex[ n[1] ].edge[ 1 ].handleVec,  ctl.in.sub[ 0 ].handleVec       );
	CbxMath::Copy2(    vertex[ n[1] ].edge[ 1 ].wingHandle, ctl.in.sub[ 0 ].wingHandle[ 0 ] );
	CbxMath::Copy2(    vertex[ n[1] ].edge[ 0 ].handleVec,  ctl.in.sub[ 0 ].wingHandle[ 1 ] );
	ctl.in.sub[ 0 ].lenRadialVec[ 0 ] = vertex[ n[1] ].edge[ 1 ].lenRadialVec;
	ctl.in.sub[ 0 ].lenRadialVec[ 1 ] = CbxMath::Distance3( vertex[ n[0] ].vertexPos, vertex[ n[1] ].vertexPos );

	CbxMath::Copy3(    vertex[ n[2] ].vertexPos,            ctl.in.sub[ 1 ].vertexPos       );
	CbxMath::CopyAxis( vertex[ n[2] ].axisVec,              ctl.in.sub[ 1 ].axisVec         );
	CbxMath::Copy2(    vertex[ n[2] ].edge[ 0 ].handleVec,  ctl.in.sub[ 1 ].handleVec       );
	CbxMath::Copy2(    vertex[ n[2] ].edge[ 1 ].handleVec,  ctl.in.sub[ 1 ].wingHandle[ 0 ] );
	CbxMath::Copy2(    vertex[ n[2] ].edge[ 0 ].wingHandle, ctl.in.sub[ 1 ].wingHandle[ 1 ] );
	ctl.in.sub[ 1 ].lenRadialVec[ 0 ] = CbxMath::Distance3( vertex[ n[2] ].vertexPos, vertex[ n[3] ].vertexPos );
	ctl.in.sub[ 1 ].lenRadialVec[ 1 ] = vertex[ n[2] ].edge[ 0 ].lenRadialVec;

	ctl.in.ctlPlaneDeg = vertex[ n[1] ].edge[ 1 ].ctlPlaneDeg;

	CbxMath::Copy2( vertex[ n[1] ].edge[ 1 ].ctlVec, ctl.out.ctlVec );

	result = CbxOrgTo3Dex::GetCtlInCtlVec( &ctl, vec );
	if( result == true )
		return( true );

	return( false );
}


bool  CbxVector::
Handle2( const short no, const short numVertex, const CbxCore::SurfaceForPickVertex vertex[], double vec[ 3 ] )
{
	bool   result;
	short  n[ 4 ], i;
	CbxCore::CtlVecForTrans  ctl;
	
	for( i=0; i<4; i++ )
		n[ i ] = ( no + i + numVertex - 1 ) % numVertex;

	CbxMath::Copy3(    vertex[ n[2] ].vertexPos,            ctl.in.sub[ 0 ].vertexPos       );
	CbxMath::CopyAxis( vertex[ n[2] ].axisVec,              ctl.in.sub[ 0 ].axisVec         );
	CbxMath::Copy2(    vertex[ n[2] ].edge[ 0 ].handleVec,  ctl.in.sub[ 0 ].handleVec       );
	CbxMath::Copy2(    vertex[ n[2] ].edge[ 1 ].handleVec,  ctl.in.sub[ 0 ].wingHandle[ 0 ] );
	CbxMath::Copy2(    vertex[ n[2] ].edge[ 0 ].wingHandle, ctl.in.sub[ 0 ].wingHandle[ 1 ] );
	ctl.in.sub[ 0 ].lenRadialVec[ 0 ] = CbxMath::Distance3( vertex[ n[2] ].vertexPos, vertex[ n[3] ].vertexPos );
	ctl.in.sub[ 0 ].lenRadialVec[ 1 ] = vertex[ n[2] ].edge[ 0 ].lenRadialVec;

	CbxMath::Copy3(    vertex[ n[1] ].vertexPos,            ctl.in.sub[ 1 ].vertexPos       );
	CbxMath::CopyAxis( vertex[ n[1] ].axisVec,              ctl.in.sub[ 1 ].axisVec         );
	CbxMath::Copy2(    vertex[ n[1] ].edge[ 1 ].handleVec,  ctl.in.sub[ 1 ].handleVec       );
	CbxMath::Copy2(    vertex[ n[1] ].edge[ 1 ].wingHandle, ctl.in.sub[ 1 ].wingHandle[ 0 ] );
	CbxMath::Copy2(    vertex[ n[1] ].edge[ 0 ].handleVec,  ctl.in.sub[ 1 ].wingHandle[ 1 ] );
	ctl.in.sub[ 1 ].lenRadialVec[ 0 ] = vertex[ n[1] ].edge[ 1 ].lenRadialVec;
	ctl.in.sub[ 1 ].lenRadialVec[ 1 ] = CbxMath::Distance3( vertex[ n[0] ].vertexPos, vertex[ n[1] ].vertexPos );

	ctl.in.ctlPlaneDeg = vertex[ n[2] ].edge[ 0 ].ctlPlaneDeg;

	CbxMath::Copy2( vertex[ n[2] ].edge[ 0 ].ctlVec, ctl.out.ctlVec );

	result = CbxOrgTo3Dex::GetCtlInCtlVec( &ctl, vec );
	if( result == true )
		return( true );

	return( false );
}


bool  CbxVector::
Handle3( const short no, const short numVertex, const CbxCore::SurfaceForPickVertex vertex[], double vec[ 3 ] )
{
	bool   result;
	short  n[ 2 ];
	CbxCore::HandleForTrans  handle;
	
	n[ 0 ] = ( no + 1 ) % numVertex;
	n[ 1 ] = ( no + 2 ) % numVertex;

	CbxMath::Copy3(    vertex[ n[0] ].vertexPos,           handle.in.vertexPos       );
	CbxMath::CopyAxis( vertex[ n[0] ].axisVec,             handle.in.axisVec         );
	CbxMath::Copy3(    vertex[ n[1] ].vertexPos,           handle.in.radialVertexPos );
	CbxMath::Copy2(    vertex[ n[0] ].edge[ 1 ].handleVec, handle.out.handleVec      );

	result = CbxOrgTo3Dex::GetHandle( &handle, vec );
	if( result == true )
		return( true );

	return( false );
}


bool  CbxVector::
AntiHandle( const short no, const short numVertex, const CbxCore::SurfaceForPickVertex vertex[], double vec[ 4 ][ 3 ] )
{
	if( AntiHandle0( no, numVertex, vertex, vec[ 0 ] ) == true )
		return( true );
	if( AntiHandle1( no, numVertex, vertex, vec[ 1 ] ) == true )
		return( true );
	if( AntiHandle2( no, numVertex, vertex, vec[ 2 ] ) == true )
		return( true );
	if( AntiHandle3( no, numVertex, vertex, vec[ 3 ] ) == true )
		return( true );

	return( false );
}


bool  CbxVector::
AntiHandle0( const short no, const short numVertex, const CbxCore::SurfaceForPickVertex vertex[], double vec[ 3 ] )
{
	bool    result;
	short   n[ 2 ];
	double  axisVec[ 2 ][ 3 ], lenRadialVec, iHandleVec[ 2 ];
	
	n[ 0 ] = ( no + numVertex - 1 ) % numVertex;
	n[ 1 ] = no;

	CbxMath::CopyAxis( vertex[ n[1] ].axisVec,              axisVec    );
	CbxMath::Copy2(    vertex[ n[1] ].edge[ 1 ].wingHandle, iHandleVec );
	lenRadialVec = vertex[ n[1] ].edge[ 1 ].lenRadialVec;

	result = CbxOrgTo3D::HandleVec( axisVec, lenRadialVec, iHandleVec, vec );
	if( result == true )
		return( true );

	return( false );
}


bool  CbxVector::
AntiHandle1( const short no, const short numVertex, const CbxCore::SurfaceForPickVertex vertex[], double vec[ 3 ] )
{
	bool   result;
	short  n[ 4 ], i;
	CbxCore::CtlVecForTrans  ctl;
	
	for( i=0; i<4; i++ )
		n[ i ] = ( no + i + numVertex - 1 ) % numVertex;

	CbxMath::Copy3(    vertex[ n[1] ].vertexPos,            ctl.in.sub[ 0 ].vertexPos       );
	CbxMath::CopyAxis( vertex[ n[1] ].axisVec,              ctl.in.sub[ 0 ].axisVec         );
	CbxMath::Copy2(    vertex[ n[1] ].edge[ 1 ].handleVec,  ctl.in.sub[ 0 ].handleVec       );
	CbxMath::Copy2(    vertex[ n[1] ].edge[ 1 ].wingHandle, ctl.in.sub[ 0 ].wingHandle[ 0 ] );
	CbxMath::Copy2(    vertex[ n[1] ].edge[ 0 ].handleVec,  ctl.in.sub[ 0 ].wingHandle[ 1 ] );
	ctl.in.sub[ 0 ].lenRadialVec[ 0 ] = vertex[ n[1] ].edge[ 1 ].lenRadialVec;
	ctl.in.sub[ 0 ].lenRadialVec[ 1 ] = CbxMath::Distance3( vertex[ n[0] ].vertexPos, vertex[ n[1] ].vertexPos );

	CbxMath::Copy3(    vertex[ n[2] ].vertexPos,            ctl.in.sub[ 1 ].vertexPos       );
	CbxMath::CopyAxis( vertex[ n[2] ].axisVec,              ctl.in.sub[ 1 ].axisVec         );
	CbxMath::Copy2(    vertex[ n[2] ].edge[ 0 ].handleVec,  ctl.in.sub[ 1 ].handleVec       );
	CbxMath::Copy2(    vertex[ n[2] ].edge[ 1 ].handleVec,  ctl.in.sub[ 1 ].wingHandle[ 0 ] );
	CbxMath::Copy2(    vertex[ n[2] ].edge[ 0 ].wingHandle, ctl.in.sub[ 1 ].wingHandle[ 1 ] );
	ctl.in.sub[ 1 ].lenRadialVec[ 0 ] = CbxMath::Distance3( vertex[ n[2] ].vertexPos, vertex[ n[3] ].vertexPos );
	ctl.in.sub[ 1 ].lenRadialVec[ 1 ] = vertex[ n[2] ].edge[ 0 ].lenRadialVec;

	ctl.in.ctlPlaneDeg = vertex[ n[1] ].edge[ 1 ].ctlPlaneDeg;

	CbxMath::Copy2( vertex[ n[1] ].edge[ 1 ].antiCtlVec, ctl.out.ctlVec );

	result = CbxOrgTo3Dex::GetCtlInCtlVec( &ctl, vec );
	if( result == true )
		return( true );

	return( false );
}


bool  CbxVector::
AntiHandle2( const short no, const short numVertex, const CbxCore::SurfaceForPickVertex vertex[], double vec[ 3 ] )
{
	bool   result;
	short  n[ 4 ], i;
	CbxCore::CtlVecForTrans  ctl;
	
	for( i=0; i<4; i++ )
		n[ i ] = ( no + i + numVertex - 1 ) % numVertex;

	CbxMath::Copy3(    vertex[ n[2] ].vertexPos,            ctl.in.sub[ 0 ].vertexPos       );
	CbxMath::CopyAxis( vertex[ n[2] ].axisVec,              ctl.in.sub[ 0 ].axisVec         );
	CbxMath::Copy2(    vertex[ n[2] ].edge[ 0 ].handleVec,  ctl.in.sub[ 0 ].handleVec       );
	CbxMath::Copy2(    vertex[ n[2] ].edge[ 1 ].handleVec,  ctl.in.sub[ 0 ].wingHandle[ 0 ] );
	CbxMath::Copy2(    vertex[ n[2] ].edge[ 0 ].wingHandle, ctl.in.sub[ 0 ].wingHandle[ 1 ] );
	ctl.in.sub[ 0 ].lenRadialVec[ 0 ] = CbxMath::Distance3( vertex[ n[2] ].vertexPos, vertex[ n[3] ].vertexPos );
	ctl.in.sub[ 0 ].lenRadialVec[ 1 ] = vertex[ n[2] ].edge[ 0 ].lenRadialVec;

	CbxMath::Copy3(    vertex[ n[1] ].vertexPos,            ctl.in.sub[ 1 ].vertexPos       );
	CbxMath::CopyAxis( vertex[ n[1] ].axisVec,              ctl.in.sub[ 1 ].axisVec         );
	CbxMath::Copy2(    vertex[ n[1] ].edge[ 1 ].handleVec,  ctl.in.sub[ 1 ].handleVec       );
	CbxMath::Copy2(    vertex[ n[1] ].edge[ 1 ].wingHandle, ctl.in.sub[ 1 ].wingHandle[ 0 ] );
	CbxMath::Copy2(    vertex[ n[1] ].edge[ 0 ].handleVec,  ctl.in.sub[ 1 ].wingHandle[ 1 ] );
	ctl.in.sub[ 1 ].lenRadialVec[ 0 ] = vertex[ n[1] ].edge[ 1 ].lenRadialVec;
	ctl.in.sub[ 1 ].lenRadialVec[ 1 ] = CbxMath::Distance3( vertex[ n[0] ].vertexPos, vertex[ n[1] ].vertexPos );

	ctl.in.ctlPlaneDeg = vertex[ n[2] ].edge[ 0 ].ctlPlaneDeg;

	CbxMath::Copy2( vertex[ n[2] ].edge[ 0 ].antiCtlVec, ctl.out.ctlVec );

	result = CbxOrgTo3Dex::GetCtlInCtlVec( &ctl, vec );
	if( result == true )
		return( true );

	return( false );
}


bool  CbxVector::
AntiHandle3( const short no, const short numVertex, const CbxCore::SurfaceForPickVertex vertex[], double vec[ 3 ] )
{
	bool    result;
	short   n[ 2 ];
	double  axisVec[ 2 ][ 3 ], lenRadialVec, iHandleVec[ 2 ];
	
	n[ 0 ] = ( no + 1 ) % numVertex;
	n[ 1 ] = ( no + 2 ) % numVertex;

	CbxMath::CopyAxis( vertex[ n[0] ].axisVec,              axisVec    );
	CbxMath::Copy2(    vertex[ n[0] ].edge[ 0 ].wingHandle, iHandleVec );
	lenRadialVec = vertex[ n[0] ].edge[ 0 ].lenRadialVec;

	result = CbxOrgTo3D::HandleVec( axisVec, lenRadialVec, iHandleVec, vec );
	if( result == true )
		return( true );

	return( false );
}


bool  CbxVector::
Tan( const short no, const short numVertex, const CbxCore::SurfaceForPickVertex vertex[], double vec[ 3 ][ 3 ] )
{
	if( Tan0( no, numVertex, vertex, vec[ 0 ] ) == true )
		return( true );
	if( Tan1( no, numVertex, vertex, vec[ 1 ] ) == true )
		return( true );
	if( Tan2( no, numVertex, vertex, vec[ 2 ] ) == true )
		return( true );

	return( false );
}


bool  CbxVector::
Tan0( const short no, const short numVertex, const CbxCore::SurfaceForPickVertex vertex[], double vec[ 3 ] )
{
	bool   result;
	short  n[ 2 ];
	CbxCore::HandleForTrans  handle;
	
	n[ 0 ] = no;
	n[ 1 ] = ( no + 1 ) % numVertex;

	CbxMath::Copy3(    vertex[ n[0] ].vertexPos,           handle.in.vertexPos       );
	CbxMath::CopyAxis( vertex[ n[0] ].axisVec,             handle.in.axisVec         );
	CbxMath::Copy3(    vertex[ n[1] ].vertexPos,           handle.in.radialVertexPos );
	CbxMath::Copy2(    vertex[ n[0] ].edge[ 1 ].handleVec, handle.out.handleVec      );

	result = CbxOrgTo3Dex::GetHandle( &handle, vec );
	if( result == true )
		return( true );

	return( false );
}


bool  CbxVector::
Tan1( const short no, const short numVertex, const CbxCore::SurfaceForPickVertex vertex[], double vec[ 3 ] )
{
	bool    result;
	double  p1[ 3 ], p2[ 3 ];

	result = Pos1( no, numVertex, vertex, p1 );
	if( result == true )
		return( true );
	result = Pos2( no, numVertex, vertex, p2 );
	if( result == true )
		return( true );
	
	CbxMath::Subtract3( p2, p1, vec );

	return( false );
}


bool  CbxVector::
Tan2( const short no, const short numVertex, const CbxCore::SurfaceForPickVertex vertex[], double vec[ 3 ] )
{
	bool   result;
	short  n[ 2 ];
	CbxCore::HandleForTrans  handle;
	
	n[ 0 ] = no;
	n[ 1 ] = ( no + 1 ) % numVertex;

	CbxMath::Copy3(    vertex[ n[1] ].vertexPos,           handle.in.vertexPos       );
	CbxMath::CopyAxis( vertex[ n[1] ].axisVec,             handle.in.axisVec         );
	CbxMath::Copy3(    vertex[ n[0] ].vertexPos,           handle.in.radialVertexPos );
	CbxMath::Copy2(    vertex[ n[1] ].edge[ 0 ].handleVec, handle.out.handleVec      );

	result = CbxOrgTo3Dex::GetHandle( &handle, vec );
	if( result == true )
		return( true );

	CbxMath::Negate3( vec, vec );

	return( false );
}


void  CbxVector::
Pos0ForCtlVecForDefault( const short no, const short numVertex, const CbxCore::CtlVecForDefaultInVertex vertex[],
	double pos[ 3 ] )
{
	CbxMath::Copy3( vertex[ no ].vertexPos, pos );
}


bool  CbxVector::
Pos1ForCtlVecForDefault( const short no, const short numVertex, const CbxCore::CtlVecForDefaultInVertex vertex[],
	double pos[ 3 ] )
{
	bool    result;
	double  org[ 3 ], tan[ 3 ];

	Pos0ForCtlVecForDefault( no, numVertex, vertex, org );

	result = Tan0ForCtlVecForDefault( no, numVertex, vertex, tan );
	if( result == true )
		return( true );
	
	CbxMath::Add3( org, tan, pos );

	return( false );
}


bool  CbxVector::
Pos2ForCtlVecForDefault( const short no, const short numVertex, const CbxCore::CtlVecForDefaultInVertex vertex[],
	double pos[ 3 ] )
{
	bool    result;
	double  org[ 3 ], tan[ 3 ];

	Pos3ForCtlVecForDefault( no, numVertex, vertex, org );

	result = Tan2ForCtlVecForDefault( no, numVertex, vertex, tan );
	if( result == true )
		return( true );
	
	CbxMath::Subtract3( org, tan, pos );

	return( false );
}


void  CbxVector::
Pos3ForCtlVecForDefault( const short no, const short numVertex, const CbxCore::CtlVecForDefaultInVertex vertex[],
	double pos[ 3 ] )
{
	short  p;

	p = ( no + 1 ) % numVertex;

	CbxMath::Copy3( vertex[ p ].vertexPos, pos );
}


bool  CbxVector::
Handle1ForCtlVecForDefault( const short no, const short numVertex, const CbxCore::CtlVecForDefaultInVertex in[],
	const CbxCore::CtlVecForDefaultOutVertex out[], double vec[ 3 ] )
{
	bool   result;
	short  n[ 4 ], i;
	CbxCore::CtlVecForTrans  ctl;
	
	for( i=0; i<4; i++ )
		n[ i ] = ( no + i + numVertex - 1 ) % numVertex;

	CbxMath::Copy3(    in[ n[1] ].vertexPos,            ctl.in.sub[ 0 ].vertexPos       );
	CbxMath::CopyAxis( in[ n[1] ].axisVec,              ctl.in.sub[ 0 ].axisVec         );
	CbxMath::Copy2(    in[ n[1] ].edge[ 1 ].handleVec,  ctl.in.sub[ 0 ].handleVec       );
	CbxMath::Copy2(    in[ n[1] ].edge[ 1 ].wingHandle, ctl.in.sub[ 0 ].wingHandle[ 0 ] );
	CbxMath::Copy2(    in[ n[1] ].edge[ 0 ].handleVec,  ctl.in.sub[ 0 ].wingHandle[ 1 ] );
	ctl.in.sub[ 0 ].lenRadialVec[ 0 ] = in[ n[1] ].edge[ 1 ].lenRadialVec;
	ctl.in.sub[ 0 ].lenRadialVec[ 1 ] = CbxMath::Distance3( in[ n[0] ].vertexPos, in[ n[1] ].vertexPos );

	CbxMath::Copy3(    in[ n[2] ].vertexPos,            ctl.in.sub[ 1 ].vertexPos       );
	CbxMath::CopyAxis( in[ n[2] ].axisVec,              ctl.in.sub[ 1 ].axisVec         );
	CbxMath::Copy2(    in[ n[2] ].edge[ 0 ].handleVec,  ctl.in.sub[ 1 ].handleVec       );
	CbxMath::Copy2(    in[ n[2] ].edge[ 1 ].handleVec,  ctl.in.sub[ 1 ].wingHandle[ 0 ] );
	CbxMath::Copy2(    in[ n[2] ].edge[ 0 ].wingHandle, ctl.in.sub[ 1 ].wingHandle[ 1 ] );
	ctl.in.sub[ 1 ].lenRadialVec[ 0 ] = CbxMath::Distance3( in[ n[2] ].vertexPos, in[ n[3] ].vertexPos );
	ctl.in.sub[ 1 ].lenRadialVec[ 1 ] = in[ n[2] ].edge[ 0 ].lenRadialVec;

	ctl.in.ctlPlaneDeg = in[ n[1] ].edge[ 1 ].ctlPlaneDeg;

	CbxMath::Copy2( out[ n[1] ].edge[ 1 ].ctlVec, ctl.out.ctlVec );

	result = CbxOrgTo3Dex::GetCtlInCtlVec( &ctl, vec );
	if( result == true )
		return( true );

	return( false );
}


bool  CbxVector::
Handle2ForCtlVecForDefault( const short no, const short numVertex, const CbxCore::CtlVecForDefaultInVertex in[],
	const CbxCore::CtlVecForDefaultOutVertex out[], double vec[ 3 ] )
{
	bool   result;
	short  n[ 4 ], i;
	CbxCore::CtlVecForTrans  ctl;
	
	for( i=0; i<4; i++ )
		n[ i ] = ( no + i + numVertex - 1 ) % numVertex;

	CbxMath::Copy3(    in[ n[2] ].vertexPos,            ctl.in.sub[ 0 ].vertexPos       );
	CbxMath::CopyAxis( in[ n[2] ].axisVec,              ctl.in.sub[ 0 ].axisVec         );
	CbxMath::Copy2(    in[ n[2] ].edge[ 0 ].handleVec,  ctl.in.sub[ 0 ].handleVec       );
	CbxMath::Copy2(    in[ n[2] ].edge[ 1 ].handleVec,  ctl.in.sub[ 0 ].wingHandle[ 0 ] );
	CbxMath::Copy2(    in[ n[2] ].edge[ 0 ].wingHandle, ctl.in.sub[ 0 ].wingHandle[ 1 ] );
	ctl.in.sub[ 0 ].lenRadialVec[ 0 ] = CbxMath::Distance3( in[ n[2] ].vertexPos, in[ n[3] ].vertexPos );
	ctl.in.sub[ 0 ].lenRadialVec[ 1 ] = in[ n[2] ].edge[ 0 ].lenRadialVec;

	CbxMath::Copy3(    in[ n[1] ].vertexPos,            ctl.in.sub[ 1 ].vertexPos       );
	CbxMath::CopyAxis( in[ n[1] ].axisVec,              ctl.in.sub[ 1 ].axisVec         );
	CbxMath::Copy2(    in[ n[1] ].edge[ 1 ].handleVec,  ctl.in.sub[ 1 ].handleVec       );
	CbxMath::Copy2(    in[ n[1] ].edge[ 1 ].wingHandle, ctl.in.sub[ 1 ].wingHandle[ 0 ] );
	CbxMath::Copy2(    in[ n[1] ].edge[ 0 ].handleVec,  ctl.in.sub[ 1 ].wingHandle[ 1 ] );
	ctl.in.sub[ 1 ].lenRadialVec[ 0 ] = in[ n[1] ].edge[ 1 ].lenRadialVec;
	ctl.in.sub[ 1 ].lenRadialVec[ 1 ] = CbxMath::Distance3( in[ n[0] ].vertexPos, in[ n[1] ].vertexPos );

	ctl.in.ctlPlaneDeg = in[ n[2] ].edge[ 0 ].ctlPlaneDeg;

	CbxMath::Copy2( out[ n[2] ].edge[ 0 ].ctlVec, ctl.out.ctlVec );

	result = CbxOrgTo3Dex::GetCtlInCtlVec( &ctl, vec );
	if( result == true )
		return( true );

	return( false );
}


bool  CbxVector::
Tan0ForCtlVecForDefault( const short no, const short numVertex, const CbxCore::CtlVecForDefaultInVertex vertex[],
	double vec[ 3 ] )
{
	bool   result;
	short  n[ 2 ];
	CbxCore::HandleForTrans  handle;
	
	n[ 0 ] = no;
	n[ 1 ] = ( no + 1 ) % numVertex;

	CbxMath::Copy3(    vertex[ n[0] ].vertexPos,           handle.in.vertexPos       );
	CbxMath::CopyAxis( vertex[ n[0] ].axisVec,             handle.in.axisVec         );
	CbxMath::Copy3(    vertex[ n[1] ].vertexPos,           handle.in.radialVertexPos );
	CbxMath::Copy2(    vertex[ n[0] ].edge[ 1 ].handleVec, handle.out.handleVec      );

	result = CbxOrgTo3Dex::GetHandle( &handle, vec );
	if( result == true )
		return( true );

	return( false );
}


bool  CbxVector::
Tan2ForCtlVecForDefault( const short no, const short numVertex, const CbxCore::CtlVecForDefaultInVertex vertex[],
	double vec[ 3 ] )
{
	bool   result;
	short  n[ 2 ];
	CbxCore::HandleForTrans  handle;
	
	n[ 0 ] = no;
	n[ 1 ] = ( no + 1 ) % numVertex;

	CbxMath::Copy3(    vertex[ n[1] ].vertexPos,           handle.in.vertexPos       );
	CbxMath::CopyAxis( vertex[ n[1] ].axisVec,             handle.in.axisVec         );
	CbxMath::Copy3(    vertex[ n[0] ].vertexPos,           handle.in.radialVertexPos );
	CbxMath::Copy2(    vertex[ n[1] ].edge[ 0 ].handleVec, handle.out.handleVec      );

	result = CbxOrgTo3Dex::GetHandle( &handle, vec );
	if( result == true )
		return( true );

	CbxMath::Negate3( vec, vec );

	return( false );
}


bool  CbxVector::
GetDefaultCtlVec( const double posVec[ 3 ], const double srcCtlVec[ 3 ], double dstCtlVec[ 3 ] )
{
	bool    result;
	double  tanE[ 3 ];

	CbxMath::Normalize3( srcCtlVec, CbxMath::zero3, tanE );
	if( CbxMath::Length3( tanE ) < CbxMath::nearZero )
		return( true );

	result = GetDefaultHandle( posVec, tanE, dstCtlVec );
	if( result == true )
		return( true );

	return ( false );
}


bool  CbxVector::
GetDefaultHandle( const double radialVec[ 3 ], const double tanE[ 3 ], double tan[ 3 ] )
{
	bool  result;

	result = IsLongerRadius( radialVec, tanE );
	if( result == true ) {
		GetDefaultHandleLonger( radialVec, tanE, tan );
		return( false );
	}
	else
		return( GetDefaultHandleShorter( radialVec, tanE, tan ) );
}


bool  CbxVector::
IsLongerRadius( const double radialVec[ 3 ], const double tanE[ 3 ] )
{
	double  radialVecE[ 3 ], cross[ 3 ], sinP;

	CbxMath::Normalize3( radialVec, CbxMath::zero3, radialVecE );
	CbxMath::Cross3( radialVecE, tanE, cross );
	sinP = CbxMath::Length3( cross );
	if( sinP < minSinForShorterRadius )
		return( true );

	return( false );
}


void  CbxVector::
GetDefaultHandleLonger( const double radialVec[ 3 ], const double tanE[ 3 ], double tan[ 3 ] )
{
	double  len;

	len = CbxMath::Length3( radialVec );
	CbxMath::Scale3( tanE, len/3.0, tan );
}


bool  CbxVector::
GetDefaultHandleShorter( const double radialVec[ 3 ], const double tanE[ 3 ], double tan[ 3 ] )
{
	double  tmp[ 3 ], cd[ 3 ], c[ 3 ], d[ 3 ], de[ 3 ], l, lenB2, h, lenC, lenD, dot, m, s;

	CbxMath::Cross3( tanE, radialVec, tmp );
	CbxMath::Cross3( tmp, tanE, cd );

	l     = CbxMath::Dot3( cd, radialVec );
	lenB2 = CbxMath::LengthSquared3( radialVec );
	h = lenB2 / ( 2 * l );

	CbxMath::Scale3( cd, h, c );
	lenC = CbxMath::Length3( c );

	CbxMath::Scale3( radialVec, 1.0/2.0, tmp );
	CbxMath::Subtract3( tmp, c, d );
	lenD = CbxMath::Length3( d );
	if( lenD < CbxMath::nearZero )
		return( true );

	CbxMath::Scale3( d, 1.0/lenD, de );
	m = lenC - lenD;

	dot = CbxMath::Dot3( tanE, de );
	s   = ( 4.0 * m ) / ( 3.0 * dot );

	CbxMath::Scale3( tanE, s, tan );

	return( false );
}


bool  CbxVector::
SetHandle1( const short no, const short numVertex, const CbxCore::CtlVecForDefaultInVertex in[],
	const double vec[ 3 ], CbxCore::CtlVecForDefaultOutVertex out[] )
{
	bool    result;
	short   n[ 4 ], i;
	double  axis[ 2 ][ 3 ], lenBaseCtlVec;
	CbxCore::CtlVecForTrans  ctl;
	
	for( i=0; i<4; i++ )
		n[ i ] = ( no + i + numVertex - 1 ) % numVertex;

	CbxMath::Copy3(    in[ n[1] ].vertexPos,            ctl.in.sub[ 0 ].vertexPos       );
	CbxMath::CopyAxis( in[ n[1] ].axisVec,              ctl.in.sub[ 0 ].axisVec         );
	CbxMath::Copy2(    in[ n[1] ].edge[ 1 ].handleVec,  ctl.in.sub[ 0 ].handleVec       );
	CbxMath::Copy2(    in[ n[1] ].edge[ 1 ].wingHandle, ctl.in.sub[ 0 ].wingHandle[ 0 ] );
	CbxMath::Copy2(    in[ n[1] ].edge[ 0 ].handleVec,  ctl.in.sub[ 0 ].wingHandle[ 1 ] );
	ctl.in.sub[ 0 ].lenRadialVec[ 0 ] = in[ n[1] ].edge[ 1 ].lenRadialVec;
	ctl.in.sub[ 0 ].lenRadialVec[ 1 ] = CbxMath::Distance3( in[ n[0] ].vertexPos, in[ n[1] ].vertexPos );

	CbxMath::Copy3(    in[ n[2] ].vertexPos,            ctl.in.sub[ 1 ].vertexPos       );
	CbxMath::CopyAxis( in[ n[2] ].axisVec,              ctl.in.sub[ 1 ].axisVec         );
	CbxMath::Copy2(    in[ n[2] ].edge[ 0 ].handleVec,  ctl.in.sub[ 1 ].handleVec       );
	CbxMath::Copy2(    in[ n[2] ].edge[ 1 ].handleVec,  ctl.in.sub[ 1 ].wingHandle[ 0 ] );
	CbxMath::Copy2(    in[ n[2] ].edge[ 0 ].wingHandle, ctl.in.sub[ 1 ].wingHandle[ 1 ] );
	ctl.in.sub[ 1 ].lenRadialVec[ 0 ] = CbxMath::Distance3( in[ n[2] ].vertexPos, in[ n[3] ].vertexPos );
	ctl.in.sub[ 1 ].lenRadialVec[ 1 ] = in[ n[2] ].edge[ 0 ].lenRadialVec;

	ctl.in.ctlPlaneDeg = in[ n[1] ].edge[ 1 ].ctlPlaneDeg;

	result = ElementAxisInCtlVecPlus( ctl.in.sub, ctl.in.ctlPlaneDeg, axis, &lenBaseCtlVec );
	if( result == true )
		return( true );

	result = Cbx3DtoOrg::CtlVec( axis, lenBaseCtlVec, vec, out[ n[1] ].edge[ 1 ].ctlVec );
	if( result == true )
		return( true );

	return( false );
}


bool  CbxVector::
SetHandle2( const short no, const short numVertex, const CbxCore::CtlVecForDefaultInVertex in[],
	const double vec[ 3 ], CbxCore::CtlVecForDefaultOutVertex out[] )
{
	bool    result;
	short   n[ 4 ], i;
	double  axis[ 2 ][ 3 ], lenBaseCtlVec;
	CbxCore::CtlVecForTrans  ctl;
	
	for( i=0; i<4; i++ )
		n[ i ] = ( no + i + numVertex - 1 ) % numVertex;

	CbxMath::Copy3(    in[ n[2] ].vertexPos,            ctl.in.sub[ 0 ].vertexPos       );
	CbxMath::CopyAxis( in[ n[2] ].axisVec,              ctl.in.sub[ 0 ].axisVec         );
	CbxMath::Copy2(    in[ n[2] ].edge[ 0 ].handleVec,  ctl.in.sub[ 0 ].handleVec       );
	CbxMath::Copy2(    in[ n[2] ].edge[ 1 ].handleVec,  ctl.in.sub[ 0 ].wingHandle[ 0 ] );
	CbxMath::Copy2(    in[ n[2] ].edge[ 0 ].wingHandle, ctl.in.sub[ 0 ].wingHandle[ 1 ] );
	ctl.in.sub[ 0 ].lenRadialVec[ 0 ] = CbxMath::Distance3( in[ n[2] ].vertexPos, in[ n[3] ].vertexPos );
	ctl.in.sub[ 0 ].lenRadialVec[ 1 ] = in[ n[2] ].edge[ 0 ].lenRadialVec;

	CbxMath::Copy3(    in[ n[1] ].vertexPos,            ctl.in.sub[ 1 ].vertexPos       );
	CbxMath::CopyAxis( in[ n[1] ].axisVec,              ctl.in.sub[ 1 ].axisVec         );
	CbxMath::Copy2(    in[ n[1] ].edge[ 1 ].handleVec,  ctl.in.sub[ 1 ].handleVec       );
	CbxMath::Copy2(    in[ n[1] ].edge[ 1 ].wingHandle, ctl.in.sub[ 1 ].wingHandle[ 0 ] );
	CbxMath::Copy2(    in[ n[1] ].edge[ 0 ].handleVec,  ctl.in.sub[ 1 ].wingHandle[ 1 ] );
	ctl.in.sub[ 1 ].lenRadialVec[ 0 ] = in[ n[1] ].edge[ 1 ].lenRadialVec;
	ctl.in.sub[ 1 ].lenRadialVec[ 1 ] = CbxMath::Distance3( in[ n[0] ].vertexPos, in[ n[1] ].vertexPos );

	ctl.in.ctlPlaneDeg = in[ n[2] ].edge[ 0 ].ctlPlaneDeg;

	result = ElementAxisInCtlVecPlus( ctl.in.sub, ctl.in.ctlPlaneDeg, axis, &lenBaseCtlVec );
	if( result == true )
		return( true );

	result = Cbx3DtoOrg::CtlVec( axis, lenBaseCtlVec, vec, out[ n[2] ].edge[ 0 ].ctlVec );
	if( result == true )
		return( true );

	return( false );
}


bool  CbxVector::
ElementAxisInCtlVecPlus( const CbxCore::CtlVecForTransSub sub[ 2 ], const double ctlPlaneDeg, double axis[ 2 ][ 3 ],
	double* lenBaseCtlVec )
{
	bool    result, errResult;
	double  handleVec[ 2 ][ 3 ], crossBoundOnVertex[ 2 ][ 3 ];
	double  axisInCtlPlaneDeg[ 3 ], normalInCtlPlaneDeg[ 3 ], baseCtlVec[ 3 ];

	result = CbxElementEx::ElementOnCtlVec( sub, ctlPlaneDeg, handleVec, crossBoundOnVertex,
				 axisInCtlPlaneDeg, normalInCtlPlaneDeg );
	if( result == true )
		return( true );

	result = CbxElement::BaseCtlVec( handleVec, crossBoundOnVertex, normalInCtlPlaneDeg, axisInCtlPlaneDeg, baseCtlVec );
	if( result == true )
		return( true );

	*lenBaseCtlVec = CbxMath::Length3( baseCtlVec );

	CbxMath::Copy3( baseCtlVec, axis[ 0 ] );
	CbxMath::Copy3( axisInCtlPlaneDeg, axis[ 1 ] );
	CbxMath::RegularAxis( axis, axis, &errResult );
	if( errResult == true )
		return( true );

	return( false );
}
