/* bxCore.h */

#if !defined( BX_CORE_H )
#define  BX_CORE_H

#include "../bxType.h"


class CbxCore {
	public:
		struct AxisForDefaultSub {
			double  vertexPos[ 3 ];
		};
		struct AxisForDefaultIn {
			double              vertexPos[ 3 ];
			ushort              numRadial;
			bool                loopFlag;
			AxisForDefaultSub*  radial;
		};
		struct AxisForDefaultOut {
			double  axisVec[ 2 ][ 3 ];
		};
		struct AxisForDefault {
			AxisForDefaultIn   in;
			AxisForDefaultOut  out;
		};

		struct HandleForTransIn {
			double  vertexPos[ 3 ];
			double  axisVec[ 2 ][ 3 ];
			double  radialVertexPos[ 3 ];
		};
		struct HandleForTransOut {
			double  handleVec[ 2 ];
		};
		struct HandleForTrans {
			HandleForTransIn   in;
			HandleForTransOut  out;
		};

		struct HandleForTransPlusIn {
			double  vertexPos[ 3 ];
			double  radialVertexPos[ 3 ];
			double  wingHandle[ 2 ][ 2 ];
			double  lenRadialVec[ 2 ];
		};
		struct HandleForTransPlusOut {
			double  axisVec[ 2 ][ 3 ];
			double  handleVec[ 2 ];
		};
		struct HandleForTransPlus {
			HandleForTransPlusIn   in;
			HandleForTransPlusOut  out;
		};

		struct VertexForRotOut {
			double  vertexPos[ 3 ];
			double  axisVec[ 2 ][ 3 ];
		};
		struct VertexForRot {
			VertexForRotOut  out;
		};

		struct CtlPlaneDegForRotSub {
			double  vertexPos[ 3 ];
			double  axisVec[ 2 ][ 3 ];
			double  handleVec[ 2 ];
		};
		struct CtlPlaneDegForRotIn {
			CtlPlaneDegForRotSub  sub[ 2 ];
		};
		struct CtlPlaneDegForRotOut {
			double  ctlPlaneDeg;
		};
		struct CtlPlaneDegForRot {
			CtlPlaneDegForRotIn   in;
			CtlPlaneDegForRotOut  out;
		};

		struct CtlVecForTransSub {
			double  vertexPos[ 3 ];
			double  axisVec[ 2 ][ 3 ];
			double  handleVec[ 2 ];
			double  wingHandle[ 2 ][ 2 ];
			double  lenRadialVec[ 2 ];
		};

		struct CtlVecForTransIn {
			CtlVecForTransSub  sub[ 2 ];
			double             ctlPlaneDeg;
		};
		struct CtlVecForTransOut {
			double  ctlVec[ 2 ];
		};
		struct CtlVecForTrans {
			CtlVecForTransIn   in;
			CtlVecForTransOut  out;
		};

		struct CtlVecForTransPlusIn {
			short              minusPlusNo;
			CtlVecForTransSub  sub[ 2 ];
		};
		struct CtlVecForTransPlusOut {
			double  ctlPlaneDeg;
			double  ctlVec[ 2 ];
		};
		struct CtlVecForTransPlus {
			CtlVecForTransPlusIn   in;
			CtlVecForTransPlusOut  out;
		};

		struct CtlVecForDefaultInEdge {
			double  handleVec[ 2 ];
			double  wingHandle[ 2 ];
			double  lenRadialVec;
			double  ctlPlaneDeg;
			double  antiCtlVec[ 2 ];
		};
		struct CtlVecForDefaultInVertex {
			double                  vertexPos[ 3 ];
			double                  axisVec[ 2 ][ 3 ];
			CtlVecForDefaultInEdge  edge[ 2 ];
		};
		struct CtlVecForDefaultIn {
			short                     numVertex;
			CtlVecForDefaultInVertex  vertex[ 4 ];
		};
		struct CtlVecForDefaultOutEdge {
			double  ctlVec[ 2 ];
		};
		struct CtlVecForDefaultOutVertex {
			CtlVecForDefaultOutEdge  edge[ 2 ];
		};
		struct CtlVecForDefaultOut {
			CtlVecForDefaultOutVertex  vertex[ 4 ];
		};
		struct CtlVecForDefault {
			CtlVecForDefaultIn   in;
			CtlVecForDefaultOut  out;
		};

		struct SurfaceForPickEdge {
			double  handleVec[ 2 ];
			double  wingHandle[ 2 ];
			double  lenRadialVec;
			double  ctlPlaneDeg;
			double  ctlVec[ 2 ];
			double  antiCtlVec[ 2 ];
		};
		struct SurfaceForPickVertex {
			double              vertexPos[ 3 ];
			double              axisVec[ 2 ][ 3 ];
			SurfaceForPickEdge  edge[ 2 ];
		};
		struct SurfaceForPick {
			short                 numVertex;
			SurfaceForPickVertex  vertex[ 4 ];
		};

		struct PatchPnt {
			double  pos[ 3 ];
		};

		struct Patch5 {
			PatchPnt  p[ 5 ][ 5 ];
		};

		struct Patch6 {
			PatchPnt  p[ 6 ][ 6 ];
		};

		struct RectPatch {
			Patch5  data[ 2 ];
		};

		struct TriPatch {
			Patch5  form[ 3 ];
			Patch6  join[ 3 ];
		};

		struct PatchRec {
			short  numVertex;
			union {
				RectPatch  rect;
				TriPatch   tri[ 3 ];
			};
		};

	public:
		static bool
		ConvSurface( const SurfaceForPick* src, bool* dispChk, PatchRec* dst );
		static bool
		PosNormalSurface( const PatchRec* patch, const double uv[ 2 ], double pos[ 3 ], double normalVec[ 3 ] );
		static bool
		PosSurface( const PatchRec* patch, const double uv[ 2 ], double pos[ 3 ] );
		static bool
		NormalSurface( const PatchRec* patch, const double uv[ 2 ], double normalVec[ 3 ] );

	protected:
		static bool
		GetDefaultAxis( const AxisForDefaultIn* in, AxisForDefaultOut* out );
		static bool
		GetDefaultHandle( const HandleForTransIn* in, HandleForTransOut* out );
		static bool
		GetDefaultCtlVec( const CtlVecForDefaultIn* in, bool* dispChk, CtlVecForDefaultOut* out );

		static void
		SetVertexPos( const double newPos[ 3 ], double vertexPos[ 3 ] );
		static bool
		RotVertex( const double origin[ 3 ], const double axis[ 3 ], const double deg, VertexForRotOut* rot );
		static bool
		ScaleVertex( const double origin[ 3 ], const double scale, double vertexPos[ 3 ] );
		static void
		SetHandleVec( const double newHandle[ 2 ], double handleVec[ 2 ] );
		static bool
		SetHandleInfo( const double newHandleVec[ 3 ], const HandleForTransPlusIn* in, bool* dispChk,
			HandleForTransPlusOut* out );
		static void
		SetCtlPlaneDeg( const double newCtl, double* ctlPlaneDeg );
		static void
		SetCtlVec( const double newCtl[ 2 ][ 2 ], double ctlVec[ 2 ][ 2 ] );
		static bool
		SetCtlInfo( const double newCtlVec[ 3 ], const CtlVecForTransPlusIn* in, bool* dispChk,
			CtlVecForTransPlusOut* out );

		static void
		GetNormalOnVertex( const VertexForRot* data, double normalVec[ 3 ] );
		static bool
		GetHandle( const HandleForTrans* handle, bool* dispChk, double handleVec[ 3 ] );
		static bool
		GetAxisInCtlPlaneDeg( const CtlPlaneDegForRot* ctl, bool* dispChk, double axisVec[ 3 ] );
		static bool
		GetNormalInCtlPlaneDeg( const CtlPlaneDegForRot* ctl, bool* dispChk, double normalVec[ 3 ] );
		static bool
		GetAxisInCtlVec( const CtlVecForTrans* ctl, bool* dispChk, double axis[ 2 ][ 3 ] );
		static bool
		GetCtlInCtlVec( const CtlVecForTrans* ctl, bool* dispChk, double ctlVec[ 3 ] );

		static bool
		ConvEdge( const HandleForTrans vertex[ 2 ], bool* dispChk, double bezierLine[ 4 ][ 3 ] );

		static bool
		ErrChkAxisVec( const double axisVec[ 2 ][ 3 ] );
		static bool
		ErrChkHandleVec1( const double handleVec1 /* handleVec[ 1 ] */ );
		static bool
		ErrChkCtlPlaneDeg( const double ctlPlaneDeg );
		static bool
		ErrChkCtlVec( const double ctlVec[ 2 ][ 2 ] );

		static ushort
		MaxNumRadial( void );

		static bool
		DispChkOnEdgeSurface( const SurfaceForPick* data, bool* dispChk );
};


#endif
