/* bxDB.cpp */

#include "bxDB.h"
#include "../Math/bxMath.h"


ulong                       CbxDB::numVertexAry;
ulong                       CbxDB::numEdgeAry;
CbxDB::VertexRec*           CbxDB::vertexAry;
CbxDB::EdgeRec*             CbxDB::edgeAry;
CbxDB::EdgeForPickSurface*  CbxDB::edgePick;

const double  CbxDB::base_ctlPlaneDeg = 0.0;
const double  CbxDB::base_ctlVec0[ 2 ] = {   0.0, 1.0 };
const double  CbxDB::base_ctlVec1[ 2 ] = { 180.0, 1.0 };

bool   CbxDB::lockFlag			= false;
ulong  CbxDB::lockVertexNo;
ulong  CbxDB::lockEdgeNo;
short  CbxDB::lockBeginEndNo;
short  CbxDB::lockMinusPlusNo;



bool  CbxDB::
AllocData( const ulong numVertex, const ulong numEdge )
{
	numVertexAry = numVertex;
	numEdgeAry   = numEdge;

	if( AllocVertex() == true )
		return( true );
	if( AllocEdge() == true )
		return( true );

	return( false );
}


bool  CbxDB::
AllocVertex( void )
{
	vertexAry = new VertexRec[ numVertexAry ];
	if( vertexAry == 0 )
		return( true );

	return( false );
}


bool  CbxDB::
AllocEdge( void )
{
	edgeAry = new EdgeRec[ numEdgeAry ];
	if( edgeAry == 0 )
		return( true );

	return( false );
}


bool  CbxDB::
AllocEdgeForPickSurface( void )
{
	ulong  i;
	short  j;

	edgePick = new EdgeForPickSurface[ numEdgeAry ];
	if( edgePick == 0 )
		return( true );

	for( i=0; i<numEdgeAry; i++ ) {
		for( j=0; j<2; j++ )
			edgePick[ i ].alreadyPick[ j ] = false;
	}

	return( false );
}


void  CbxDB::
FreeData( void )
{
	FreeVertex();
	FreeEdge();
}


void  CbxDB::
FreeVertex( void )
{
	delete [] vertexAry;
}


void  CbxDB::
FreeEdge( void )
{
	delete [] edgeAry;
}


void  CbxDB::
FreeEdgeForPickSurface( void )
{
	delete [] edgePick;
}


void  CbxDB::
InitilizeEdgeRec( EdgeRec* edge )
{
	short  i;

	for( i=0; i<2; i++ ) {
		edge->vertex[ i ].ctlPlaneDeg = base_ctlPlaneDeg;
		CbxMath::Copy2( base_ctlVec0, edge->vertex[ i ].ctlVec[ 0 ] );
		CbxMath::Copy2( base_ctlVec1, edge->vertex[ i ].ctlVec[ 1 ] );
	}
}

		
bool  CbxDB::
CancelLock( void )
{
	if( lockFlag == false )
		return( true );
	lockFlag = false;

	return( false );
}


void  CbxDB::
LoadVertexRec( const ulong vertexNo, VertexRec* vertex )
{
	*vertex = vertexAry[ vertexNo ];
}


bool  CbxDB::
LockVertexRec( const ulong vertexNo, VertexRec* vertex )
{
	if( lockFlag == true )
		return( true );
	lockFlag     = true;
	lockVertexNo = vertexNo;

	*vertex = vertexAry[ vertexNo ];

	return( false );
}


bool  CbxDB::
UnlockVertexRec( const VertexRec* vertex )
{
	if( lockFlag == false )
		return( true );
	lockFlag = false;

	vertexAry[ lockVertexNo ] = *vertex;

	return( false );
}

		
void  CbxDB::
LoadEdgeRec( const ulong edgeNo, EdgeRec* edge )
{
	*edge = edgeAry[ edgeNo ];
}


bool  CbxDB::
LockEdgeRec( const ulong edgeNo, EdgeRec* edge )
{
	if( lockFlag == true )
		return( true );
	lockFlag   = true;
	lockEdgeNo = edgeNo;

	*edge = edgeAry[ edgeNo ];

	return( false );
}


bool  CbxDB::
UnlockEdgeRec( const EdgeRec* edge )
{
	if( lockFlag == false )
		return( true );
	lockFlag = false;

	edgeAry[ lockEdgeNo ] = *edge;

	return( false );
}

		
bool  CbxDB::
GetEdgeForPickSurface( const ulong edgeNo, const short beginEndNo )
{
	return( edgePick[ edgeNo ].alreadyPick[ beginEndNo ] );
}


void  CbxDB::
SetEdgeForPickSurface( const ulong edgeNo, const short beginEndNo )
{
	edgePick[ edgeNo ].alreadyPick[ beginEndNo ] = true;
}

		
void  CbxDB::
NumRadial( const ulong vertexNo, ushort* numRadial )
{
	*numRadial = vertexAry[ vertexNo ].numRadial;
}


bool  CbxDB::
LoadAxisForDefault( const ulong vertexNo, const CbxCore::AxisForDefaultSub buf[/*numRadial*/],
	CbxCore::AxisForDefault* axis )
{
	ushort  i;
	ulong   edgeNo, radialVertexNo;
	short   beginEndNo;
	bool    result;

	CbxMath::Copy3( vertexAry[ vertexNo ].vertexPos, axis->in.vertexPos );
	axis->in.numRadial = vertexAry[ vertexNo ].numRadial;
	axis->in.loopFlag  = vertexAry[ vertexNo ].loopFlag;
	axis->in.radial = ( CbxCore::AxisForDefaultSub* )buf;
	CbxMath::CopyAxis( vertexAry[ vertexNo ].axisVec, axis->out.axisVec );

	for( i=0; i<axis->in.numRadial; i++ ) {
		result = VertexRadialToEdge( vertexNo, i, &edgeNo, &beginEndNo );
		if( result == true )
			return( true );
		radialVertexNo = edgeAry[ edgeNo ].vertex[ 1 - beginEndNo ].vertexNo;
		CbxMath::Copy3( vertexAry[ radialVertexNo ].vertexPos, axis->in.radial[ i ].vertexPos );
	}

	return( false );
}


bool  CbxDB::
LockAxisForDefault( const ulong vertexNo, const CbxCore::AxisForDefaultSub buf[/*numRadial*/],
	CbxCore::AxisForDefault* axis )
{
	if( lockFlag == true )
		return( true );
	lockFlag     = true;
	lockVertexNo = vertexNo;

	return( LoadAxisForDefault( vertexNo, buf, axis ) );
}


bool  CbxDB::
VertexRadialToEdge( const ulong vertexNo, const ushort radialNo, ulong* edgeNo, short* beginEndNo )
{
	ulong  i;
	short  j;

	for( i=0; i<numEdgeAry; i++ ) {
		for( j=0; j<2; j++ ) {
			if( edgeAry[ i ].vertex[ j ].vertexNo == vertexNo &&
				edgeAry[ i ].vertex[ j ].radialNo == radialNo )
			{
				*edgeNo     = i;
				*beginEndNo = j;
				return( false );
			}
		}
	}

	return( true );
}


bool  CbxDB::
UnlockAxisForDefault( const CbxCore::AxisForDefaultOut* axis )
{
	if( lockFlag == false )
		return( true );
	lockFlag = false;

	CbxMath::CopyAxis( axis->axisVec, vertexAry[ lockVertexNo ].axisVec );

	return( false );
}


void  CbxDB::
LoadHandleForTrans( const ulong edgeNo, const short beginEndNo, CbxCore::HandleForTrans* handle )
{
	ulong  vertexNo, radialVertexNo;

	vertexNo       = edgeAry[ edgeNo ].vertex[ beginEndNo ].vertexNo;
	radialVertexNo = edgeAry[ edgeNo ].vertex[ 1 - beginEndNo ].vertexNo;

	CbxMath::Copy3( vertexAry[ vertexNo ].vertexPos, handle->in.vertexPos );
	CbxMath::CopyAxis( vertexAry[ vertexNo ].axisVec, handle->in.axisVec );
	CbxMath::Copy3( vertexAry[ radialVertexNo ].vertexPos, handle->in.radialVertexPos );

	CbxMath::Copy2( edgeAry[ edgeNo ].vertex[ beginEndNo ].handleVec, handle->out.handleVec );
}


bool  CbxDB::
LockHandleForTrans( const ulong edgeNo, const short beginEndNo, CbxCore::HandleForTrans* handle )
{
	if( lockFlag == true )
		return( true );
	lockFlag       = true;
	lockEdgeNo     = edgeNo;
	lockBeginEndNo = beginEndNo;

	LoadHandleForTrans( edgeNo, beginEndNo, handle );

	return( false );
}


bool  CbxDB::
UnlockHandleForTrans( const CbxCore::HandleForTransOut* handle )
{
	if( lockFlag == false )
		return( true );
	lockFlag = false;

	CbxMath::Copy2( handle->handleVec, edgeAry[ lockEdgeNo ].vertex[ lockBeginEndNo ].handleVec );

	return( false );
}


bool  CbxDB::
LoadHandleForTransPlus( const ulong edgeNo, const short beginEndNo, CbxCore::HandleForTransPlus* handle )
{
	ulong  vertexNo, radialVertexNo;
	bool   result;

	vertexNo       = edgeAry[ edgeNo ].vertex[ beginEndNo ].vertexNo;
	radialVertexNo = edgeAry[ edgeNo ].vertex[ 1 - beginEndNo ].vertexNo;

	CbxMath::Copy3( vertexAry[ vertexNo ].vertexPos, handle->in.vertexPos );
	CbxMath::Copy3( vertexAry[ radialVertexNo ].vertexPos, handle->in.radialVertexPos );

	result = LoadHandleForTransPlusSub( edgeNo, beginEndNo, &handle->in );
	if( result == true )
		return( true );

	CbxMath::CopyAxis( vertexAry[ vertexNo ].axisVec, handle->out.axisVec );
	CbxMath::Copy2( edgeAry[ edgeNo ].vertex[ beginEndNo ].handleVec, handle->out.handleVec );

	return( false );
}


bool  CbxDB::
LoadHandleForTransPlusSub( const ulong edgeNo, const short beginEndNo, CbxCore::HandleForTransPlusIn* handle )
{
	ulong   vertexNo;
	short   i;
	double  wing[ 2 ][ 2 ], lenRadialVec[ 2 ];
	ushort  numRadial;
	bool    loopFlag, result;

	vertexNo = edgeAry[ edgeNo ].vertex[ beginEndNo ].vertexNo;

	numRadial = vertexAry[ vertexNo ].numRadial;
	loopFlag  = vertexAry[ vertexNo ].loopFlag;

	result = LoadCtlVecForTransSubDecWing( edgeNo, beginEndNo, numRadial, loopFlag, wing[ 0 ], &lenRadialVec[ 0 ] );
	if( result == true )
		return( true );
	result = LoadCtlVecForTransSubIncWing( edgeNo, beginEndNo, numRadial, loopFlag, wing[ 1 ], &lenRadialVec[ 1 ] );
	if( result == true )
		return( true );

	for( i=0; i<2; i++ ) {
		CbxMath::Copy2( wing[ i ], handle->wingHandle[ i ] );
		handle->lenRadialVec[ i ] = lenRadialVec[ i ];
	}

	return( false );
}


bool  CbxDB::
LockHandleForTransPlus( const ulong edgeNo, const short beginEndNo, CbxCore::HandleForTransPlus* handle )
{
	if( lockFlag == true )
		return( true );
	lockFlag       = true;
	lockEdgeNo     = edgeNo;
	lockBeginEndNo = beginEndNo;

	return( LoadHandleForTransPlus( edgeNo, beginEndNo, handle ) );
}


bool  CbxDB::
UnlockHandleForTransPlus( const CbxCore::HandleForTransPlusOut* handle )
{
	ulong  vertexNo;

	if( lockFlag == false )
		return( true );
	lockFlag = false;

	vertexNo = edgeAry[ lockEdgeNo ].vertex[ lockBeginEndNo ].vertexNo;

	CbxMath::CopyAxis( handle->axisVec, vertexAry[ vertexNo ].axisVec );
	CbxMath::Copy2( handle->handleVec, edgeAry[ lockEdgeNo ].vertex[ lockBeginEndNo ].handleVec );

	return( false );
}


void  CbxDB::
LoadVertexForRot( const ulong vertexNo, CbxCore::VertexForRot* rot )
{
	CbxMath::Copy3( vertexAry[ vertexNo ].vertexPos, rot->out.vertexPos );
	CbxMath::CopyAxis( vertexAry[ vertexNo ].axisVec, rot->out.axisVec );
}


bool  CbxDB::
LockVertexForRot( const ulong vertexNo, CbxCore::VertexForRot* rot )
{
	if( lockFlag == true )
		return( true );
	lockFlag     = true;
	lockVertexNo = vertexNo;

	LoadVertexForRot( vertexNo, rot );

	return( false );
}


bool  CbxDB::
UnlockVertexForRot( const CbxCore::VertexForRotOut* rot )
{
	if( lockFlag == false )
		return( true );
	lockFlag = false;

	CbxMath::Copy3( rot->vertexPos, vertexAry[ lockVertexNo ].vertexPos );
	CbxMath::CopyAxis( rot->axisVec, vertexAry[ lockVertexNo ].axisVec );

	return( false );
}


void  CbxDB::
LoadCtlPlaneDegForRot( const ulong edgeNo, const short beginEndNo, CbxCore::CtlPlaneDegForRot* ctl )
{
	LoadCtlPlaneDegForRotSub( edgeNo, beginEndNo, ctl->in.sub );
	ctl->out.ctlPlaneDeg = edgeAry[ edgeNo ].vertex[ beginEndNo ].ctlPlaneDeg;
}


bool  CbxDB::
LockCtlPlaneDegForRot( const ulong edgeNo, const short beginEndNo, CbxCore::CtlPlaneDegForRot* ctl )
{
	if( lockFlag == true )
		return( true );
	lockFlag       = true;
	lockEdgeNo     = edgeNo;
	lockBeginEndNo = beginEndNo;

	LoadCtlPlaneDegForRot( edgeNo, beginEndNo, ctl );

	return( false );
}


void  CbxDB::
LoadCtlPlaneDegForRotSub( const ulong edgeNo, const short beginEndNo, CbxCore::CtlPlaneDegForRotSub sub[ 2 ] )
{
	ulong   vertexNo;
	short   i, indexNo;

	for( i=0; i<2; i++ ) {
		indexNo  = ( beginEndNo + i ) % 2;
		vertexNo = edgeAry[ edgeNo ].vertex[ indexNo ].vertexNo;

		CbxMath::Copy3( vertexAry[ vertexNo ].vertexPos, sub[ i ].vertexPos );
		CbxMath::CopyAxis( vertexAry[ vertexNo ].axisVec, sub[ i ].axisVec );
		CbxMath::Copy2( edgeAry[ edgeNo ].vertex[ indexNo ].handleVec, sub[ i ].handleVec );
	}
}


bool  CbxDB::
UnlockCtlPlaneDegForRot( const CbxCore::CtlPlaneDegForRotOut* ctl )
{
	if( lockFlag == false )
		return( true );
	lockFlag = false;

	edgeAry[ lockEdgeNo ].vertex[ lockBeginEndNo ].ctlPlaneDeg = ctl->ctlPlaneDeg;

	return( false );
}


bool  CbxDB::
LoadCtlVecForTrans( const ulong edgeNo, const short beginEndNo, const short minusPlusNo,
	CbxCore::CtlVecForTrans* ctl )
{
	bool  result;

	result = LoadCtlVecForTransSub( edgeNo, beginEndNo, ctl->in.sub );
	if( result == true )
		return( true );

	ctl->in.ctlPlaneDeg = edgeAry[ edgeNo ].vertex[ beginEndNo ].ctlPlaneDeg;
	CbxMath::Copy2( edgeAry[ edgeNo ].vertex[ beginEndNo ].ctlVec[ minusPlusNo ], ctl->out.ctlVec );

	return( false );
}


bool  CbxDB::
LockCtlVecForTrans( const ulong edgeNo, const short beginEndNo, const short minusPlusNo,
	CbxCore::CtlVecForTrans* ctl )
{
	if( lockFlag == true )
		return( true );
	lockFlag        = true;
	lockEdgeNo      = edgeNo;
	lockBeginEndNo  = beginEndNo;
	lockMinusPlusNo = minusPlusNo;

	return( LoadCtlVecForTrans( edgeNo, beginEndNo, minusPlusNo, ctl ) );
}


bool  CbxDB::
LoadCtlVecForTransSub( const ulong edgeNo, const short beginEndNo, CbxCore::CtlVecForTransSub sub[ 2 ] )
{
	ulong   vertexNo;
	short   i, indexNo, j;
	double  wing[ 2 ][ 2 ], lenRadialVec[ 2 ];
	ushort  numRadial;
	bool    loopFlag, result;

	for( i=0; i<2; i++ ) {
		indexNo  = ( beginEndNo + i ) % 2;
		vertexNo = edgeAry[ edgeNo ].vertex[ indexNo ].vertexNo;

		CbxMath::Copy3( vertexAry[ vertexNo ].vertexPos, sub[ i ].vertexPos );
		CbxMath::CopyAxis( vertexAry[ vertexNo ].axisVec, sub[ i ].axisVec );
		CbxMath::Copy2( edgeAry[ edgeNo ].vertex[ indexNo ].handleVec, sub[ i ].handleVec );

		numRadial = vertexAry[ vertexNo ].numRadial;
		loopFlag  = vertexAry[ vertexNo ].loopFlag;

		result = LoadCtlVecForTransSubDecWing( edgeNo, indexNo, numRadial, loopFlag, wing[ 0 ], &lenRadialVec[ 0 ] );
		if( result == true )
			return( true );
		result = LoadCtlVecForTransSubIncWing( edgeNo, indexNo, numRadial, loopFlag, wing[ 1 ], &lenRadialVec[ 1 ] );
		if( result == true )
			return( true );

		for( j=0; j<2; j++ ) {
			CbxMath::Copy2( wing[ j ], sub[ i ].wingHandle[ j ] );
			sub[ i ].lenRadialVec[ j ] = lenRadialVec[ j ];
		}
	}

	return( false );
}


bool  CbxDB::
LoadCtlVecForTransSubDecWing( const ulong edgeNo, const short beginEndNo, const ushort numRadial,
	const bool loopFlag, double wing[ 2 ], double* lenRadialVec )
{
	ulong   vertexNo, edgeNo2, vertexNo2;
	ushort  radialNo, radialDec, radialInc;
	short   beginEndNo2;
	bool    result;

	vertexNo  = edgeAry[ edgeNo ].vertex[ beginEndNo ].vertexNo;
	radialNo  = edgeAry[ edgeNo ].vertex[ beginEndNo ].radialNo;
	radialDec = ( radialNo + numRadial - 1 ) % numRadial;
	radialInc = ( radialNo + 1 ) % numRadial;

	if( radialNo > 0 || loopFlag == true ) {
		result = VertexRadialToEdge( vertexNo, radialDec, &edgeNo2, &beginEndNo2 );
		if( result == true )
			return( true );
		CbxMath::Copy2( edgeAry[ edgeNo2 ].vertex[ beginEndNo2 ].handleVec, wing );
	}
	else {
		result = VertexRadialToEdge( vertexNo, radialInc, &edgeNo2, &beginEndNo2 );
		if( result == true )
			return( true );
		CbxMath::Copy2( edgeAry[ edgeNo2 ].vertex[ beginEndNo2 ].handleVec, wing );
		wing[ 0 ] = CbxMath::NegateDeg( wing[ 0 ] );
	}

	vertexNo2     = edgeAry[ edgeNo2 ].vertex[ 1-beginEndNo2 ].vertexNo;
	*lenRadialVec = CbxMath::Distance3( vertexAry[ vertexNo ].vertexPos, vertexAry[ vertexNo2 ].vertexPos );

	return( false );
}


bool  CbxDB::
LoadCtlVecForTransSubIncWing( const ulong edgeNo, const short beginEndNo, const ushort numRadial,
	const bool loopFlag, double wing[ 2 ], double* lenRadialVec )
{
	ulong   vertexNo, edgeNo2, vertexNo2;
	ushort  radialNo, radialDec, radialInc;
	short   beginEndNo2;
	bool    result;

	vertexNo  = edgeAry[ edgeNo ].vertex[ beginEndNo ].vertexNo;
	radialNo  = edgeAry[ edgeNo ].vertex[ beginEndNo ].radialNo;
	radialDec = ( radialNo + numRadial - 1 ) % numRadial;
	radialInc = ( radialNo + 1 ) % numRadial;

	if( radialNo < ( numRadial-1 ) || loopFlag == true ) {
		result = VertexRadialToEdge( vertexNo, radialInc, &edgeNo2, &beginEndNo2 );
		if( result == true )
			return( true );
		CbxMath::Copy2( edgeAry[ edgeNo2 ].vertex[ beginEndNo2 ].handleVec, wing );
	}
	else {
		result = VertexRadialToEdge( vertexNo, radialDec, &edgeNo2, &beginEndNo2 );
		if( result == true )
			return( true );
		CbxMath::Copy2( edgeAry[ edgeNo2 ].vertex[ beginEndNo2 ].handleVec, wing );
		wing[ 0 ] = CbxMath::NegateDeg( wing[ 0 ] );
	}

	vertexNo2     = edgeAry[ edgeNo2 ].vertex[ 1-beginEndNo2 ].vertexNo;
	*lenRadialVec = CbxMath::Distance3( vertexAry[ vertexNo ].vertexPos, vertexAry[ vertexNo2 ].vertexPos );

	return( false );
}


bool  CbxDB::
UnlockCtlVecForTrans( const CbxCore::CtlVecForTransOut* ctl )
{
	if( lockFlag == false )
		return( true );
	lockFlag = false;

	CbxMath::Copy2( ctl->ctlVec, edgeAry[ lockEdgeNo ].vertex[ lockBeginEndNo ].ctlVec[ lockMinusPlusNo ] );

	return( false );
}


bool  CbxDB::
LoadCtlVecForTransPlus( const ulong edgeNo, const short beginEndNo, const short minusPlusNo,
	CbxCore::CtlVecForTransPlus* ctl )
{
	bool  result;

	result = LoadCtlVecForTransSub( edgeNo, beginEndNo, ctl->in.sub );
	if( result == true )
		return( true );

	ctl->in.minusPlusNo = minusPlusNo;

	ctl->out.ctlPlaneDeg = edgeAry[ edgeNo ].vertex[ beginEndNo ].ctlPlaneDeg;
	CbxMath::Copy2( edgeAry[ edgeNo ].vertex[ beginEndNo ].ctlVec[ minusPlusNo ], ctl->out.ctlVec );

	return( false );
}


bool  CbxDB::
LockCtlVecForTransPlus( const ulong edgeNo, const short beginEndNo, const short minusPlusNo,
	CbxCore::CtlVecForTransPlus* ctl )
{
	if( lockFlag == true )
		return( true );
	lockFlag        = true;
	lockEdgeNo      = edgeNo;
	lockBeginEndNo  = beginEndNo;
	lockMinusPlusNo = minusPlusNo;

	return( LoadCtlVecForTransPlus( edgeNo, beginEndNo, minusPlusNo, ctl ) );
}


bool  CbxDB::
UnlockCtlVecForTransPlus( const CbxCore::CtlVecForTransPlusOut* ctl )
{
	if( lockFlag == false )
		return( true );
	lockFlag = false;

	edgeAry[ lockEdgeNo ].vertex[ lockBeginEndNo ].ctlPlaneDeg = ctl->ctlPlaneDeg;
	CbxMath::Copy2( ctl->ctlVec, edgeAry[ lockEdgeNo ].vertex[ lockBeginEndNo ].ctlVec[ lockMinusPlusNo ] );

	return( false );
}


bool  CbxDB::
GetSurfaceVertex( const ulong edgeNo, const short beginEndNo, short* chkSurfExist, short* numAngle,
	EdgeVertexRadial surfaceVertex[ 4 ] )
{
	short  cnt, nextBeginEndNo;
	ulong  nextEdgeNo;
	bool   result, radialEdge;

	*chkSurfExist = surfNonExist;

	surfaceVertex[ 0 ].edgeNo     = edgeNo;
	surfaceVertex[ 0 ].beginEndNo = beginEndNo;

	cnt = 0;
	for(;;) {
		result = NextEdgeNo( surfaceVertex[ cnt ].edgeNo, surfaceVertex[ cnt ].beginEndNo, &radialEdge,
					 &nextEdgeNo, &nextBeginEndNo );
		if( result == true )
			return( true );
		else if( radialEdge == true ) {
			if( cnt == 0 ) {
				*chkSurfExist = surfSchRadialEdge;
				return( false );
			}
			return( true );
		}

		cnt++;
		
		if( nextEdgeNo == edgeNo && nextBeginEndNo == beginEndNo ) {
			if( cnt == 3 || cnt == 4 )
				*chkSurfExist = surfExist;
		}

		if( *chkSurfExist == surfExist || cnt == 4 )
			break;

		surfaceVertex[ cnt ].edgeNo     = nextEdgeNo;
		surfaceVertex[ cnt ].beginEndNo = nextBeginEndNo;
	}

	if( *chkSurfExist == surfExist )
		*numAngle = cnt;

	return( false );
}


bool  CbxDB::
NextEdgeNo( const ulong orgEdgeNo, const short orgBeginEndNo, bool* radialEdge, ulong* nextEdgeNo, short* nextBeginEndNo )
{
	ulong   vertexNo;
	ushort  radialNo, radialDec, numRadial;
	bool    result, loopFlag;

	*radialEdge = false;

	vertexNo  = edgeAry[ orgEdgeNo ].vertex[ 1-orgBeginEndNo ].vertexNo;
	radialNo  = edgeAry[ orgEdgeNo ].vertex[ 1-orgBeginEndNo ].radialNo;

	numRadial = vertexAry[ vertexNo ].numRadial;
	loopFlag  = vertexAry[ vertexNo ].loopFlag;
	radialDec = ( radialNo + numRadial - 1 ) % numRadial;

	if( radialNo > 0 || loopFlag == true ) {
		result = VertexRadialToEdge( vertexNo, radialDec, nextEdgeNo, nextBeginEndNo );
		if( result == true )
			return( true );
	}
	else {
		*radialEdge = true;
		return( false );
	}

	return( false );
}


bool  CbxDB::
LockCtlVecForDefault( const ulong edgeNo, const short beginEndNo, CbxCore::CtlVecForDefault* surface )
{
	if( lockFlag == true )
		return( true );
	lockFlag       = true;
	lockEdgeNo     = edgeNo;
	lockBeginEndNo = beginEndNo;

	return( LoadCtlVecForDefault( edgeNo, beginEndNo, surface ) );
}


bool  CbxDB::
LoadCtlVecForDefault( const ulong edgeNo, const short beginEndNo, CbxCore::CtlVecForDefault* surface )
{
	bool              result;
	short             numAngle, i, chkSurfExist;
	EdgeVertexRadial  surfaceVertex[ 4 ];

	result = GetSurfaceVertex( edgeNo, beginEndNo, &chkSurfExist, &numAngle, surfaceVertex );
	if( result == true || chkSurfExist != surfExist )
		return( true );

	surface->in.numVertex = numAngle;

	for( i=0; i<numAngle; i++ ) {
		result = LoadCtlVecForDefaultVertex( &surfaceVertex[ i ], &surface->in.vertex[ i ], &surface->out.vertex[ i ] );
		if( result == true )
			return( true );
	}
	
	return( false );
}


bool  CbxDB::
LoadCtlVecForDefaultVertex( const EdgeVertexRadial* surfaceVertex, CbxCore::CtlVecForDefaultInVertex* in,
	 CbxCore::CtlVecForDefaultOutVertex* out )
{
	ulong   edgeNo0, vertexNo, edgeNo1;
	ushort  radialNo, radialInc, numRadial;
	short   beginEndNo0, beginEndNo1;
	bool    result, loopFlag;

	edgeNo0     = surfaceVertex->edgeNo;
	beginEndNo0 = surfaceVertex->beginEndNo;
	vertexNo    = edgeAry[ edgeNo0 ].vertex[ beginEndNo0 ].vertexNo;
	radialNo    = edgeAry[ edgeNo0 ].vertex[ beginEndNo0 ].radialNo;

	numRadial = vertexAry[ vertexNo ].numRadial;
	loopFlag  = vertexAry[ vertexNo ].loopFlag;
	radialInc = ( radialNo + 1 ) % numRadial;

	if( radialNo < ( numRadial-1 ) || loopFlag == true ) {
		result = VertexRadialToEdge( vertexNo, radialInc, &edgeNo1, &beginEndNo1 );
		if( result == true )
			return( true );
	}
	else
		return( true );

	CbxMath::Copy3( vertexAry[ vertexNo ].vertexPos, in->vertexPos );
	CbxMath::CopyAxis( vertexAry[ vertexNo ].axisVec, in->axisVec );

	result = LoadCtlVecForDefaultEdge( edgeNo1, beginEndNo1, 0, numRadial, loopFlag, &in->edge[ 0 ], &out->edge[ 0 ] );
	if( result == true )
		return( true );
	result = LoadCtlVecForDefaultEdge( edgeNo0, beginEndNo0, 1, numRadial, loopFlag, &in->edge[ 1 ], &out->edge[ 1 ] );
	if( result == true )
		return( true );

	return( false );
}


bool  CbxDB::
LoadCtlVecForDefaultEdge( const ulong edgeNo, const short beginEndNo, const short minusPlusNo, const ushort numRadial,
	const bool loopFlag, CbxCore::CtlVecForDefaultInEdge* in, CbxCore::CtlVecForDefaultOutEdge* out )
{
	bool  result;

	CbxMath::Copy2( edgeAry[ edgeNo ].vertex[ beginEndNo ].handleVec, in->handleVec );

	if( minusPlusNo == 0 ) {
		result = LoadCtlVecForTransSubIncWing( edgeNo, beginEndNo, numRadial, loopFlag,
					 in->wingHandle, &in->lenRadialVec );
		if( result == true )
			return( true );
	}
	else {
		result = LoadCtlVecForTransSubDecWing( edgeNo, beginEndNo, numRadial, loopFlag,
					 in->wingHandle, &in->lenRadialVec );
		if( result == true )
			return( true );
	}

	in->ctlPlaneDeg = edgeAry[ edgeNo ].vertex[ beginEndNo ].ctlPlaneDeg;
	CbxMath::Copy2( edgeAry[ edgeNo ].vertex[ beginEndNo ].ctlVec[ minusPlusNo ],   out->ctlVec    );
	CbxMath::Copy2( edgeAry[ edgeNo ].vertex[ beginEndNo ].ctlVec[ 1-minusPlusNo ], in->antiCtlVec );

	return( false );
}


bool  CbxDB::
UnlockCtlVecForDefault( CbxCore::CtlVecForDefaultOut* surface )
{
	if( lockFlag == false )
		return( true );
	lockFlag = false;

	return( SaveCtlVecForDefault( lockEdgeNo, lockBeginEndNo, surface ) );
}


bool  CbxDB::
SaveCtlVecForDefault( const ulong edgeNo, const short beginEndNo, CbxCore::CtlVecForDefaultOut* surface )
{
	bool              result;
	short             numAngle, i, chkSurfExist;
	EdgeVertexRadial  surfaceVertex[ 4 ];

	result = GetSurfaceVertex( edgeNo, beginEndNo, &chkSurfExist, &numAngle, surfaceVertex );
	if( result == true || chkSurfExist != surfExist )
		return( true );

	for( i=0; i<numAngle; i++ ) {
		result = SaveCtlVecForDefaultVertex( &surfaceVertex[ i ], &surface->vertex[ i ] );
		if( result == true )
			return( true );
	}
	
	return( false );
}


bool  CbxDB::
SaveCtlVecForDefaultVertex( const EdgeVertexRadial* surfaceVertex, CbxCore::CtlVecForDefaultOutVertex* vertex )
{
	ulong   edgeNo0, vertexNo, edgeNo1;
	ushort  radialNo, radialInc, numRadial;
	short   beginEndNo0, beginEndNo1;
	bool    result, loopFlag;

	edgeNo0     = surfaceVertex->edgeNo;
	beginEndNo0 = surfaceVertex->beginEndNo;
	vertexNo    = edgeAry[ edgeNo0 ].vertex[ beginEndNo0 ].vertexNo;
	radialNo    = edgeAry[ edgeNo0 ].vertex[ beginEndNo0 ].radialNo;

	numRadial = vertexAry[ vertexNo ].numRadial;
	loopFlag  = vertexAry[ vertexNo ].loopFlag;
	radialInc = ( radialNo + 1 ) % numRadial;

	if( radialNo < ( numRadial-1 ) || loopFlag == true ) {
		result = VertexRadialToEdge( vertexNo, radialInc, &edgeNo1, &beginEndNo1 );
		if( result == true )
			return( true );
	}
	else
		return( true );

	SaveCtlVecForDefaultEdge( edgeNo1, beginEndNo1, 0, &vertex->edge[ 0 ] );
	SaveCtlVecForDefaultEdge( edgeNo0, beginEndNo0, 1, &vertex->edge[ 1 ] );

	return( false );
}


void  CbxDB::
SaveCtlVecForDefaultEdge( const ulong edgeNo, const short beginEndNo, const short minusPlusNo,
	CbxCore::CtlVecForDefaultOutEdge* edge )
{
	CbxMath::Copy2( edge->ctlVec, edgeAry[ edgeNo ].vertex[ beginEndNo ].ctlVec[ minusPlusNo ] );
}

		
bool  CbxDB::
LoadSurfaceForPick( const ulong edgeNo, const short beginEndNo, CbxCore::SurfaceForPick* surface )
{
	bool              result;
	short             numAngle, i, chkSurfExist;
	EdgeVertexRadial  surfaceVertex[ 4 ];

	result = GetSurfaceVertex( edgeNo, beginEndNo, &chkSurfExist, &numAngle, surfaceVertex );
	if( result == true || chkSurfExist != surfExist )
		return( true );

	surface->numVertex = numAngle;

	for( i=0; i<numAngle; i++ ) {
		result = LoadSurfaceForPickVertex( &surfaceVertex[ i ], &surface->vertex[ i ] );
		if( result == true )
			return( true );
	}
	
	return( false );
}


bool  CbxDB::
LoadSurfaceForPickVertex( const EdgeVertexRadial* surfaceVertex, CbxCore::SurfaceForPickVertex* vertex )
{
	ulong   edgeNo0, vertexNo, edgeNo1;
	ushort  radialNo, radialInc, numRadial;
	short   beginEndNo0, beginEndNo1;
	bool    result, loopFlag;

	edgeNo0     = surfaceVertex->edgeNo;
	beginEndNo0 = surfaceVertex->beginEndNo;
	vertexNo    = edgeAry[ edgeNo0 ].vertex[ beginEndNo0 ].vertexNo;
	radialNo    = edgeAry[ edgeNo0 ].vertex[ beginEndNo0 ].radialNo;

	numRadial = vertexAry[ vertexNo ].numRadial;
	loopFlag  = vertexAry[ vertexNo ].loopFlag;
	radialInc = ( radialNo + 1 ) % numRadial;

	if( radialNo < ( numRadial-1 ) || loopFlag == true ) {
		result = VertexRadialToEdge( vertexNo, radialInc, &edgeNo1, &beginEndNo1 );
		if( result == true )
			return( true );
	}
	else
		return( true );

	CbxMath::Copy3( vertexAry[ vertexNo ].vertexPos, vertex->vertexPos );
	CbxMath::CopyAxis( vertexAry[ vertexNo ].axisVec, vertex->axisVec );

	result = LoadSurfaceForPickEdge( edgeNo1, beginEndNo1, 0, numRadial, loopFlag, &vertex->edge[ 0 ] );
	if( result == true )
		return( true );
	result = LoadSurfaceForPickEdge( edgeNo0, beginEndNo0, 1, numRadial, loopFlag, &vertex->edge[ 1 ] );
	if( result == true )
		return( true );

	return( false );
}


bool  CbxDB::
LoadSurfaceForPickEdge( const ulong edgeNo, const short beginEndNo, const short minusPlusNo, const ushort numRadial,
	const bool loopFlag, CbxCore::SurfaceForPickEdge* edge )
{
	bool  result;

	CbxMath::Copy2( edgeAry[ edgeNo ].vertex[ beginEndNo ].handleVec, edge->handleVec );

	if( minusPlusNo == 0 ) {
		result = LoadCtlVecForTransSubIncWing( edgeNo, beginEndNo, numRadial, loopFlag,
					 edge->wingHandle, &edge->lenRadialVec );
		if( result == true )
			return( true );
	}
	else {
		result = LoadCtlVecForTransSubDecWing( edgeNo, beginEndNo, numRadial, loopFlag,
					 edge->wingHandle, &edge->lenRadialVec );
		if( result == true )
			return( true );
	}

	edge->ctlPlaneDeg = edgeAry[ edgeNo ].vertex[ beginEndNo ].ctlPlaneDeg;
	CbxMath::Copy2( edgeAry[ edgeNo ].vertex[ beginEndNo ].ctlVec[ minusPlusNo ],   edge->ctlVec     );
	CbxMath::Copy2( edgeAry[ edgeNo ].vertex[ beginEndNo ].ctlVec[ 1-minusPlusNo ], edge->antiCtlVec );

	return( false );
}
