/* bxDB.h */

#if !defined( BX_DB_H )
#define  BX_DB_H

#include "../bxType.h"
#include "../Core/bxCore.h"


class CbxDB {
	public:
		enum {
			minNumVertex = 3,
			minNumEdge   = 3,
			minNumRadial = 2
		};

		enum {
			surfExist,
			surfSchRadialEdge,
			surfNonExist
		};

	public:
		struct VertexRec {
			double  vertexPos[ 3 ];
			ushort  numRadial;
			bool    loopFlag;
			double  axisVec[ 2 ][ 3 ];
		};

		struct EdgeVertex {
			ulong   vertexNo;
			ushort  radialNo;
			double  handleVec[ 2 ];
			double  ctlPlaneDeg;
			double  ctlVec[ 2 ][ 2 ];
		};

		struct EdgeRec {
			EdgeVertex  vertex[ 2 ];
		};

		struct EdgeForPickSurface {
			bool  alreadyPick[ 2 ];
		};

		struct EdgeVertexRadial {
			ulong  edgeNo;
			short  beginEndNo;
		};

	public:
		static bool
		AllocData( const ulong numVertex, const ulong numEdge );
		static bool
		AllocEdgeForPickSurface( void );
		static void
		FreeData( void );
		static void
		FreeEdgeForPickSurface( void );

		static ulong
		NumVertexAry( void )	{  return( numVertexAry );  }
		static ulong
		NumEdgeAry( void )		{  return( numEdgeAry );    }

		static void
		InitilizeEdgeRec( EdgeRec* edge );
		static bool
		CancelLock( void );
		static void
		LoadVertexRec( const ulong vertexNo, VertexRec* vertex );
		static bool
		LockVertexRec( const ulong vertexNo, VertexRec* vertex );
		static bool
		UnlockVertexRec( const VertexRec* vertex );
		static void
		LoadEdgeRec( const ulong edgeNo, EdgeRec* edge );
		static bool
		LockEdgeRec( const ulong edgeNo, EdgeRec* edge );
		static bool
		UnlockEdgeRec( const EdgeRec* edge );
		static bool
		GetEdgeForPickSurface( const ulong edgeNo, const short beginEndNo );
		static void
		SetEdgeForPickSurface( const ulong edgeNo, const short beginEndNo );

		static bool
		GetSurfaceVertex( const ulong edgeNo, const short beginEndNo, short* chkSurfExist, short* numAngle,
			EdgeVertexRadial surfaceVertex[ 4 ] );

		static bool
		VertexRadialToEdge( const ulong vertexNo, const ushort radialNo, ulong* edgeNo, short* beginEndNo );

	protected:
		static void
		NumRadial( const ulong vertexNo, ushort* numRadial );
		static bool
		LockAxisForDefault( const ulong vertexNo, const CbxCore::AxisForDefaultSub buf[/*numRadial*/],
			CbxCore::AxisForDefault* axis );
		static bool
		UnlockAxisForDefault( const CbxCore::AxisForDefaultOut* axis );
		static void
		LoadHandleForTrans( const ulong edgeNo, const short beginEndNo, CbxCore::HandleForTrans* handle );
		static bool
		LockHandleForTrans( const ulong edgeNo, const short beginEndNo, CbxCore::HandleForTrans* handle );
		static bool
		UnlockHandleForTrans( const CbxCore::HandleForTransOut* handle );
		static bool
		LoadHandleForTransPlus( const ulong edgeNo, const short beginEndNo, CbxCore::HandleForTransPlus* handle );
		static bool
		LockHandleForTransPlus( const ulong edgeNo, const short beginEndNo, CbxCore::HandleForTransPlus* handle );
		static bool
		UnlockHandleForTransPlus( const CbxCore::HandleForTransPlusOut* handle );
		static void
		LoadVertexForRot( const ulong vertexNo, CbxCore::VertexForRot* rot );
		static bool
		LockVertexForRot( const ulong vertexNo, CbxCore::VertexForRot* rot );
		static bool
		UnlockVertexForRot( const CbxCore::VertexForRotOut* rot );
		static void
		LoadCtlPlaneDegForRot( const ulong edgeNo, const short beginEndNo, CbxCore::CtlPlaneDegForRot* ctl );
		static bool
		LockCtlPlaneDegForRot( const ulong edgeNo, const short beginEndNo, CbxCore::CtlPlaneDegForRot* ctl );
		static bool
		UnlockCtlPlaneDegForRot( const CbxCore::CtlPlaneDegForRotOut* ctl );
		static bool
		LoadCtlVecForTrans( const ulong edgeNo, const short beginEndNo, const short minusPlusNo,
			CbxCore::CtlVecForTrans* ctl );
		static bool
		LockCtlVecForTrans( const ulong edgeNo, const short beginEndNo, const short minusPlusNo,
			CbxCore::CtlVecForTrans* ctl );
		static bool
		UnlockCtlVecForTrans( const CbxCore::CtlVecForTransOut* ctl );
		static bool
		LoadCtlVecForTransPlus( const ulong edgeNo, const short beginEndNo, const short minusPlusNo,
			CbxCore::CtlVecForTransPlus* ctl );
		static bool
		LockCtlVecForTransPlus( const ulong edgeNo, const short beginEndNo, const short minusPlusNo,
			CbxCore::CtlVecForTransPlus* ctl );
		static bool
		UnlockCtlVecForTransPlus( const CbxCore::CtlVecForTransPlusOut* ctl );
		static bool
		LockCtlVecForDefault( const ulong edgeNo, const short beginEndNo, CbxCore::CtlVecForDefault* surface );
		static bool
		UnlockCtlVecForDefault( CbxCore::CtlVecForDefaultOut* surface );
		static bool
		LoadSurfaceForPick( const ulong edgeNo, const short beginEndNo, CbxCore::SurfaceForPick* surface );

	private:
		static bool
		AllocVertex( void );
		static bool
		AllocEdge( void );
		static void
		FreeVertex( void );
		static void
		FreeEdge( void );

		static bool
		LoadAxisForDefault( const ulong vertexNo, const CbxCore::AxisForDefaultSub buf[/*numRadial*/],
			CbxCore::AxisForDefault* axis );
		static bool
		LoadHandleForTransPlusSub( const ulong edgeNo, const short beginEndNo, CbxCore::HandleForTransPlusIn* handle );
		static void
		LoadCtlPlaneDegForRotSub( const ulong edgeNo, const short beginEndNo, CbxCore::CtlPlaneDegForRotSub sub[ 2 ] );
		static bool
		LoadCtlVecForTransSub( const ulong edgeNo, const short beginEndNo, CbxCore::CtlVecForTransSub sub[ 2 ] );
		static bool
		LoadCtlVecForTransSubDecWing( const ulong edgeNo, const short beginEndNo, const ushort numRadial,
			const bool loopFlag, double wing[ 2 ], double* lenRadialVec );
		static bool
		LoadCtlVecForTransSubIncWing( const ulong edgeNo, const short beginEndNo, const ushort numRadial,
			const bool loopFlag, double wing[ 2 ], double* lenRadialVec );
		static bool
		NextEdgeNo( const ulong orgEdgeNo, const short orgBeginEndNo, bool* radialEdge, ulong* nextEdgeNo,
			short* nextBeginEndNo );
		static bool
		LoadCtlVecForDefault( const ulong edgeNo, const short beginEndNo, CbxCore::CtlVecForDefault* surface );
		static bool
		LoadCtlVecForDefaultVertex( const EdgeVertexRadial* surfaceVertex, CbxCore::CtlVecForDefaultInVertex* in,
			 CbxCore::CtlVecForDefaultOutVertex* out );
		static bool
		LoadCtlVecForDefaultEdge( const ulong edgeNo, const short beginEndNo, const short minusPlusNo, const ushort numRadial,
			const bool loopFlag, CbxCore::CtlVecForDefaultInEdge* in, CbxCore::CtlVecForDefaultOutEdge* out );
		static bool
		SaveCtlVecForDefault( const ulong edgeNo, const short beginEndNo, CbxCore::CtlVecForDefaultOut* surface );
		static bool
		SaveCtlVecForDefaultVertex( const EdgeVertexRadial* surfaceVertex, CbxCore::CtlVecForDefaultOutVertex* vertex );
		static void
		SaveCtlVecForDefaultEdge( const ulong edgeNo, const short beginEndNo, const short minusPlusNo,
			CbxCore::CtlVecForDefaultOutEdge* edge );
		static bool
		LoadSurfaceForPickVertex( const EdgeVertexRadial* surfaceVertex, CbxCore::SurfaceForPickVertex* vertex );
		static bool
		LoadSurfaceForPickEdge( const ulong edgeNo, const short beginEndNo, const short minusPlusNo, const ushort numRadial,
			const bool loopFlag, CbxCore::SurfaceForPickEdge* edge );

	private:
		static ulong                numVertexAry;
		static ulong                numEdgeAry;
		static VertexRec*           vertexAry;
		static EdgeRec*             edgeAry;
		static EdgeForPickSurface*  edgePick;

		static const double  base_ctlPlaneDeg;
		static const double  base_ctlVec0[ 2 ];
		static const double  base_ctlVec1[ 2 ];

	private:
		static bool   lockFlag;
		static ulong  lockVertexNo;
		static ulong  lockEdgeNo;
		static short  lockBeginEndNo;
		static short  lockMinusPlusNo;
};


#endif
