/* bxMath.cpp */

#include <math.h>
#include "bxMath.h"


const double  CbxMath::pi =
	3.1415926535897932;

const double  CbxMath::zero2[ 2 ] = {
	0.0, 0.0
};
const double  CbxMath::zero3[ 3 ] = {
	0.0, 0.0, 0.0
};

const double  CbxMath::nearZero =
	1.0E-6;
const double  CbxMath::nearZero2[ 2 ] = {
	1.0E-6, 1.0E-6
};
const double  CbxMath::nearZero3[ 3 ] = {
	1.0E-6, 1.0E-6, 1.0E-6
};
const double  CbxMath::minusNearZero2[ 2 ] = {
	-1.0E-6, -1.0E-6
};
const double  CbxMath::minusNearZero3[ 3 ] = {
	-1.0E-6, -1.0E-6, -1.0E-6
};

const double  CbxMath::identity3x3[ 3 ][ 3 ] = {
	{ 1.0, 0.0, 0.0 },
	{ 0.0, 1.0, 0.0 },
	{ 0.0, 0.0, 1.0 }
};
const double  CbxMath::identity4x4[ 4 ][ 4 ] = {
	{ 1.0, 0.0, 0.0, 0.0 },
	{ 0.0, 1.0, 0.0, 0.0 },
	{ 0.0, 0.0, 1.0, 0.0 },
	{ 0.0, 0.0, 0.0, 1.0 }
};

const double  CbxMath::worldAxis[ 3/*axis*/ ][ 3/*element*/ ] = {
	{ 1.0, 0.0, 0.0 },
	{ 0.0, 1.0, 0.0 },
	{ 0.0, 0.0, 1.0 }
};




double*  CbxMath::
Set2( const double a, const double b, double z[ 2 ] )
{
	z[ 0 ] = a;		z[ 1 ] = b;

	return( z );
}

double*  CbxMath::
Set3( const double a, const double b, const double c, double z[ 3 ] )
{
	z[ 0 ] = a;		z[ 1 ] = b;		z[ 2 ] = c;

	return( z );
}

double*  CbxMath::
Copy2( const double a[ 2 ], double z[ 2 ] )
{
	z[ 0 ] = a[ 0 ];
	z[ 1 ] = a[ 1 ];

	return( z );
}

double*  CbxMath::
Copy3( const double a[ 3 ], double z[ 3 ] )
{
	z[ 0 ] = a[ 0 ];
	z[ 1 ] = a[ 1 ];
	z[ 2 ] = a[ 2 ];

	return( z );
}

double*  CbxMath::
Add2( const double a[ 2 ], const double b[ 2 ], double z[ 2 ] )
{
	z[ 0 ] = a[ 0 ] + b[ 0 ];
	z[ 1 ] = a[ 1 ] + b[ 1 ];

	return( z );
}

double*  CbxMath::
Add3( const double a[ 3 ], const double b[ 3 ], double z[ 3 ] )
{
	z[ 0 ] = a[ 0 ] + b[ 0 ];
	z[ 1 ] = a[ 1 ] + b[ 1 ];
	z[ 2 ] = a[ 2 ] + b[ 2 ];

	return( z );
}

double*  CbxMath::
Subtract2( const double a[ 2 ], const double b[ 2 ], double z[ 2 ] )
{
	z[ 0 ] = a[ 0 ] - b[ 0 ];
	z[ 1 ] = a[ 1 ] - b[ 1 ];
	return( z );
}

double*  CbxMath::
Subtract3( const double a[ 3 ], const double b[ 3 ], double z[ 3 ] )
{
	z[ 0 ] = a[ 0 ] - b[ 0 ];
	z[ 1 ] = a[ 1 ] - b[ 1 ];
	z[ 2 ] = a[ 2 ] - b[ 2 ];
	return( z );
}

double  CbxMath::
Distance2( const double a[ 2 ], const double b[ 2 ] )
{
	double  c[ 2 ], d;

	c[ 0 ] = a[ 0 ] - b[ 0 ];
	c[ 1 ] = a[ 1 ] - b[ 1 ];

	d = sqrt( c[ 0 ] * c[ 0 ] + c[ 1 ] * c[ 1 ] );

	return( d );
}

double  CbxMath::
Distance3( const double a[ 3 ], const double b[ 3 ] )
{
	double  c[ 3 ], d;

	c[ 0 ] = a[ 0 ] - b[ 0 ];
	c[ 1 ] = a[ 1 ] - b[ 1 ];
	c[ 2 ] = a[ 2 ] - b[ 2 ];

	d = sqrt( c[ 0 ] * c[ 0 ] + c[ 1 ] * c[ 1 ] + c[ 2 ] * c[ 2 ] );

	return( d );
}

double  CbxMath::
DistanceSquared2( const double a[ 2 ], const double b[ 2 ] )
{
	double  c[ 2 ], d;

	c[ 0 ] = a[ 0 ] - b[ 0 ];
	c[ 1 ] = a[ 1 ] - b[ 1 ];

	d = c[ 0 ] * c[ 0 ] + c[ 1 ] * c[ 1 ];

	return( d );
}

double  CbxMath::
DistanceSquared3( const double a[ 3 ], const double b[ 3 ] )
{
	double  c[ 3 ], d;

	c[ 0 ] = a[ 0 ] - b[ 0 ];
	c[ 1 ] = a[ 1 ] - b[ 1 ];
	c[ 2 ] = a[ 2 ] - b[ 2 ];

	d = c[ 0 ] * c[ 0 ] + c[ 1 ] * c[ 1 ] + c[ 2 ] * c[ 2 ];

	return( d );
}

double*  CbxMath::
Ratio2( const double a[ 2 ], const double b[ 2 ], const double t, double z[ 2 ] )
{
	z[ 0 ] = ( 1.0 - t ) * a[ 0 ] + t * b[ 0 ];
	z[ 1 ] = ( 1.0 - t ) * a[ 1 ] + t * b[ 1 ];

	return( z );
}

double*  CbxMath::
Ratio3( const double a[ 3 ], const double b[ 3 ], const double t, double z[ 3 ] )
{
	z[ 0 ] = ( 1.0 - t ) * a[ 0 ] + t * b[ 0 ];
	z[ 1 ] = ( 1.0 - t ) * a[ 1 ] + t * b[ 1 ];
	z[ 2 ] = ( 1.0 - t ) * a[ 2 ] + t * b[ 2 ];

	return( z );
}

double*  CbxMath::
Scale2( const double a[ 2 ], const double scale, double z[ 2 ] )
{
	z[ 0 ] = scale * a[ 0 ];
	z[ 1 ] = scale * a[ 1 ];

	return( z );
}

double*  CbxMath::
Scale3( const double a[ 3 ], const double scale, double z[ 3 ] )
{
	z[ 0 ] = scale * a[ 0 ];
	z[ 1 ] = scale * a[ 1 ];
	z[ 2 ] = scale * a[ 2 ];

	return( z );
}

double  CbxMath::
Length2( const double a[ 2 ] )
{
	double  b;

	b = sqrt( a[ 0 ] * a[ 0 ] + a[ 1 ] * a[ 1 ] );

	return( b );
}

double  CbxMath::
Length3( const double a[ 3 ] )
{
	double  b;

	b = sqrt( a[ 0 ] * a[ 0 ] + a[ 1 ] * a[ 1 ] + a[ 2 ] * a[ 2 ] );

	return( b );
}

double  CbxMath::
LengthSquared2( const double a[ 2 ] )
{
	double  b;

	b = a[ 0 ] * a[ 0 ] + a[ 1 ] * a[ 1 ];

	return( b );
}

double  CbxMath::
LengthSquared3( const double a[ 3 ] )
{
	double  b;

	b = a[ 0 ] * a[ 0 ] + a[ 1 ] * a[ 1 ] + a[ 2 ] * a[ 2 ];

	return( b );
}

double*  CbxMath::
Normalize2( const double a[ 2 ], const double* valueWhenErr/*[ 2 ]*/, double z[ 2 ], bool* errResult /*= 0*/ )
{
	double  l;

	l = sqrt( a[ 0 ] * a[ 0 ] + a[ 1 ] * a[ 1 ] );

	if( l < nearZero ) {
		if( errResult != 0 )
			*errResult = true;
		if( valueWhenErr == 0 )
			return( 0 );
		z[ 0 ] = valueWhenErr[ 0 ];		z[ 1 ] = valueWhenErr[ 1 ];
	}
	else {
		if( errResult != 0 )
			*errResult = false;
		z[ 0 ] = a[ 0 ] / l;
		z[ 1 ] = a[ 1 ] / l;
	}
	
	return( z );
}

double*  CbxMath::
Normalize3( const double a[ 3 ], const double* valueWhenErr/*[ 3 ]*/, double z[ 3 ], bool* errResult /*= 0*/ )
{
	double  l;

	l = sqrt( a[ 0 ] * a[ 0 ] + a[ 1 ] * a[ 1 ] + a[ 2 ] * a[ 2 ] );

	if( l < nearZero ) {
		if( errResult != 0 )
			*errResult = true;
		if( valueWhenErr == 0 )
			return( 0 );
		z[ 0 ] = valueWhenErr[ 0 ];		z[ 1 ] = valueWhenErr[ 1 ];;		z[ 2 ] = valueWhenErr[ 2 ];
	}
	else {
		if( errResult != 0 )
			*errResult = false;
		z[ 0 ] = a[ 0 ] / l;
		z[ 1 ] = a[ 1 ] / l;
		z[ 2 ] = a[ 2 ] / l;
	}
	
	return( z );
}

double  CbxMath::
Cross2( const double a[ 2 ], const double b[ 2 ] )
{
	double  c;

	c = a[ 0 ] * b[ 1 ] - a[ 1 ] * b[ 0 ];

	return( c );
}

double*  CbxMath::
Cross3( const double a[ 3 ], const double b[ 3 ], double z[ 3 ] )
{
	double  c[ 3 ];

	c[ 0 ] = a[ 1 ] * b[ 2 ] - a[ 2 ] * b[ 1 ];
	c[ 1 ] = a[ 2 ] * b[ 0 ] - a[ 0 ] * b[ 2 ];
	c[ 2 ] = a[ 0 ] * b[ 1 ] - a[ 1 ] * b[ 0 ];

	z[ 0 ] = c[ 0 ];	z[ 1 ] = c[ 1 ];	z[ 2 ] = c[ 2 ];

	return( z );
}

double  CbxMath::
Dot2( const double a[ 2 ], const double b[ 2 ] )
{
	double  c;

	c = a[ 0 ] * b[ 0 ] + a[ 1 ] * b[ 1 ];

	return( c );
}

double  CbxMath::
Dot3( const double a[ 3 ], const double b[ 3 ] )
{
	double  c;

	c = a[ 0 ] * b[ 0 ] + a[ 1 ] * b[ 1 ] + a[ 2 ] * b[ 2 ];

	return( c );
}

double*  CbxMath::
Transform2( const double a[ 2 ], const double b[ 3 ][ 3 ], double z[ 2 ] )
{
	double  c[ 2 ];

	c[ 0 ] = a[ 0 ] * b[ 0 ][ 0 ] + a[ 1 ] * b[ 1 ][ 0 ] + b[ 2 ][ 0 ];
	c[ 1 ] = a[ 0 ] * b[ 0 ][ 1 ] + a[ 1 ] * b[ 1 ][ 1 ] + b[ 2 ][ 1 ];

	z[ 0 ] = c[ 0 ];	z[ 1 ] = c[ 1 ];

	return( z );
}

double*  CbxMath::
Transform3( const double a[ 3 ], const double b[ 4 ][ 4 ], double z[ 3 ] )
{
	double  c[ 3 ];

	c[ 0 ] = a[ 0 ] * b[ 0 ][ 0 ] + a[ 1 ] * b[ 1 ][ 0 ] + a[ 2 ] * b[ 2 ][ 0 ] + b[ 3 ][ 0 ];
	c[ 1 ] = a[ 0 ] * b[ 0 ][ 1 ] + a[ 1 ] * b[ 1 ][ 1 ] + a[ 2 ] * b[ 2 ][ 1 ] + b[ 3 ][ 1 ];
	c[ 2 ] = a[ 0 ] * b[ 0 ][ 2 ] + a[ 1 ] * b[ 1 ][ 2 ] + a[ 2 ] * b[ 2 ][ 2 ] + b[ 3 ][ 2 ];

	z[ 0 ] = c[ 0 ];	z[ 1 ] = c[ 1 ];	z[ 2 ] = c[ 2 ];

	return( z );
}

double*  CbxMath::
Negate2( const double a[ 2 ], double z[ 2 ] )
{
	z[ 0 ] = -a[ 0 ];		z[ 1 ] = -a[ 1 ];

	return( z );
}

double*  CbxMath::
Negate3( const double a[ 3 ], double z[ 2 ] )
{
	z[ 0 ] = -a[ 0 ];		z[ 1 ] = -a[ 1 ];		z[ 2 ] = -a[ 2 ];

	return( z );
}

double*  CbxMath::
Abs2( const double a[ 2 ], double z[ 2 ] )
{
	if( a[ 0 ] >= 0.0 )
		z[ 0 ] =  a[ 0 ];
	else
		z[ 0 ] = -a[ 0 ];

	if( a[ 1 ] >= 0.0 )
		z[ 1 ] =  a[ 1 ];
	else
		z[ 1 ] = -a[ 1 ];

	return( z );
}

double*  CbxMath::
Abs3( const double a[ 3 ], double z[ 3 ] )
{
	if( a[ 0 ] >= 0.0 )
		z[ 0 ] =  a[ 0 ];
	else
		z[ 0 ] = -a[ 0 ];

	if( a[ 1 ] >= 0.0 )
		z[ 1 ] =  a[ 1 ];
	else
		z[ 1 ] = -a[ 1 ];

	if( a[ 2 ] >= 0.0 )
		z[ 2 ] =  a[ 2 ];
	else
		z[ 2 ] = -a[ 2 ];

	return( z );
}



CbxMath::double3*  CbxMath::
Copy3x3( const double a[ 3 ][ 3 ], double z[ 3 ][ 3 ] )
{
	short  u, v;

	for( v=0; v<3; v++ ) {
		for( u=0; u<3; u++ )
			z[ u ][ v ] = a[ u ][ v ];
	}

	return( z );
}

CbxMath::double4*  CbxMath::
Copy4x4( const double a[ 4 ][ 4 ], double z[ 4 ][ 4 ] )
{
	short  u, v;

	for( v=0; v<4; v++ ) {
		for( u=0; u<4; u++ )
			z[ u ][ v ] = a[ u ][ v ];
	}

	return( z );
}

CbxMath::double3*  CbxMath::
MakeIdentity3x3( double z[ 3 ][ 3 ] )
{
	z[ 0 ][ 0 ] = 1.0;	z[ 1 ][ 0 ] = 0.0;	z[ 2 ][ 0 ] = 0.0; 
	z[ 0 ][ 1 ] = 0.0;	z[ 1 ][ 1 ] = 1.0;	z[ 2 ][ 1 ] = 0.0; 
	z[ 0 ][ 2 ] = 0.0;	z[ 1 ][ 2 ] = 0.0;	z[ 2 ][ 2 ] = 1.0; 

	return( z );
}

CbxMath::double4*  CbxMath::
MakeIdentity4x4( double z[ 4 ][ 4 ] )
{
	z[ 0 ][ 0 ] = 1.0;	z[ 1 ][ 0 ] = 0.0;	z[ 2 ][ 0 ] = 0.0;	z[ 3 ][ 0 ] = 0.0; 
	z[ 0 ][ 1 ] = 0.0;	z[ 1 ][ 1 ] = 1.0;	z[ 2 ][ 1 ] = 0.0;	z[ 3 ][ 1 ] = 0.0; 
	z[ 0 ][ 2 ] = 0.0;	z[ 1 ][ 2 ] = 0.0;	z[ 2 ][ 2 ] = 1.0;	z[ 3 ][ 2 ] = 0.0; 
	z[ 0 ][ 3 ] = 0.0;	z[ 1 ][ 3 ] = 0.0;	z[ 2 ][ 3 ] = 0.0;	z[ 3 ][ 3 ] = 1.0; 

	return( z );
}

CbxMath::double3*  CbxMath::
Multiply3x3( const double a[ 3 ][ 3 ], const double b[ 3 ][ 3 ], double z[ 3 ][ 3 ] )
{
	double  c[ 3 ][ 3 ];
	short   u, v;

	for( v=0; v<3; v++ ) {
		for( u=0; u<3; u++ ) {
			c[ u ][ v ] = a[ 0 ][ v ] * b[ u ][ 0 ]
				        + a[ 1 ][ v ] * b[ u ][ 1 ]
				        + a[ 2 ][ v ] * b[ u ][ 2 ];
		}
	}

	for( v=0; v<3; v++ ) {
		for( u=0; u<3; u++ )
			z[ u ][ v ] = c[ u ][ v ];
	}

	return( z );
}

CbxMath::double4*  CbxMath::
Multiply4x4( const double a[ 4 ][ 4 ], const double b[ 4 ][ 4 ], double z[ 4 ][ 4 ] )
{
	double  c[ 4 ][ 4 ];
	short   u, v;

	for( v=0; v<4; v++ ) {
		for( u=0; u<4; u++ ) {
			c[ u ][ v ] = a[ 0 ][ v ] * b[ u ][ 0 ]
				        + a[ 1 ][ v ] * b[ u ][ 1 ]
				        + a[ 2 ][ v ] * b[ u ][ 2 ]
				        + a[ 3 ][ v ] * b[ u ][ 3 ];
		}
	}

	for( v=0; v<4; v++ ) {
		for( u=0; u<4; u++ )
			z[ u ][ v ] = c[ u ][ v ];
	}

	return( z );
}

CbxMath::double3*  CbxMath::
Translete3x3( const double x, const double y, double result[ 3 ][ 3 ] )
{
	MakeIdentity3x3( result );
	result[ 2 ][ 0 ] = x;
	result[ 2 ][ 1 ] = y;

	return( result );
}

CbxMath::double4*  CbxMath::
Translete4x4( const double x, const double y, const double z, double result[ 4 ][ 4 ] )
{
	MakeIdentity4x4( result );
	result[ 3 ][ 0 ] = x;
	result[ 3 ][ 1 ] = y;
	result[ 3 ][ 2 ] = z;

	return( result );
}

CbxMath::double3*  CbxMath::
Scale3x3( const double x, const double y, double result[ 3 ][ 3 ] )
{
	MakeIdentity3x3( result );
	result[ 0 ][ 0 ] = x;
	result[ 1 ][ 1 ] = y;

	return( result );
}

CbxMath::double4*  CbxMath::
Scale4x4( const double x, const double y, const double z, double result[ 4 ][ 4 ] )
{
	MakeIdentity4x4( result );
	result[ 3 ][ 0 ] = x;
	result[ 3 ][ 1 ] = y;
	result[ 3 ][ 2 ] = z;

	return( result );
}

CbxMath::double3*  CbxMath::
Rotate3x3( const double a, double z[ 3 ][ 3 ] )
{
	double  b;

	b = a * pi / 180.0;

	RotateRad3x3( b, z );

	return( z );
}

CbxMath::double3*  CbxMath::
RotateRad3x3( const double a, double z[ 3 ][ 3 ] )
{
	double  dtSin, dtCos;

	dtSin = sin( a );
	dtCos = cos( a );

	MakeIdentity3x3( z );
	z[ 0 ][ 0 ] =  dtCos;	z[ 1 ][ 0 ] = -dtSin;
	z[ 0 ][ 1 ] =  dtSin;	z[ 1 ][ 1 ] =  dtCos;

	return( z );
}

CbxMath::double4*  CbxMath::
RotateX4x4( const double x, double result[ 4 ][ 4 ] )
{
	double  x2;

	x2 = x * pi / 180.0;

	RotateXRad4x4( x2, result );

	return( result );
}

CbxMath::double4*  CbxMath::
RotateY4x4( const double y, double result[ 4 ][ 4 ] )
{
	double  y2;

	y2 = y * pi / 180.0;

	RotateYRad4x4( y2, result );

	return( result );
}

CbxMath::double4*  CbxMath::
RotateZ4x4( const double z, double result[ 4 ][ 4 ] )
{
	double  z2;

	z2 = z * pi / 180.0;

	RotateZRad4x4( z2, result );

	return( result );
}

CbxMath::double4*  CbxMath::
RotateXRad4x4( const double x, double result[ 4 ][ 4 ] )
{
	double  dtSin, dtCos;

	dtSin = sin( x );
	dtCos = cos( x );

	MakeIdentity4x4( result );
	result[ 1 ][ 1 ] =  dtCos;	result[ 2 ][ 1 ] = -dtSin;
	result[ 1 ][ 2 ] =  dtSin;	result[ 2 ][ 2 ] =  dtCos;

	return( result );
}

CbxMath::double4*  CbxMath::
RotateYRad4x4( const double y, double result[ 4 ][ 4 ] )
{
	double  dtSin, dtCos;

	dtSin = sin( y );
	dtCos = cos( y );

	MakeIdentity4x4( result );
	result[ 2 ][ 2 ] =  dtCos;	result[ 0 ][ 2 ] = -dtSin;
	result[ 2 ][ 0 ] =  dtSin;	result[ 0 ][ 0 ] =  dtCos;

	return( result );
}

CbxMath::double4*  CbxMath::
RotateZRad4x4( const double z, double result[ 4 ][ 4 ] )
{
	double  dtSin, dtCos;

	dtSin = sin( z );
	dtCos = cos( z );

	MakeIdentity4x4( result );
	result[ 0 ][ 0 ] =  dtCos;	result[ 1 ][ 0 ] = -dtSin;
	result[ 0 ][ 1 ] =  dtSin;	result[ 1 ][ 1 ] =  dtCos;

	return( result );
}

CbxMath::double4*  CbxMath::
RotateAxis4x4( const double angle, const double axis[ 3 ], double result[ 4 ][ 4 ], bool* errResult /*= 0*/ )
{
	double  angle2;
	bool    err;

	if( errResult != 0 )
		*errResult = false;

	angle2 = angle * pi / 180.0;

	RotateAxisRad4x4( angle2, axis, result, &err );
	if( err == true ) {
		if( errResult != 0 )
			*errResult = true;
		return( result );
	}

	return( result );
}

CbxMath::double4*  CbxMath::
RotateAxisRad4x4( const double angle, const double axis[ 3 ], double result[ 4 ][ 4 ], bool* errResult /*= 0*/ )
{
	double  axis2[ 3 ], l, a[ 3 ][ 3 ], pSin, pCos, mCos;
	short   u, v;

	if( errResult != 0 )
		*errResult = false;
	
	l = Length3( axis );
	if( l < nearZero ) {
		if( errResult != 0 )
			*errResult = true;
		return( result );
	}
	Scale3( axis, 1.0/l, axis2 );

	pSin = sin( angle );
	pCos = cos( angle );
	mCos = 1.0 - pCos;

	a[ 0 ][ 0 ] =  pCos;
	a[ 1 ][ 0 ] = -pSin * axis2[ 2 ];
	a[ 2 ][ 0 ] =  pSin * axis2[ 1 ];

	a[ 0 ][ 1 ] =  pSin * axis2[ 2 ];
	a[ 1 ][ 1 ] =  pCos;
	a[ 2 ][ 1 ] = -pSin * axis2[ 0 ];

	a[ 0 ][ 2 ] = -pSin * axis2[ 1 ];
	a[ 1 ][ 2 ] =  pSin * axis2[ 0 ];
	a[ 2 ][ 2 ] =  pCos;

	for( v=0; v<3; v++ ) {
		for( u=0; u<3; u++ )
			result[ u ][ v ] = a[ u ][ v ] + mCos * axis2[ u ] * axis2[ v ];
		result[ 3 ][ v ] = 0.0;
	}
	result[ 0 ][ 3 ] = 0.0;		result[ 1 ][ 3 ] = 0.0;
	result[ 2 ][ 3 ] = 0.0;		result[ 3 ][ 3 ] = 1.0;

	return( result );
}



CbxMath::double3*  CbxMath::
SetAxis( const double mainAxis[ 3 ], const double subAxis[ 3 ], double result[ 2 ][ 3 ] )
{
	short  i;

	for( i=0; i<3; i++ ) {
		result[ 0 ][ i ] = mainAxis[ i ];
		result[ 1 ][ i ] = subAxis[ i ];
	}

	return( result );
}

CbxMath::double3*  CbxMath::
SetAxisXYZ( const double xAxis[ 3 ], const double yAxis[ 3 ], const double zAxis[ 3 ], double result[ 3 ][ 3 ] )
{
	short  i;

	for( i=0; i<3; i++ ) {
		result[ 0 ][ i ] = xAxis[ i ];
		result[ 1 ][ i ] = yAxis[ i ];
		result[ 2 ][ i ] = zAxis[ i ];
	}

	return( result );
}

CbxMath::double3*  CbxMath::
CopyAxis( const double a[ 2 ][ 3 ], double z[ 2 ][ 3 ] )
{
	short  p, i;

	for( p=0; p<2; p++ ) {
		for( i=0; i<3; i++ )
			z[ p ][ i ] = a[ p ][ i ];
	}

	return( z );
}

CbxMath::double3*  CbxMath::
CopyAxisXYZ( const double a[ 3 ][ 3 ], double z[ 3 ][ 3 ] )
{
	short  p, i;

	for( p=0; p<3; p++ ) {
		for( i=0; i<3; i++ )
			z[ p ][ i ] = a[ p ][ i ];
	}

	return( z );
}

CbxMath::double3*  CbxMath::
RegularAxis( const double a[ 2 ][ 3 ], double z[ 2 ][ 3 ], bool* errResult /*= 0*/ )
{
	double  l, b[ 3 ];

	if( errResult != 0 )
		*errResult = false;

	l = Length3( a[ 0 ] );
	if( l < nearZero ) {
		if( errResult != 0 )
			*errResult = true;
		return( z );
	}
	Scale3( a[ 0 ], 1.0/l, z[ 0 ] );

	Cross3( z[ 0 ], a[ 1 ], b );
	Cross3( b, z[ 0 ], b );
	l = Length3( b );
	if( l < nearZero ) {
		if( errResult != 0 )
			*errResult = true;
		return( z );
	}
	Scale3( b, 1.0/l, z[ 1 ] );

	return( z );
}

CbxMath::double3*  CbxMath::
MakeSubAxis( const double mainAxis[ 3 ], double result[ 2 ][ 3 ], bool* errResult /*= 0*/ )
{
	double  a[ 3 ], b[ 2 ][ 3 ];
	short   kindAxis;
	bool    err;

	if( errResult != 0 )
		*errResult = false;

	Abs3( mainAxis, a );
	if( a[ 0 ] >= a[ 1 ] ) {
		if( a[ 0 ] >= a[ 2 ] )
			kindAxis = 0;
		else
			kindAxis = 2;
	}
	else {
		if( a[ 1 ] >= a[ 2 ] )
			kindAxis = 1;
		else
			kindAxis = 2;
	}

	switch( kindAxis ) {
		case  0:
			Copy3( worldAxis[ 1 ], a );
			break;
		case  1:
			Copy3( worldAxis[ 2 ], a );
			break;
		case  2:
			Copy3( worldAxis[ 0 ], a );
			break;
	}

	SetAxis( mainAxis, a, b );
	RegularAxis( b, result, &err );
	if( err == true ) {
		if( errResult != 0 )
			*errResult = true;
		return( result );
	}

	return( result );
}

CbxMath::double4*  CbxMath::
RotWorldToLocal( const double axis[ 2 ][ 3 ], double result[ 4 ][ 4 ], bool* errResult /*= 0*/ )
{
	double  dstAxis[ 3 ][ 3 ];
	bool    err;

	if( errResult != 0 )
		*errResult = false;

	RegularAxis( axis, dstAxis, &err );
	if( err == true ) {
		if( errResult != 0 )
			*errResult = true;
		return( result );
	}
	Cross3( dstAxis[ 0 ], dstAxis[ 1 ], dstAxis[ 2 ] );

	RotWorld_Local_Sub( worldAxis, dstAxis, result );

	return( result );
}

CbxMath::double4*  CbxMath::
RotLocalToWorld( const double axis[ 2 ][ 3 ], double result[ 4 ][ 4 ], bool* errResult /*= 0*/ )
{
	double  srcAxis[ 3 ][ 3 ];
	bool    err;

	if( errResult != 0 )
		*errResult = false;

	RegularAxis( axis, srcAxis, &err );
	if( err == true ) {
		if( errResult != 0 )
			*errResult = true;
		return( result );
	}
	Cross3( srcAxis[ 0 ], srcAxis[ 1 ], srcAxis[ 2 ] );

	RotWorld_Local_Sub( srcAxis, worldAxis, result );

	return( result );
}

void  CbxMath::
RotWorld_Local_Sub( const double srcAxis[ 3 ][ 3 ], const double dstAxis[ 3 ][ 3 ], double result[ 4 ][ 4 ] )
{
	short  u, v;

	for( v=0; v<3; v++ ) {
		for( u=0; u<3; u++ )
			result[ u ][ v ] = Dot3( srcAxis[ v ], dstAxis[ u ] );
		result[ 3 ][ v ] = 0.0;
	}
	result[ 0 ][ 3 ] = 0.0;		result[ 1 ][ 3 ] = 0.0;
	result[ 2 ][ 3 ] = 0.0;		result[ 3 ][ 3 ] = 1.0;
}

CbxMath::double4*  CbxMath::
RotateAxis( const double srcAxis[ 2 ][ 3 ], const double dstAxis[ 2 ][ 3 ], double result[ 4 ][ 4 ],
	bool* errResult /*= 0*/ )
{
	double  mat0[ 4 ][ 4 ], mat1[ 4 ][ 4 ];
	bool    err;

	if( errResult != 0 )
		*errResult = false;

	RotLocalToWorld( srcAxis, mat0, &err );
	if( err == true ) {
		if( errResult != 0 )
			*errResult = true;
		return( result );
	}

	RotWorldToLocal( dstAxis, mat1, &err );
	if( err == true ) {
		if( errResult != 0 )
			*errResult = true;
		return( result );
	}
	
	Multiply4x4( mat0, mat1, result );

	return( result );
}



bool  CbxMath::
IsMinusSin( const double axis[ 3 ], const double zeroDegVec[ 3 ], const double vecOnPlane[ 3 ] )
{
	double  dot, cross[ 3 ];

	CbxMath::Cross3( axis, zeroDegVec, cross );

	dot = Dot3( cross, vecOnPlane );
	if( dot < 0 )
		return( true );

	return( false );
}


double  CbxMath::
CosToDeg( const double cosP, const bool minusSin, bool* errResult /*= 0*/ )
{
	double  rad;
	bool    err;

	rad = CosToRad( cosP, minusSin, &err );
	if( err == true ) {
		if( errResult != 0 )
			*errResult = true;
		return( 0.0 );
	}

	if( errResult != 0 )
		*errResult = false;

	return( rad * 180.0 / pi );
}
		
		
double  CbxMath::
CosToRad( const double cosP, const bool minusSin, bool* errResult /*= 0*/ )
{
	double  cosP2, rad;

	if( ( -1.0 - nearZero ) > cosP || cosP > ( 1.0 + nearZero ) ) {
		if( errResult != 0 )
			*errResult = true;
		return( 0.0 );
	}

	if( errResult != 0 )
		*errResult = false;

	cosP2 = cosP;
	if( cosP > 1.0 )
		cosP2 = 1.0;
	else if( cosP < -1.0 )
		cosP2 = -1.0;

	rad = acos( cosP2 );
	if( minusSin == true )
		rad = -rad;

	return( rad );
}


double  CbxMath::
NegateRad( const double rad )
{
	double  deg, deg2, rad2;

	deg  = rad * 180.0 / pi;
	deg2 = RegularDeg( deg + 180.0 );
	rad2 = deg2 * pi / 180.0;

	return( rad2 );
}


double  CbxMath::
NegateDeg( const double deg )
{
	return( RegularDeg( deg + 180.0 ) );
}


double  CbxMath::
RegularRad( const double rad )
{
	double  deg, deg2, rad2;

	deg  = rad * 180.0 / pi;
	deg2 = RegularDeg( deg );
	rad2 = deg2 * pi / 180.0;

	return( rad2 );
}


double  CbxMath::
RegularDeg( const double deg )
{
	double  deg2;

	deg2 = UnsignedRegularDeg( deg );

	if( deg2 > 180.0 )
		deg2 = 360.0 - deg2;

	return( deg2 );
}


double  CbxMath::
UnsignedRegularRad( const double rad )
{
	double  deg, deg2, rad2;

	deg  = rad * 180.0 / pi;
	deg2 = UnsignedRegularDeg( deg );
	rad2 = deg2 * pi / 180.0;

	return( rad2 );
}


double  CbxMath::
UnsignedRegularDeg( const double deg )
{
	double  deg2;

	if( deg >= 0 )
		deg2 = fmod( deg, 360.0 );
	else
		deg2 = 360.0 - fmod( -deg, 360.0 );

	return( deg2 );
}
