/* bxMath.h */

class CbxMath {
		// point[ 2 ] or vector[ 2 ] ...   ex. a[ i ]      : i = 0:X-elm, 1:Y-elm
		// point[ 3 ] or vector[ 3 ] ...   ex. a[ i ]      : i = 0:X-elm, 1:Y-elm, 2:Z-elm
		// matrix3x3[ 3 ][ 3 ]       ...   ex. a[ u ][ v ] : u = raw,  v = column
		// matrix4x4[ 4 ][ 4 ]       ...   ex. a[ u ][ v ] : u = raw,  v = column
		// axisVector[ 2 ][ 3 ]      ...   ex. a[ p ][ i ] : p = 0:mainAxis, 1:subAxis
	    //                                                 : i = 0:X-elm, 1:Y-elm, 2:Z-elm
		// axisXYZvector[ 3 ][ 3 ]   ...   ex. a[ p ][ i ] : p = 0:X-axis, 1:Y-axis, 2:Z-axis
	    //                                                 : i = 0:X-elm, 1:Y-elm, 2:Z-elm
	public:
		typedef double  double3[ 3 ];
		typedef double  double4[ 4 ];

	public:
		static double*
		Set2( const double a, const double b, double z[ 2 ] );
		static double*
		Set3( const double a, const double b, const double c, double z[ 3 ] );
		static double*
		Copy2( const double a[ 2 ], double z[ 2 ] );
		static double*
		Copy3( const double a[ 3 ], double z[ 3 ] );
		static double*
		Add2( const double a[ 2 ], const double b[ 2 ], double z[ 2 ] );
		static double*
		Add3( const double a[ 3 ], const double b[ 3 ], double z[ 3 ] );
		static double*
		Subtract2( const double a[ 2 ], const double b[ 2 ], double z[ 2 ] );
		static double*
		Subtract3( const double a[ 3 ], const double b[ 3 ], double z[ 3 ] );
		static double
		Distance2( const double a[ 2 ], const double b[ 2 ] );
		static double
		Distance3( const double a[ 3 ], const double b[ 3 ] );
		static double
		DistanceSquared2( const double a[ 2 ], const double b[ 2 ] );
		static double
		DistanceSquared3( const double a[ 3 ], const double b[ 3 ] );
		static double*
		Ratio2( const double a[ 2 ], const double b[ 2 ], const double t, double z[ 2 ] );
		static double*
		Ratio3( const double a[ 3 ], const double b[ 3 ], const double t, double z[ 3 ] );
		static double*
		Scale2( const double a[ 2 ], const double scale, double z[ 2 ] );
		static double*
		Scale3( const double a[ 3 ], const double scale, double z[ 3 ] );
		static double
		Length2( const double a[ 2 ] );
		static double
		Length3( const double a[ 3 ] );
		static double
		LengthSquared2( const double a[ 2 ] );
		static double
		LengthSquared3( const double a[ 3 ] );
		static double*
		Normalize2( const double a[ 2 ], const double* valueWhenErr/*[ 2 ]*/, double z[ 2 ], bool* errResult = 0 );
		static double*
		Normalize3( const double a[ 3 ], const double* valueWhenErr/*[ 3 ]*/, double z[ 3 ], bool* errResult = 0 );
		static double
		Cross2( const double a[ 2 ], const double b[ 2 ] );
		static double*
		Cross3( const double a[ 3 ], const double b[ 3 ], double z[ 3 ] );
		static double
		Dot2( const double a[ 2 ], const double b[ 2 ] );
		static double
		Dot3( const double a[ 3 ], const double b[ 3 ] );
		static double*
		Transform2( const double a[ 2 ], const double b[ 3 ][ 3 ], double z[ 2 ] );
		static double*
		Transform3( const double a[ 3 ], const double b[ 4 ][ 4 ], double z[ 3 ] );
		static double*
		Negate2( const double a[ 2 ], double z[ 2 ] );
		static double*
		Negate3( const double a[ 3 ], double z[ 2 ] );
		static double*
		Abs2( const double a[ 2 ], double z[ 2 ] );
		static double*
		Abs3( const double a[ 3 ], double z[ 2 ] );

		static double3*
		Copy3x3( const double a[ 3 ][ 3 ], double z[ 3 ][ 3 ] );
		static double4*
		Copy4x4( const double a[ 4 ][ 4 ], double z[ 4 ][ 4 ] );
		static double3*
		MakeIdentity3x3( double z[ 3 ][ 3 ] );
		static double4*
		MakeIdentity4x4( double z[ 4 ][ 4 ] );
		static double3*
		Multiply3x3( const double a[ 3 ][ 3 ], const double b[ 3 ][ 3 ], double z[ 3 ][ 3 ] );
		static double4*
		Multiply4x4( const double a[ 4 ][ 4 ], const double b[ 4 ][ 4 ], double z[ 4 ][ 4 ] );
		static double3*
		Translete3x3( const double x, const double y, double result[ 3 ][ 3 ] );
		static double4*
		Translete4x4( const double x, const double y, const double z, double result[ 4 ][ 4 ] );
		static double3*
		Scale3x3( const double x, const double y, double result[ 3 ][ 3 ] );
		static double4*
		Scale4x4( const double x, const double y, const double z, double result[ 4 ][ 4 ] );
		static double3*
		Rotate3x3( const double a, double z[ 3 ][ 3 ] );
		static double3*
		RotateRad3x3( const double a, double z[ 3 ][ 3 ] );
		static double4*
		RotateX4x4( const double x, double result[ 4 ][ 4 ] );
		static double4*
		RotateY4x4( const double y, double result[ 4 ][ 4 ] );
		static double4*
		RotateZ4x4( const double z, double result[ 4 ][ 4 ] );
		static double4*
		RotateXRad4x4( const double x, double result[ 4 ][ 4 ] );
		static double4*
		RotateYRad4x4( const double y, double result[ 4 ][ 4 ] );
		static double4*
		RotateZRad4x4( const double z, double result[ 4 ][ 4 ] );
		static double4*
		RotateAxis4x4( const double angle, const double axis[ 3 ], double result[ 4 ][ 4 ], bool* errResult = 0 );
		static double4*
		RotateAxisRad4x4( const double angle, const double axis[ 3 ], double result[ 4 ][ 4 ], bool* errResult = 0 );

		static double3*
		SetAxis( const double mainAxis[ 3 ], const double subAxis[ 3 ], double result[ 2 ][ 3 ] );
		static double3*
		SetAxisXYZ( const double xAxis[ 3 ], const double yAxis[ 3 ], const double zAxis[ 3 ], double result[ 3 ][ 3 ] );
		static double3*
		CopyAxis( const double a[ 2 ][ 3 ], double z[ 2 ][ 3 ] );
		static double3*
		CopyAxisXYZ( const double a[ 3 ][ 3 ], double z[ 3 ][ 3 ] );
		static double3*
		RegularAxis( const double a[ 2 ][ 3 ], double z[ 2 ][ 3 ], bool* errResult = 0 );
		static double3*
		MakeSubAxis( const double mainAxis[ 3 ], double result[ 2 ][ 3 ], bool* errResult = 0 );
		static double4*
		RotWorldToLocal( const double axis[ 2 ][ 3 ], double result[ 4 ][ 4 ], bool* errResult = 0 );
		static double4*
		RotLocalToWorld( const double axis[ 2 ][ 3 ], double result[ 4 ][ 4 ], bool* errResult = 0 );
		static double4*
		RotateAxis( const double srcAxis[ 2 ][ 3 ], const double dstAxis[ 2 ][ 3 ], double result[ 4 ][ 4 ],
			bool* errResult = 0 );

		static bool
		IsMinusSin( const double axis[ 3 ], const double zeroDegVec[ 3 ], const double vecOnPlane[ 3 ] );
		static double
		CosToDeg( const double cosP, const bool minusSin, bool* errResult = 0 );
		static double
		CosToRad( const double cosP, const bool minusSin, bool* errResult = 0 );
		static double
		NegateRad( const double rad );
		static double
		NegateDeg( const double deg );
		static double
		RegularRad( const double rad );
		static double
		RegularDeg( const double deg );
		static double
		UnsignedRegularRad( const double rad );
		static double
		UnsignedRegularDeg( const double deg );

	private:
		static void
		RotWorld_Local_Sub( const double srcAxis[ 3 ][ 3 ], const double dstAxis[ 3 ][ 3 ], double result[ 4 ][ 4 ] );

	public:
		static const double  pi;

		static const double  zero2[ 2 ];
		static const double  zero3[ 3 ];

		static const double  nearZero;
		static const double  nearZero2[ 2 ];
		static const double  nearZero3[ 3 ];
		static const double  minusNearZero2[ 2 ];
		static const double  minusNearZero3[ 3 ];

		static const double  identity3x3[ 3 ][ 3 ];
		static const double  identity4x4[ 4 ][ 4 ];

		static const double  worldAxis[ 3/*axis*/ ][ 3/*element*/ ];
};
