/* bxProc.cpp */

#include "bxProc.h"
#include "../Core/bxCore.h"
#include "../Math/bxMath.h"



bool  CbxProc::
DefaultAxis( const ulong vertexNo )
{
	ushort  numRadial;
	bool    result;
	CbxCore::AxisForDefaultSub*  buf;
	CbxCore::AxisForDefault      axis;

	CbxDB::NumRadial( vertexNo, &numRadial );
	result = AllocDefaultAxis( numRadial, &buf );
	if( result == true )
		return( true );

	result = CbxDB::LockAxisForDefault( vertexNo, buf, &axis );
	if( result == true )
		return( true );

	result = CbxCore::GetDefaultAxis( &axis.in, &axis.out );
	if( result == true ) {
		CbxDB::CancelLock();
		return( true );
	}

	CbxDB::UnlockAxisForDefault( &axis.out );

	FreeDefaultAxis( buf );

	return( false );
}


bool  CbxProc::
AllocDefaultAxis( const ushort numRadial, CbxCore::AxisForDefaultSub** buf )
{
	*buf = new CbxCore::AxisForDefaultSub[ numRadial ];
	if( *buf == 0 )
		return( true );

	return( false );
}


void  CbxProc::
FreeDefaultAxis( CbxCore::AxisForDefaultSub* buf )
{
	delete [] buf;
}


bool  CbxProc::
DefaultHandle( const ulong edgeNo, const short beginEndNo )
{
	bool  result;
	CbxCore::HandleForTrans  handle;

	result = CbxDB::LockHandleForTrans( edgeNo, beginEndNo, &handle );
	if( result == true )
		return( true );

	result = CbxCore::GetDefaultHandle( &handle.in, &handle.out );
	if( result == true ) {
		CbxDB::CancelLock();
		return( true );
	}

	CbxDB::UnlockHandleForTrans( &handle.out );

	return( false );
}


bool  CbxProc::
DefaultCtlVec( const ulong edgeNo, const short beginEndNo, bool* dispChk )
{
	bool  result;
	CbxCore::CtlVecForDefault  surface;

	result = CbxDB::LockCtlVecForDefault( edgeNo, beginEndNo, &surface );
	if( result == true )
		return( true );

	result = CbxCore::GetDefaultCtlVec( &surface.in, dispChk, &surface.out );
	if( result == true ) {
		CbxDB::CancelLock();
		return( true );
	}
	else if( *dispChk == false ) {
		CbxDB::CancelLock();
		return( false );
	}

	CbxDB::UnlockCtlVecForDefault( &surface.out );

	return( false );
}


bool  CbxProc::
SetVertexPos( const ulong vertexNo, const double newPos[ 3 ] )
{
	bool  result;
	CbxDB::VertexRec  vertex;

	result = CbxDB::LockVertexRec( vertexNo, &vertex );
	if( result == true )
		return( true );

	CbxCore::SetVertexPos( newPos, vertex.vertexPos );

	CbxDB::UnlockVertexRec( &vertex );

	return( false );
}


bool  CbxProc::
RotVertex( const ulong vertexNo, const double origin[ 3 ], const double axis[ 3 ], const double deg )
{
	bool  result;
	CbxCore::VertexForRot  rot;

	result = CbxDB::LockVertexForRot( vertexNo, &rot );
	if( result == true )
		return( true );

	result = CbxCore::RotVertex( origin, axis, deg, &rot.out );
	if( result == true ) {
		CbxDB::CancelLock();
		return( true );
	}

	CbxDB::UnlockVertexForRot( &rot.out );

	return( false );
}


bool  CbxProc::
ScaleVertex( const ulong vertexNo, const double origin[ 3 ], const double scale )
{
	bool  result;
	CbxDB::VertexRec  vertex;

	result = CbxDB::LockVertexRec( vertexNo, &vertex );
	if( result == true )
		return( true );

	result = CbxCore::ScaleVertex( origin, scale, vertex.vertexPos );
	if( result == true ) {
		CbxDB::CancelLock();
		return( true );
	}

	CbxDB::UnlockVertexRec( &vertex );

	return( false );
}


bool  CbxProc::
SetHandleVec( const ulong edgeNo, const short beginEndNo, const double newHandleVec[ 2 ] )
{
	bool  result;
	CbxDB::EdgeRec  edge;

	result = CbxDB::LockEdgeRec( edgeNo, &edge );
	if( result == true )
		return( true );

	CbxCore::SetHandleVec( newHandleVec, edge.vertex[ beginEndNo ].handleVec );

	CbxDB::UnlockEdgeRec( &edge );

	return( false );
}


bool  CbxProc::
SetHandleInfo( const ulong edgeNo, const short beginEndNo, const double newHandleVec[ 3 ] )
{
	bool  result, dispChk;
	CbxCore::HandleForTransPlus  trans;

	result = CbxDB::LockHandleForTransPlus( edgeNo, beginEndNo, &trans );
	if( result == true )
		return( true );

	result = CbxCore::SetHandleInfo( newHandleVec, &trans.in, &dispChk, &trans.out );
	if( result == true ) {
		CbxDB::CancelLock();
		return( true );
	}
	else if( dispChk == false ) {
		CbxDB::CancelLock();
		return( false );
	}

	CbxDB::UnlockHandleForTransPlus( &trans.out );

	return( false );
}


bool  CbxProc::
SetCtlPlaneDeg( const ulong edgeNo, const short beginEndNo, const double newCtlPlaneDeg )
{
	bool  result;
	CbxDB::EdgeRec  edge;

	result = CbxDB::LockEdgeRec( edgeNo, &edge );
	if( result == true )
		return( true );

	CbxCore::SetCtlPlaneDeg( newCtlPlaneDeg, &edge.vertex[ beginEndNo ].ctlPlaneDeg );

	CbxDB::UnlockEdgeRec( &edge );

	return( false );
}


bool  CbxProc::
SetCtlVec( const ulong edgeNo, const short beginEndNo, const double newCtlVec[ 2 ][ 2 ] )
{
	bool  result;
	CbxDB::EdgeRec  edge;

	result = CbxDB::LockEdgeRec( edgeNo, &edge );
	if( result == true )
		return( true );

	CbxCore::SetCtlVec( newCtlVec, edge.vertex[ beginEndNo ].ctlVec );

	CbxDB::UnlockEdgeRec( &edge );

	return( false );
}


bool  CbxProc::
SetCtlInfo( const ulong edgeNo, const short beginEndNo, const short minusPlusNo, const double newCtlVec[ 3 ] )
{
	bool  result, dispChk;
	CbxCore::CtlVecForTransPlus  ctl;

	result = CbxDB::LockCtlVecForTransPlus( edgeNo, beginEndNo, minusPlusNo, &ctl );
	if( result == true )
		return( true );

	result = CbxCore::SetCtlInfo( newCtlVec, &ctl.in, &dispChk, &ctl.out );
	if( result == true ) {
		CbxDB::CancelLock();
		return( true );
	}
	else if( dispChk == false ) {
		CbxDB::CancelLock();
		return( false );
	}

	CbxDB::UnlockCtlVecForTransPlus( &ctl.out );

	return( false );
}


void  CbxProc::
GetVertexPos( const ulong vertexNo, double vertexPos[ 3 ] )
{
	CbxDB::VertexRec  vertex;

	CbxDB::LoadVertexRec( vertexNo, &vertex );
	CbxMath::Copy3( vertex.vertexPos, vertexPos );
}


void  CbxProc::
GetNormalOnVertex( const ulong vertexNo, double normalVec[ 3 ] )
{
	CbxCore::VertexForRot  rot;

	CbxDB::LoadVertexForRot( vertexNo, &rot );
	CbxCore::GetNormalOnVertex( &rot, normalVec );
}


bool  CbxProc::
GetHandle( const ulong edgeNo, const short beginEndNo, bool* dispChk, double handleVec[ 3 ] )
{
	bool  result;
	CbxCore::HandleForTrans  handle;

	CbxDB::LoadHandleForTrans( edgeNo, beginEndNo, &handle );
	result = CbxCore::GetHandle( &handle, dispChk, handleVec );
	if( result == true )
		return( true );

	return( false );
}


bool  CbxProc::
ConvEdge( const ulong edgeNo, bool* dispChk, double bezierLine[ 4 ][ 3 ] )
{
	bool                     result;
	short                    i;
	CbxCore::HandleForTrans  vertex[ 2 ];

	for( i=0; i<2; i++ )
		CbxDB::LoadHandleForTrans( edgeNo, i, &vertex[ i ] );

	result = CbxCore::ConvEdge( vertex, dispChk, bezierLine );
	if( result == true )
		return( true );

	return( false );
}


bool  CbxProc::
GetAxisInCtlPlaneDeg( const ulong edgeNo, const short beginEndNo, bool* dispChk, double axisVec[ 3 ] )
{
	bool  result;
	CbxCore::CtlPlaneDegForRot  ctl;

	CbxDB::LoadCtlPlaneDegForRot( edgeNo, beginEndNo, &ctl );

	result = CbxCore::GetAxisInCtlPlaneDeg( &ctl, dispChk, axisVec );
	if( result == true )
		return( true );

	return( false );
}


bool  CbxProc::
GetNormalInCtlPlaneDeg( const ulong edgeNo, const short beginEndNo, bool* dispChk, double normalVec[ 3 ] )
{
	bool  result;
	CbxCore::CtlPlaneDegForRot  ctl;

	CbxDB::LoadCtlPlaneDegForRot( edgeNo, beginEndNo, &ctl );

	result = CbxCore::GetNormalInCtlPlaneDeg( &ctl, dispChk, normalVec );
	if( result == true )
		return( true );

	return( false );
}


bool  CbxProc::
GetAxisInCtlVec( const ulong edgeNo, const short beginEndNo, const short minusPlusNo, bool* dispChk, double axis[ 2 ][ 3 ] )
{
	bool  result;
	CbxCore::CtlVecForTrans  ctl;

	result = CbxDB::LoadCtlVecForTrans( edgeNo, beginEndNo, minusPlusNo, &ctl );
	if( result == true )
		return( true );

	result = CbxCore::GetAxisInCtlVec( &ctl, dispChk, axis );
	if( result == true )
		return( true );

	return( false );
}


bool  CbxProc::
GetCtlInCtlVec( const ulong edgeNo, const short beginEndNo, const short minusPlusNo, bool* dispChk, double ctlVec[ 3 ] )
{
	bool  result;
	CbxCore::CtlVecForTrans  ctl;
	
	result = CbxDB::LoadCtlVecForTrans( edgeNo, beginEndNo, minusPlusNo, &ctl );
	if( result == true )
		return( true );

	result = CbxCore::GetCtlInCtlVec( &ctl, dispChk, ctlVec );
	if( result == true )
		return( true );

	return( false );
}


bool  CbxProc::
ErrChk( void )
{
	ulong  i;
	short  j;

	if( CbxDB::NumVertexAry() < 3 )
		return( true );
	if( CbxDB::NumEdgeAry() < 3 )
		return( true );

	for( i=0; i<CbxDB::NumVertexAry(); i++ ) {
		if( ErrChkAxisVec( i ) == true )
			return( true );
	}

	for( i=0; i<CbxDB::NumEdgeAry(); i++ ) {
		for( j=0; j<2; j++ ) {
			if( ErrChkHandleVec1( i, j ) == true )
				return( true );
			if( ErrChkCtlPlaneDeg( i, j ) == true )
				return( true );
			if( ErrChkCtlVec( i, j ) == true )
				return( true );
		}
	}

	return( false );
}


bool  CbxProc::
ErrChkAxisVec( const ulong vertexNo )
{
	CbxDB::VertexRec  vertex;

	CbxDB::LoadVertexRec( vertexNo, &vertex );

	return( CbxCore::ErrChkAxisVec( vertex.axisVec ) );
}

bool  CbxProc::
ErrChkHandleVec1( const ulong edgeNo, const short beginEndNo )
{
	double  handleVec1;
	CbxDB::EdgeRec  edge;

	CbxDB::LoadEdgeRec( edgeNo, &edge );
	handleVec1 = edge.vertex[ beginEndNo ].handleVec[ 1 ];

	return( CbxCore::ErrChkHandleVec1( handleVec1 ) );
}


bool  CbxProc::
ErrChkCtlPlaneDeg( const ulong edgeNo, const short beginEndNo )
{
	double  ctlPlaneDeg;
	CbxDB::EdgeRec  edge;

	CbxDB::LoadEdgeRec( edgeNo, &edge );
	ctlPlaneDeg = edge.vertex[ beginEndNo ].ctlPlaneDeg;

	return( CbxCore::ErrChkCtlPlaneDeg( ctlPlaneDeg ) );
}


bool  CbxProc::
ErrChkCtlVec( const ulong edgeNo, const short beginEndNo )
{
	double  ctlVec[ 2 ][ 2 ];
	short   i;
	CbxDB::EdgeRec  edge;

	CbxDB::LoadEdgeRec( edgeNo, &edge );

	for( i=0; i<2; i++ )
		CbxMath::Copy2( edge.vertex[ beginEndNo ].ctlVec[ i ], ctlVec[ i ] );

	return( CbxCore::ErrChkCtlVec( ctlVec ) );
}


bool  CbxProc::
LoadSurfaceForPick( const ulong edgeNo, const short beginEndNo, CbxCore::SurfaceForPick* surface, bool* dispChk )
{
	bool  result;
	CbxCore::SurfaceForPick  surface2;

	result = CbxDB::LoadSurfaceForPick( edgeNo, beginEndNo, &surface2 );
	if( result == true )
		return( true );

	result = CbxCore::DispChkOnEdgeSurface( &surface2, dispChk );
	if( result == true )
		return( true );
	else if( *dispChk == false )
		return( false );

	*surface = surface2;

	return( false );
}
