/* bvAfterRead.cpp */

#include <stdlib.h>
#include "../BeziexView.h"



bool  CbvAfterRead::
Proc( void )
{
	bool  result;

	result = ConvReadData();
	if( result == true )
		return( true );

	result = ReadDataToBeziexDB();
	if( result == true )
		return( true );

	CbvRead::FreeData();

	result = GetDefaultAxis();
	if( result == true )
		return( true );

	result = GetDefaultHandle();
	if( result == true )
		return( true );

	result = GetDefaultCtlVec();
	if( result == true )
		return( true );

	result = ErrChkVertexPos();
	if( result == true )
		return( true );

	result = CbxProc::ErrChk();
	if( result == true )
		return( true );

	return( false );
}


bool  CbvAfterRead::
ConvReadData( void )
{
	RadialForSort*  sortRadial;
	bool            result;

	result = AllocSortRadial( &sortRadial );
	if( result == true )
		return( true );

	SetSortRadial( sortRadial );

	result = AfterMain( sortRadial );
	if( result == true ) {
		FreeSortRadial( sortRadial );
		return( true );
	}

	FreeSortRadial( sortRadial );

	return( false );
}


bool  CbvAfterRead::
AllocSortRadial( RadialForSort** radial )
{
	*radial = new RadialForSort[ CbvRead::numRadialAry ];
	if( *radial == 0 )
		return( true );

	return( false );
}


void  CbvAfterRead::
SetSortRadial( RadialForSort* radial )
{
	ulong  i;

	for( i=0; i<CbvRead::numRadialAry; i++ ) {
		radial[ i ].edgeNo  = CbvRead::radialAry[ i ].edgeNo;
		radial[ i ].indexNo = i;
	}

	qsort( radial, ( size_t )CbvRead::numRadialAry, sizeof( RadialForSort ), SortFunc );
}


int  CbvAfterRead::
SortFunc( const void* idt0, const void* idt1 )
{
	RadialForSort  *dt0, *dt1;

	dt0 = ( RadialForSort* )idt0;
	dt1 = ( RadialForSort* )idt1;

	if( dt0->edgeNo < dt1->edgeNo )
		return( -1 );
	else if( dt0->edgeNo > dt1->edgeNo )
		return( 1 );

	return( 0 );
}


bool  CbvAfterRead::
AfterMain( const RadialForSort* radial )
{
	ulong  i, id0, id1;

	if( ( CbvRead::numRadialAry % 2 ) != 0 )
		return( true );

	for( i=0; i<( CbvRead::numRadialAry/2 ); i++ ) {
		if( radial[ i*2 ].edgeNo != radial[ i*2+1 ].edgeNo )
			return( true );
	}

	for( i=0; i<( CbvRead::numRadialAry/2 ); i++ ) {
		id0 = radial[ i*2   ].indexNo;
		id1 = radial[ i*2+1 ].indexNo;

		CbvRead::radialAry[ id0 ].edgeNo = i;
		CbvRead::radialAry[ id1 ].edgeNo = i;

		CbvRead::radialAry[ id0 ].beginEndNo = 0;
		CbvRead::radialAry[ id1 ].beginEndNo = 1;
	}

	return( false );
}


void  CbvAfterRead::
FreeSortRadial( RadialForSort* radial )
{
	delete [] radial;
}


bool  CbvAfterRead::
ReadDataToBeziexDB( void )
{
	bool  result;

	result = AllocBeziexDB();
	if( result == true )
		return( true );

	ReadToDBvertex();
	ReadToDBedge();

	return( false );
}


bool  CbvAfterRead::
AllocBeziexDB( void )
{
	bool   result;
	ulong  numVertex, numEdge;

	numVertex = CbvRead::numVertexAry;
	numEdge   = CbvRead::numRadialAry / 2;

	result = CbxDB::AllocData( numVertex, numEdge );
	if( result == true )
		return( true );

	return( false );
}


void  CbvAfterRead::
ReadToDBvertex( void )
{
	ulong  i;
	CbxDB::VertexRec  vertex;

	for( i=0; i<CbvRead::numVertexAry; i++ ) {
		CbxDB::LockVertexRec( i, &vertex );
		CbxMath::Copy3( CbvRead::vertexAry[ i ].vertexPos, vertex.vertexPos );
		vertex.numRadial = CbvRead::vertexAry[ i ].numRadial;
		vertex.loopFlag  = CbvRead::vertexAry[ i ].loopFlag;
		CbxDB::UnlockVertexRec( &vertex );
	}
}


void  CbvAfterRead::
ReadToDBedge( void )
{
	ulong   i, edgeNo, beginEndNo;
	ushort  j;
	CbxDB::EdgeRec  edge;

	for( i=0; i<CbxDB::NumEdgeAry(); i++ ) {
		CbxDB::LockEdgeRec( i, &edge );
		CbxDB::InitilizeEdgeRec( &edge );
		CbxDB::UnlockEdgeRec( &edge );
	}

	for( i=0; i<CbvRead::numVertexAry; i++ ) {
		for( j=0; j<CbvRead::vertexAry[ i ].numRadial; j++ ) {
			edgeNo     = CbvRead::radialAry[ CbvRead::vertexAry[ i ].startIndex + j ].edgeNo;
			beginEndNo = CbvRead::radialAry[ CbvRead::vertexAry[ i ].startIndex + j ].beginEndNo;

			CbxDB::LockEdgeRec( edgeNo, &edge );
			edge.vertex[ beginEndNo ].vertexNo = i;
			edge.vertex[ beginEndNo ].radialNo = j;
			CbxDB::UnlockEdgeRec( &edge );
		}
	}
}


bool  CbvAfterRead::
GetDefaultAxis( void )
{
	ulong  i;
	bool   result;

	for( i=0; i<CbxDB::NumVertexAry(); i++ ) {
		result = CbxProc::DefaultAxis( i );
		if( result == true )
			return( true );
	}

	return( false );
}


bool  CbvAfterRead::
GetDefaultHandle( void )
{
	ulong  i;
	short  j;
	bool   result;

	for( i=0; i<CbxDB::NumEdgeAry(); i++ ) {
		for( j=0; j<2; j++ ) {
			result = CbxProc::DefaultHandle( i, j );
			if( result == true )
				return( true );
		}
	}

	return( false );
}


bool  CbvAfterRead::
GetDefaultCtlVec( void )
{
	bool  result, surfaceErr;

	result = CbxDB::AllocEdgeForPickSurface();
	if( result == true )
		return( true );

	result = GetDefaultCtlVecMain( &surfaceErr );
	if( result == true || surfaceErr == true ) {
		CbxDB::FreeEdgeForPickSurface();
		return( true );
	}

	CbxDB::FreeEdgeForPickSurface();

	return( false );
}


bool  CbvAfterRead::
GetDefaultCtlVecMain( bool* surfaceErr )
{
	ulong  i;
	short  j, chkSurfExist;
	bool   result, execOp;

	*surfaceErr = false;
	for( i=0; i<CbxDB::NumEdgeAry(); i++ ) {
		for( j=0; j<2; j++ ) {
			result = GetDefaultCtlVecVertex( i, j, &execOp, &chkSurfExist );
			if( result == true )
				return( true );
			else if( execOp == true && chkSurfExist == CbxDB::surfNonExist )
				*surfaceErr = true;
		}
	}

	return( false );
}


bool  CbvAfterRead::
GetDefaultCtlVecVertex( const ulong edgeNo, const short beginEndNo, bool* execOp, short* chkSurfExist )
{
	bool                     result, dispChk;
	ulong                    edgeNo2;
	short                    numAngle, i, beginEndNo2;
	CbxDB::EdgeVertexRadial  surfaceVertex[ 4 ];

	if( CbxDB::GetEdgeForPickSurface( edgeNo, beginEndNo ) == true ) {
		*execOp = false;
		return( false );
	}

	*execOp = true;

	result = CbxDB::GetSurfaceVertex( edgeNo, beginEndNo, chkSurfExist, &numAngle, surfaceVertex );
	if( result == true )
		return( true );
	else if( *chkSurfExist != CbxDB::surfExist )
		return( false );

	for( i=0; i<numAngle; i++ ) {
		edgeNo2     = surfaceVertex[ i ].edgeNo;
		beginEndNo2 = surfaceVertex[ i ].beginEndNo;

		CbxDB::SetEdgeForPickSurface( edgeNo2, beginEndNo2 );
	}

	result = CbxProc::DefaultCtlVec( edgeNo, beginEndNo, &dispChk );
	if( result == true )
		return( true );

	return( false );
}


bool  CbvAfterRead::
ErrChkVertexPos( void )
{
	ulong  i;
	bool   result;
	CbxDB::VertexRec  vertex;

	for( i=0; i<CbxDB::NumVertexAry(); i++ ) {
		CbxDB::LoadVertexRec( i, &vertex );
		result = CbvGL::ErrChkVertexPos( vertex.vertexPos );
		if( result == true )
			return( true );
	}

	return( false );
}
