/* bvRead.cpp */

#include <stdio.h>
#include <string.h>
#include "../BeziexView.h"

ulong         CbvRead::numVertexAry;
ulong         CbvRead::numRadialAry;
CbvRead::VertexRec*  CbvRead::vertexAry;
CbvRead::RadialRec*  CbvRead::radialAry;

const char    CbvRead::fileExt[ 16 ]        = "bxb";
const double  CbvRead::fileVersion          = 1.00;
const char    CbvRead::loopStr[ 2 ][ 16 ]   = { "no", "yes" };


bool  CbvRead::
Proc( HWND hWnd, bool* cancel )
{
	bool  result;

	result = ReadProc( hWnd, cancel );
	if( result == true && *cancel == false )
		CbvMessage::MessageOK( IDS_NOTOPENBFILE );
	if( result == true || *cancel == true )
		return( result );

	result = CbvAfterRead::Proc();
	if( result == true ) {
		CbvMessage::MessageOK( IDS_NOTOPENBFILE );
		return( true );
	}

	return( result );
}

			
bool  CbvRead::
ReadProc( HWND hWnd, bool* cancel )
{
	bool  result;

	result = OpenFile( hWnd, cancel );
	if( result == true )
		return( true );
	else if( *cancel == true )
		return( false );

	result = ReadMain();

	fclose( CbvReadCore::GetFP() );

	if( result == true )
		return( true );

	return( false );
}


bool  CbvRead::
OpenFile( HWND hWnd, bool* cancel )
{
	bool  result;
	char  fileName[ MAX_PATH ];

	result = GetFileName( hWnd, fileName, cancel );
	if( result == true )
		return( true );
	else if( *cancel == true )
		return( false );

	result = OpenFileMain( fileName );
	if( result == true )
		return( true );

	return( false );
}


bool  CbvRead::
GetFileName( HWND hWnd, char fileName[ MAX_PATH ], bool* cancel )
{
	OPENFILENAME  ofn;
	char          filter[ 256 ];
	short         stIndex;
	BOOL          result;
	DWORD         errNo;
	char          fileDlogFilter[ 256 ];

	*cancel = false;

	strcpy( fileName, "" );

	LoadString( CbvMain::HInst(), IDS_FILEFILTER, fileDlogFilter, 256-1 );
	memset( filter, 0, 256 );
	sprintf( filter, fileDlogFilter, fileExt );
	stIndex = strlen( filter ) + 1;
	sprintf( &filter[ stIndex ], "*.%s", fileExt );

	memset( &ofn, 0, sizeof( OPENFILENAME ) );
	ofn.lStructSize = sizeof( OPENFILENAME );
	ofn.hwndOwner   = hWnd;
	ofn.lpstrFilter = filter;
	ofn.lpstrFile   = fileName;
	ofn.nMaxFile    = MAX_PATH;
	ofn.Flags       = OFN_FILEMUSTEXIST | OFN_HIDEREADONLY;
	ofn.lpstrDefExt = fileExt;

	result = GetOpenFileName( &ofn );
	if( result == FALSE ) {
		errNo = CommDlgExtendedError() % 16;
		if( errNo != 0 )
			return( true );
		else {
			*cancel = true;
			return( false );
		}
	}

	return( false );
}


bool  CbvRead::
OpenFileMain( const char fileName[ MAX_PATH ] )
{
	FILE*  fp;

	fp = fopen( fileName, "rt" );
	if( fp == NULL )
		return( true );

	CbvReadCore::SetFP( fp );

	return( false );
}


bool  CbvRead::
ReadMain( void )
{
	bool  result;

	rewind( CbvReadCore::GetFP() );
	result = ReadHeader();
	if( result == true )
		return( true );
	result = ReadData( false );
	if( result == true )
		return( true );

	if( AllocVertex() == true )
		return( true );
	if( AllocRadial() == true )
		return( true );

	rewind( CbvReadCore::GetFP() );
	result = ReadHeader();
	if( result == true )
		return( true );
	result = ReadData( true );
	if( result == true )
		return( true );

	return( false );
}


bool  CbvRead::
ReadHeader( void )
{
	bool    result, equ;
	double  version;

	result = CbvReadCore::ReadCmpStr1( fileExt, &equ );
	if( result == true || equ == false )
		return( true );

	result = CbvReadCore::ReadFloat( &version );
	if( result == true || version != fileVersion )
		return( true );

	return( false );
}


bool  CbvRead::
ReadData( const bool setting )
{
	enum {
		typeBegin,
		typeEnd,
		numType
	};
	ulong      cntVertex, cntRadial, numRadial;
	VertexRec  vertex;
	bool       result, equ;
	char*      str[ numType ];
	short      no;

	result = CbvReadCore::ReadCmpStr1( CbvReadCore::BeginStr(), &equ );
	if( result == true || equ == false )
		return( true );

	cntVertex = cntRadial = 0;
	str[ typeBegin ] = ( char* )CbvReadCore::BeginStr();
	str[ typeEnd   ] = ( char* )CbvReadCore::EndStr();
	for(;;) {
		result = CbvReadCore::ReadCmpStrN( numType, ( const char** )str, &equ, &no );
		if( result == true || equ == false )
			return( true );
		else if( no == typeEnd )
			break;

		result = ReadVertexHeader( &vertex );
		if( result == true )
			return( true );

		result = ReadVertexRadial( setting, cntRadial, &numRadial );
		if( result == true )
			return( true );

		vertex.numRadial  = ( short )numRadial;
		vertex.startIndex = cntRadial;
		
		SetVertexAry( setting, cntVertex, &vertex );
		cntVertex++;
		cntRadial += numRadial;

		result = CbvReadCore::ReadCmpStr1( CbvReadCore::EndStr(), &equ );
		if( result == true || equ == false )
			return( true );
	}

	numVertexAry = cntVertex;
	numRadialAry = cntRadial;
	if( numVertexAry < CbxDB::minNumVertex || numRadialAry < CbxDB::minNumEdge*2 )
		return( true );

	return( false );
}


bool  CbvRead::
ReadVertexHeader( VertexRec* vertex )
{
	bool  result;

	result = CbvReadCore::ReadFloat3( vertex->vertexPos );
	if( result == true )
		return( true );

	result = CbvReadCore::ReadBool( loopStr[ 0 ], loopStr[ 1 ], &vertex->loopFlag );
	if( result == true )
		return( true );

	return( false );
}


bool  CbvRead::
ReadVertexRadial( const bool setting, const ulong startIndex, ulong* numAry )
{
	bool  result;

	result = CbvReadCore::ReadUIntAry( setting, startIndex, SetRadialAry, numAry );
	if( result == true || *numAry < CbxDB::minNumRadial )
		return( true );

	return( false );
}


void  CbvRead::
SetRadialAry( const bool setting, const ulong no, const ulong edgeNo )
{
	if( setting == true )
		radialAry[ no ].edgeNo = edgeNo;
}


void  CbvRead::
SetVertexAry( const bool setting, const ulong no, const VertexRec* vertex )
{
	if( setting == true )
		vertexAry[ no ] = *vertex;
}


bool  CbvRead::
AllocVertex( void )
{
	vertexAry = new VertexRec[ numVertexAry ];
	if( vertexAry == 0 )
		return( true );

	return( false );
}


bool  CbvRead::
AllocRadial( void )
{
	radialAry = new RadialRec[ numRadialAry ];
	if( radialAry == 0 )
		return( true );

	return( false );
}


void  CbvRead::
FreeData( void )
{
	FreeVertex();
	FreeRadial();
}


void  CbvRead::
FreeVertex( void )
{
	delete [] vertexAry;
}


void  CbvRead::
FreeRadial( void )
{
	delete [] radialAry;
}
