/* bvReadCore.cpp */

#include "../BeziexView.h"


FILE*       CbvReadCore::fileStream        = NULL;
const char  CbvReadCore::beginStr[ 16 ]    = "{";
const char  CbvReadCore::endStr[ 16 ]      = "}";
const char  CbvReadCore::nonPunctStr[ 32 ] = "-0-9+.A-Za-z_";
const char  CbvReadCore::commentChr[ 2 ]   = ";";




FILE*  CbvReadCore::
GetFP( void )
{
	return( fileStream );
}


void  CbvReadCore::
SetFP( FILE* const fp )
{
	fileStream = fp;
}


const char*  CbvReadCore::
BeginStr( void )
{
	return( beginStr );
}


const char*  CbvReadCore::
EndStr( void )
{
	return( endStr );
}


bool  CbvReadCore::
ReadStr( char str[ lenStrBuf ] )
{
	bool  result, punct, chk;

	result = ReadSelect( str, &punct );
	if( result == true )
		return( true );
	else if( punct == true )
		return( false );

	ChkName( str, &chk );
	if( chk == false )
		return( true );

	return( false );
}


bool  CbvReadCore::
ReadSelect( char str[ lenStrBuf ], bool* punct )
{
	bool  result, overflow;
	char  chr;

	ReadComment();

	result = ReadPunct( &chr );
	if( result == false ) {
		*punct = true;
		str[ 0 ] = chr;		str[ 1 ] = 0;
		return( false );
	}

	result = ReadNonPunct( str, &overflow );
	if( result == false ) {
		*punct = false;
		return( false );
	}

	return( true );
}


void  CbvReadCore::
ReadComment( void )
{
	int   result;
	char  buf[ 2 ], format[ 256 ];

	strcpy( format, "\x20%1[" );
	strcat( format, commentChr );
	strcat( format, "]" );

	for(;;) {
		result = fscanf( fileStream, format, buf );
		if( result != 1 )
			return;

		fscanf( fileStream, "%*[^\n]" );
	}
}

	
bool  CbvReadCore::
ReadPunct( char* chr )
{
	char  buf[ 2 ], format[ 256 ];
	int   result;

	strcpy( format, "\x20%1[^" );
	strcat( format, nonPunctStr );
	strcat( format, "]" );

	result = fscanf( fileStream, format, buf );
	if( result != 1 )
		return( true );

	*chr = buf[ 0 ];

	return( false );
}


bool  CbvReadCore::
ReadNonPunct( char str[ lenStrBuf ], bool* overflow )
{
	char  buf[ lenStrBuf + 1 ], format[ 256 ];
	int   result;

	*overflow = false;

	strcpy( format, "\x20%" );
	sprintf( &format[ strlen( format ) ], "%d", lenStrBuf );
	strcat( format, "[" );
	strcat( format, nonPunctStr );
	strcat( format, "]" );

	result = fscanf( fileStream, format, buf );
	if( result != 1 || strlen( buf ) == lenStrBuf ) {
		if( strlen( buf ) == lenStrBuf )
			*overflow = true;
		return( true );
	}

	_strlwr( buf );
	strcpy( str, buf );

	return( false );
}


void  CbvReadCore::
ChkName( const char str[ lenStrBuf ], bool* chk )
{
	*chk = false;
	if( isalpha( str[ 0 ] ) != 0 )
		*chk = true;
}


bool  CbvReadCore::
ReadFloat( double* value )
{
	bool   result, chkStr;
	void*  data;

	result = ReadFloatStr( &chkStr, &data );
	if( result == true || chkStr == true )
		return( true );

	*value = GetFloat( data );

	return( false );
}


double  CbvReadCore::
GetFloat( const void* data )
{
	return( *( double* )data );
}


bool  CbvReadCore::
ReadUInt( ulong* value )
{
	bool   result, chkStr;
	void*  data;

	result = ReadUIntStr( &chkStr, &data );
	if( result == true || chkStr == true )
		return( true );

	*value = GetUInt( data );

	return( false );
}


ulong  CbvReadCore::
GetUInt( const void* data )
{
	return( *( ulong* )data );
}


bool  CbvReadCore::
ReadBool( const char* falseStr, const char* trueStr, bool* boolean )
{
	enum {
		typeFalse,
		typeTrue,
		numType
	};
	char*  cmpStr[ numType ];
	bool   result, equ;
	short  no;

	cmpStr[ typeFalse ] = ( char* )falseStr;
	cmpStr[ typeTrue  ] = ( char* )trueStr;
	result = ReadCmpStrN( numType, ( const char** )cmpStr, &equ, &no );
	if( result == true || equ == false )
		return( true );

	if( no == typeFalse )
		*boolean = false;
	else
		*boolean = true;

	return( false );
}


bool  CbvReadCore::
ReadFloat3( double value[ 3 ] )
{
	short  i;
	bool   result, equ;

	result = ReadCmpStr1( beginStr, &equ );
	if( result == true || equ == false )
		return( true );

	for( i=0; i<3; i++ ) {
		result = ReadFloat( &value[ i ] );
		if( result == true )
			return( true );
	}

	result = ReadCmpStr1( endStr, &equ );
	if( result == true || equ == false )
		return( true );

	return( false );
}


bool  CbvReadCore::
ReadUIntAry( const bool setting, const ulong startIndex,
	void ( *SetAry )( const bool setting, const ulong no, const ulong data ), ulong* numAry )
{
	ulong  cnt;
	void*  data;
	bool   result, equ, chkStr;

	result = ReadCmpStr1( beginStr, &equ );
	if( result == true || equ == false )
		return( true );

	cnt = 0;
	for(;;) {
		result = ReadUIntCmpStr( endStr, &chkStr, &data );
		if( result == true )
			return( true );

		if( chkStr == true ) {
			if( GetBool( data ) == true )
				break;
			else
				return( true );
		}

		( *SetAry )( setting, cnt + startIndex, GetUInt( data ) );
		cnt++;
	}

	*numAry = cnt;

	return( false );
}


bool  CbvReadCore::
GetBool( const void* data )
{
	return( *( bool* )data );
}


bool  CbvReadCore::
ReadFloatStr( bool* chkStr, void** data )
{
	bool  result, punct, chk;
	static char    strResult[ lenStrBuf ];
	static double  floatResult;

	result = ReadSelect( strResult, &punct );
	if( result == true )
		return( true );
	else if( punct == true ) {
		*chkStr = true;
		*data   = strResult;
		return( false );
	}

	result = ConvFloat( strResult, &chk, &floatResult );
	if( result == true )
		return( true );

	if( chk == true ) {
		*chkStr = false;
		*data   = &floatResult;
	}
	else {
		*chkStr = true;
		*data   = strResult;
	}

	return( false );
}


bool  CbvReadCore::
ConvFloat( const char str[ lenStrBuf ], bool* chk, double* value )
{
	int   result;
	char  chr;

	if( isalpha( str[ 0 ] ) != 0 ) {
		*chk = false;
		return( false );
	}

	result = sscanf( str, "%lg%c", value, &chr );
	if( result != 1 )
		return( true );

	*chk = true;

	return( false );
}


bool  CbvReadCore::
ReadUIntStr( bool* chkStr, void** data )
{
	bool  result, punct, chk;
	static char   strResult[ lenStrBuf ];
	static ulong  uintResult;

	result = ReadSelect( strResult, &punct );
	if( result == true )
		return( true );
	else if( punct == true ) {
		*chkStr = true;
		*data   = strResult;
		return( false );
	}

	result = ConvUInt( strResult, &chk, &uintResult );
	if( result == true )
		return( true );

	if( chk == true ) {
		*chkStr = false;
		*data   = &uintResult;
	}
	else {
		*chkStr = true;
		*data   = strResult;
	}

	return( false );
}


bool  CbvReadCore::
ConvUInt( const char str[ lenStrBuf ], bool* chk, ulong* value )
{
	int   result;
	char  chr;

	if( isalpha( str[ 0 ] ) != 0 ) {
		*chk = false;
		return( false );
	}

	result = sscanf( str, "%lu%c", value, &chr );
	if( result != 1 )
		return( true );

	*chk = true;

	return( false );
}


bool  CbvReadCore::
ReadCmpStr1( const char cmpStr[ lenStrBuf ], bool* equ )
{
	char*  cmpStr2[ 1 ];
	short  no;

	cmpStr2[ 0 ] = ( char* )cmpStr;
	return( ReadCmpStrN( 1, ( const char** )cmpStr2, equ, &no ) );
}


bool  CbvReadCore::
ReadCmpStrN( const short numStr, const char** cmpStr, bool* equ, short* no )
{
	char   str[ lenStrBuf ];
	bool   result;
	short  i;

	result = ReadStr( str );
	if( result == true )
		return( true );

	*equ = false;
	for( i=0; i<numStr; i++ ) {
		if( strcmp( cmpStr[ i ], str ) == 0 ) {
			*equ = true;
			*no  = i;
			break;
		}
	}

	return( false );
}


bool  CbvReadCore::
ReadUIntCmpStr( const char cmpStr[], bool* chkStr, void** data )
{
	char  str[ lenStrBuf ];
	bool  result;
	static bool  equ;

	result = ReadUIntStr( chkStr, data );
	if( result == true )
		return( true );

	if( *chkStr == false )
		return( false );

	CopyStr( *data, str );
	equ = false;
	if( strcmp( cmpStr, str ) == 0 )
		equ = true;

	*data = &equ;

	return( false );
}


char*  CbvReadCore::
CopyStr( const void* data, char* str )
{
	return( strcpy( str, ( char* )data ) );
}
