/* bvMain.cpp */

#include "../BeziexView.h"


HINSTANCE  CbvMain::hInst;
HPALETTE   CbvMain::hPalette	= NULL;


int  WINAPI
WinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow )
{
	return( CbvMain::Main( hInstance, hPrevInstance, lpCmdLine, nCmdShow ) );
}


int  CbvMain::
Main( HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow )
{
	MSG       msg;
	WNDCLASS  wc;
	HWND      hWnd;
	char      title[ 256 ];
	RECT      rect;
	static const char  className[] = "BeziexView";

    wc.style         = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc   = CbvMain::WndProc;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = 0;
    wc.hInstance     = hInstance;
    wc.hIcon         = LoadIcon( hInstance, IDI_APPLICATION );
    wc.hCursor       = LoadCursor( NULL, IDC_ARROW );
    wc.hbrBackground = ( HBRUSH )( COLOR_WINDOW+1 );
    wc.lpszMenuName  = MAKEINTRESOURCE( IDR_MENU );
    wc.lpszClassName = className;

    RegisterClass( &wc );

	hInst = hInstance;

	LoadString( hInst, IDS_WINTITLE, title, 256-1 );
	rect.right  = winInitSizeX + GetNonClientWidth();
	rect.bottom = winInitSizeY + GetNonClientHeight();
	rect.left   = ( GetSystemMetrics( SM_CXSCREEN ) - rect.right  ) / 2;
	rect.top    = ( GetSystemMetrics( SM_CYSCREEN ) - rect.bottom ) / 2;

    hWnd = CreateWindow(
        className,
        title,
        WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS,
        rect.left, rect.top, rect.right, rect.bottom,
        NULL,
        NULL,
        hInstance,
        NULL );

    if( hWnd == NULL )
        return( FALSE );

	InitWin( hWnd );
	ShowWindow( hWnd, nCmdShow );

	while( GetMessage( &msg, NULL, 0, 0 ) == TRUE ) {
		TranslateMessage( &msg );
		DispatchMessage( &msg );
	}

	return( msg.wParam );
}



LRESULT CALLBACK  CbvMain::
WndProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	PAINTSTRUCT  ps;

    switch( uMsg ) {
		case WM_CREATE:
			CreateProc( hWnd );
			break;

        case WM_COMMAND:
			CbvMenu::Proc( hWnd, LOWORD( wParam ) );
            break;

        case WM_DESTROY:
			FinishProc();
            break;

		case WM_PALETTECHANGED:
		case WM_QUERYNEWPALETTE:
			PaletteProc( uMsg, ( HWND )wParam );
			break;

		case WM_PAINT:
			BeginPaint( hWnd, &ps );

			if( hPalette != NULL ) {
				SelectPalette( ps.hdc, hPalette, FALSE );
				RealizePalette( ps.hdc );
			}
			CbvPaint::Paint( hWnd, ps.hdc );

			EndPaint( hWnd, &ps );
            break;

        default:
            return( DefWindowProc( hWnd, uMsg, wParam, lParam ) );
    }

    return( 0L );
}


void  CbvMain::
InitWin( HWND hWnd )
{
	bool  needPalette;

	needPalette = CbvGL::Init( hWnd );
	if( needPalette == true )
		InitPalette();

	CbvGL::SetDirStd();
}


void  CbvMain::
FinishProc( void )
{
	CbxDB::FreeData();
	CbvPolygon::FreeData();
	CbvEdge::FreeData();
	CbvGL::Finish();

	if( hPalette != NULL )
		DeleteObject( hPalette );

    PostQuitMessage ( 0 );
}


void  CbvMain::
InitPalette( void )
{
	short       i, rNo, gNo, bNo;
	LOGPALETTE  pal;

	pal.palVersion    = 0x300;
	pal.palNumEntries = 256;

	for( i=0; i<256; i++ ) {
		rNo = i & 0x7;
		gNo = ( i >> 3 ) & 0x7;
		bNo = ( i >> 6 ) & 0x3;
		pal.palPalEntry[ i ].peRed   = ( 256 - 1 ) * rNo / ( 8 - 1 );
		pal.palPalEntry[ i ].peGreen = ( 256 - 1 ) * gNo / ( 8 - 1 );
		pal.palPalEntry[ i ].peBlue  = ( 256 - 1 ) * bNo / ( 4 - 1 );
		pal.palPalEntry[ i ].peFlags = NULL;
	}

	hPalette = CreatePalette( &pal );
}


void  CbvMain::
PaletteProc( const UINT uMsg, HWND hFocusWnd )
{
	HWND  hWnd;

	hWnd = GetActiveWindow();

	switch( uMsg ) {
		case WM_PALETTECHANGED:
			if( hFocusWnd != hWnd )
				InvalidateRect( hWnd, NULL, FALSE );
			break;

		case WM_QUERYNEWPALETTE:
			InvalidateRect( hWnd, NULL, FALSE );
			break;
	}
}


int  CbvMain::
GetNonClientWidth( void )
{
	return( GetSystemMetrics( SM_CXSIZEFRAME ) * 2 );
}


int  CbvMain::
GetNonClientHeight( void )
{
	return( GetSystemMetrics( SM_CYSIZEFRAME ) * 2
		+ GetSystemMetrics( SM_CYCAPTION ) + GetSystemMetrics( SM_CYMENU ) );
}


void   CbvMain::
CreateProc( HWND hWnd )
{
	bool  result, cancel;

	result = CbvRead::Proc( hWnd, &cancel );
	if( result == true || cancel == true ) {
		DestroyWindow( hWnd );
		return;
	}

	result = CbvPolygon::Update();
	if( result == true ) {
		DestroyWindow( hWnd );
		return;
	}

	result = CbvEdge::Update();
	if( result == true ) {
		DestroyWindow( hWnd );
		return;
	}
}
