/* bvGL.cpp */

#include <stdlib.h>
#include <GL/glut.h>
#include "../BeziexView.h"


const double  CbvGL::minPosInSpace[ 3 ] = { -3.0, -3.0, -3.0 };
const double  CbvGL::maxPosInSpace[ 3 ] = {  3.0,  3.0,  3.0 };

		
HGLRC  CbvGL::hRC;

double  CbvGL::currentMat[ 4 ][ 4 ] = {
	{ 1.0, 0.0, 0.0, 0.0 },
	{ 0.0, 1.0, 0.0, 0.0 },
	{ 0.0, 0.0, 1.0, 0.0 },
	{ 0.0, 0.0, 0.0, 1.0 }
};



bool  CbvGL::
Init( HWND hWnd )
{
	HDC  hDC;
	int  pixelFormat;
	PIXELFORMATDESCRIPTOR  pfd = {
		sizeof( PIXELFORMATDESCRIPTOR ),
		1,
		PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL,
		PFD_TYPE_RGBA,
		24,
		0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0,
		32,
		0, 0,
		PFD_MAIN_PLANE,
		0,
		0, 0, 0
	};

	hDC = GetDC( hWnd );

	pixelFormat = ChoosePixelFormat( hDC, &pfd );
	SetPixelFormat( hDC, pixelFormat, &pfd );
	DescribePixelFormat( hDC, pixelFormat, sizeof( pfd ), &pfd );

	hRC = wglCreateContext( hDC );

	ReleaseDC( hWnd, hDC );

	if( ( pfd.dwFlags & PFD_NEED_PALETTE ) != 0 )
		return( true );

	return( false );
}


void  CbvGL::
Finish( void )
{
	wglDeleteContext( hRC );
}


void  CbvGL::
Paint( HWND hWnd, HDC hDC )
{
	wglMakeCurrent( hDC, hRC );

	glClearColor( 1.0, 1.0, 1.0, 1.0 );
	glClear( GL_COLOR_BUFFER_BIT );

	DrawWire( hWnd );
	DrawSurf( hWnd );

	glFinish();
	wglMakeCurrent( hDC, NULL );
}


void  CbvGL::
DrawWire( HWND hWnd )
{
	RECT  clientRect, rect;

	GetClientRect( hWnd, &clientRect );
	rect.top    = 0;
	rect.bottom = clientRect.bottom;

	rect.left   = 0;
	rect.right  = clientRect.right / 2;

	SetView( hWnd, &rect );
	SetObj();

	CbvGLwire::Draw( hWnd );
	CbvGLelement::Draw();
}


void  CbvGL::
DrawSurf( HWND hWnd )
{
	RECT  clientRect, rect;

	GetClientRect( hWnd, &clientRect );
	rect.top    = 0;
	rect.bottom = clientRect.bottom;

	rect.left   = clientRect.right / 2 + 1;
	rect.right  = clientRect.right - rect.left;

	SetView( hWnd, &rect );
	SetObj();

	CbvGLsurf::Draw( hWnd );
	CbvGLmesh::Draw();
}


void  CbvGL::
SetView( HWND hWnd, const RECT* rect )
{
	glClearColor( 1.0, 1.0, 1.0, 1.0 );
	glShadeModel( GL_SMOOTH );
	glClear( GL_DEPTH_BUFFER_BIT );

	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	SetPerspective( hWnd );
	glViewport( rect->left, rect->top, rect->right, rect->bottom );

	gluLookAt( 0.0, 0.0, 4.5, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0 );
}


void  CbvGL::
SetPerspective( HWND hWnd )
{
	RECT    rect;
	double  sizeX, sizeY;

	GetClientRect( hWnd, &rect );
	sizeX = rect.right / 2.0;
	sizeY = rect.bottom;

	if( sizeX >= sizeY )
		glFrustum( -sizeX/sizeY, sizeX/sizeY, -1.0, 1.0, 2.0, 7.0 );
	else
		glFrustum( -1.0, 1.0, -sizeY/sizeX, sizeY/sizeX, 2.0, 7.0 );
}


void  CbvGL::
SetObj( void )
{
	SetLight();
	SetMatrix();
}


void  CbvGL::
SetLight( void )
{
	GLfloat  data[ 4 ];

	data[ 0 ] = 0.3f;	data[ 1 ] = 0.3f;	data[ 2 ] = 0.3f;	data[ 3 ] = 0.0;	
	glLightfv( GL_LIGHT0, GL_AMBIENT, data );

	data[ 0 ] = 0.7f;	data[ 1 ] = 0.7f;	data[ 2 ] = 0.7f;	data[ 3 ] = 0.0;	
	glLightfv( GL_LIGHT0, GL_DIFFUSE, data );

	data[ 0 ] = 1.0;	data[ 1 ] = 1.0;	data[ 2 ] = 1.0;	data[ 3 ] = 0.0;	
	glLightfv( GL_LIGHT0, GL_SPECULAR, data );

	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();

	data[ 0 ] = 2.0;	data[ 1 ] = 0.5;	data[ 2 ] = 3.0;	data[ 3 ] = 1.0;
	glLightfv( GL_LIGHT0, GL_POSITION, data );

	glLightModeli( GL_LIGHT_MODEL_TWO_SIDE, 1 );
	glEnable( GL_LIGHT0 );
}


void  CbvGL::
SetMatrix( void )
{
	short     u, v;
	GLdouble  m[ 16 ];

	for( v=0; v<4; v++ ) {
		for( u=0; u<4; u++ )
			m[ u*4+v ] = currentMat[ u ][ v ];
	}

	glMatrixMode( GL_MODELVIEW );
	glLoadMatrixd( m );
}


void  CbvGL::
SetDirStd( void )
{
	double  mx[ 4 ][ 4 ], my[ 4 ][ 4 ];

	CbxMath::RotateX4x4(  45.0, mx );
	CbxMath::RotateY4x4( -45.0, my );
	CbxMath::Multiply4x4( mx, my, currentMat );
}


void  CbvGL::
SetDirTop( void )
{
	CbxMath::RotateX4x4( 90.0, currentMat );
}


void  CbvGL::
SetDirFront( void )
{
	CbxMath::MakeIdentity4x4( currentMat );
}


void  CbvGL::
SetDirSide( void )
{
	CbxMath::RotateY4x4( -90.0, currentMat );
}


void  CbvGL::
RotSurfX( const double deg )
{
	double  m[ 4 ][ 4 ];

	CbxMath::RotateX4x4( deg, m );
	CbxMath::Multiply4x4( m, currentMat, currentMat );
}


void  CbvGL::
RotSurfY( const double deg )
{
	double  m[ 4 ][ 4 ];

	CbxMath::RotateY4x4( deg, m );
	CbxMath::Multiply4x4( m, currentMat, currentMat );
}


void  CbvGL::
RotSurfZ( const double deg )
{
	double  m[ 4 ][ 4 ];

	CbxMath::RotateZ4x4( deg, m );
	CbxMath::Multiply4x4( m, currentMat, currentMat );
}


void  CbvGL::
GetCurrentMat( double mat[ 4 ][ 4 ] )
{
	short  u, v;

	for( v=0; v<4; v++ ) {
		for( u=0; u<4; u++ )
			mat[ u ][ v ] = currentMat[ u ][ v ];
	}
}


bool  CbvGL::
ErrChkVertexPos( const double vertexPos[ 3 ] )
{
	short  i;

	for( i=0; i<3; i++ ) {
		if( minPosInSpace[ i ] > vertexPos[ i ] || vertexPos[ i ] > maxPosInSpace[ i ] )
			return( true );
	}

	return( false );
}
