/* bvGLelement.cpp */

#include "../BeziexView.h"

const double  CbvGLelement::lenNormalOnVertex   =  0.4;
const double  CbvGLelement::lenNormalInCtlPlane =  0.4;
const double  CbvGLelement::scaleDownRatio      =  0.9;



void  CbvGLelement::
Draw( void )
{
	glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
	glEnable( GL_DEPTH_TEST );

	glDisable( GL_LIGHTING );

	SetMaterial();
	DrawElement();
}


void  CbvGLelement::
SetMaterial( void )
{
	GLfloat  data[ 4 ];

	data[ 0 ] = 0.2f;	data[ 1 ] = 0.2f;	data[ 2 ] = 0.15f;	data[ 3 ] = 0.0;	
	glMaterialfv( GL_FRONT, GL_AMBIENT, data );

	data[ 0 ] = 0.1f;	data[ 1 ] = 0.1f;	data[ 2 ] = 0.2f;	data[ 3 ] = 0.0;	
	glMaterialfv( GL_BACK, GL_AMBIENT, data );

	data[ 0 ] = 0.8f;	data[ 1 ] = 0.8f;	data[ 2 ] = 0.6f;	data[ 3 ] = 0.0;	
	glMaterialfv( GL_FRONT, GL_DIFFUSE, data );

	data[ 0 ] = 0.4f;	data[ 1 ] = 0.4f;	data[ 2 ] = 0.8f;	data[ 3 ] = 0.0;	
	glMaterialfv( GL_BACK, GL_DIFFUSE, data );

	data[ 0 ] = 0.0;	data[ 1 ] = 0.0;	data[ 2 ] = 0.0;	data[ 3 ] = 0.0;	
	glMaterialfv( GL_FRONT_AND_BACK, GL_SPECULAR, data );

	data[ 0 ] = 0.0;	data[ 1 ] = 0.0;	data[ 2 ] = 0.0;	data[ 3 ] = 0.0;	
	glMaterialfv( GL_FRONT_AND_BACK, GL_EMISSION, data );

	data[ 0 ] = 70.0;
	glMaterialfv( GL_FRONT_AND_BACK, GL_SHININESS, data );
}


void  CbvGLelement::
DrawElement( void )
{
	if( DrawLadder() == true ) {
		CbvMessage::Error( IDS_CANNOTCONTINUE );
		return;
	}

	if( DrawHandle() == true ) {
		CbvMessage::Error( IDS_CANNOTCONTINUE );
		return;
	}

	if( DrawCtlVec() == true ) {
		CbvMessage::Error( IDS_CANNOTCONTINUE );
		return;
	}
}


bool  CbvGLelement::
DrawLadder( void )
{
	PosInfo  pos;
	bool     result, dispChk;
	short    i;

	result = GetElementPos( &pos );
	if( result == true )
		return( true );

	dispChk = true;
	for( i=0; i<2; i++ ) {
		if( pos.dt[ i ].dispChkForHandle == false ||
			pos.dt[ i ].dispChkForCtl    == false ||
			pos.dt[ i ].dispChkForWing   == false )
		{
			dispChk = false;
			break;
		}
	}

	if( dispChk == true ) {
		DrawElementRect( &pos );
		DrawElementLine( &pos );
	}

	return( false );
}


bool  CbvGLelement::
GetElementPos( PosInfo* pos )
{
	WingHandleInfo  dt[ 2 ];
	short           i;

	GetVertexOrg( pos->dt[ 0 ].vertexPos );
	if( GetVertexRadial( pos->dt[ 1 ].vertexPos ) == true )
		return( true );
	if( GetHandleOrg( &pos->dt[ 0 ].dispChkForHandle, pos->dt[ 0 ].handlePos ) == true )
		return( true );
	if( GetHandleRadial( &pos->dt[ 1 ].dispChkForHandle, pos->dt[ 1 ].handlePos ) == true )
		return( true );
	if( GetCtlVecOrg( &pos->dt[ 0 ].dispChkForCtl, pos->dt[ 0 ].ctlPos ) == true )
		return( true );
	if( GetCtlVecRadial( &pos->dt[ 1 ].dispChkForCtl, pos->dt[ 1 ].ctlPos ) == true )
		return( true );
	if( GetWingHandleOrg( &pos->dt[ 0 ].dispChkForWing, &dt[ 0 ] ) == true )
		return( true );
	if( GetWingHandleRadial( &pos->dt[ 1 ].dispChkForWing, &dt[ 1 ] ) == true )
		return( true );

	for( i=0; i<2; i++ ) {
		CbxMath::Copy3( dt[ i ].handlePosOrg,    pos->dt[ i ].wingPos  );
		pos->dt[ i ].extFlag = dt[ i ].extFlag;
		CbxMath::Copy3( dt[ i ].handlePosRadial, pos->dt[ i ].wingPos2 );
		CbxMath::Copy3( dt[ i ].vertexPosRadial, pos->dt[ i ].wingPos3 );
	}

	return( false );
}


void  CbvGLelement::
DrawElementRect( const PosInfo* pos )
{
	glEnable( GL_LIGHTING );

	DrawRect( pos->dt[ 0 ].vertexPos, pos->dt[ 0 ].wingPos, pos->dt[ 0 ].ctlPos,  pos->dt[ 0 ].handlePos );
	DrawRect( pos->dt[ 0 ].handlePos, pos->dt[ 0 ].ctlPos,  pos->dt[ 1 ].ctlPos,  pos->dt[ 1 ].handlePos );
	DrawRect( pos->dt[ 1 ].handlePos, pos->dt[ 1 ].ctlPos,  pos->dt[ 1 ].wingPos, pos->dt[ 1 ].vertexPos );
}


void  CbvGLelement::
DrawElementLine( const PosInfo* pos )
{
	short  i;

	glDisable( GL_LIGHTING );

	glColor3f( 0.8f, 0.0f, 1.0f );
	glLineWidth( 1.0 );

	DrawLine( pos->dt[ 0 ].handlePos, pos->dt[ 1 ].handlePos );
	DrawLine( pos->dt[ 1 ].handlePos, pos->dt[ 1 ].vertexPos );
 
	DrawLine( pos->dt[ 0 ].vertexPos, pos->dt[ 0 ].wingPos   );
	DrawLine( pos->dt[ 1 ].handlePos, pos->dt[ 1 ].ctlPos    );
	DrawLine( pos->dt[ 1 ].vertexPos, pos->dt[ 1 ].wingPos   );

	DrawLine( pos->dt[ 0 ].wingPos,   pos->dt[ 0 ].ctlPos    );
	DrawLine( pos->dt[ 0 ].ctlPos,    pos->dt[ 1 ].ctlPos    );
	DrawLine( pos->dt[ 1 ].ctlPos,    pos->dt[ 1 ].wingPos   );

	for( i=0; i<2; i++ ) {
		if( pos->dt[ i ].extFlag == true ) {
			DrawLine( pos->dt[ i ].wingPos,  pos->dt[ i ].wingPos2 );
			DrawLine( pos->dt[ i ].wingPos2, pos->dt[ i ].wingPos3 );
		}
	}
}


void  CbvGLelement::
DrawRect( const double p0[ 3 ], const double p1[ 3 ], const double p2[ 3 ], const double p3[ 3 ] )
{
	double  g[ 3 ], p[ 4 ][ 3 ];
	short   i;

	CbxMath::Copy3( p0, p[ 0 ] );
	CbxMath::Copy3( p1, p[ 1 ] );
	CbxMath::Copy3( p2, p[ 2 ] );
	CbxMath::Copy3( p3, p[ 3 ] );

	CbxMath::Copy3( CbxMath::zero3, g );
	for( i=0; i<4; i++ )
		CbxMath::Add3( g, p[ i ], g );
	CbxMath::Scale3( g, 1/4.0, g );
	for( i=0; i<4; i++ )
		CbxMath::Ratio3( g, p[ i ], scaleDownRatio, p[ i ] );

	DrawTri( p[ 0 ], p[ 1 ], g );
	DrawTri( p[ 1 ], p[ 2 ], g );
	DrawTri( p[ 2 ], p[ 3 ], g );
	DrawTri( p[ 3 ], p[ 0 ], g );
}


void  CbvGLelement::
DrawTri( const double p0[ 3 ], const double p1[ 3 ], const double p2[ 3 ] )
{
	double  vec[ 2 ][ 3 ], normal[ 3 ];

	CbxMath::Subtract3( p1, p0, vec[ 0 ] );
	CbxMath::Subtract3( p2, p0, vec[ 1 ] );
	CbxMath::Cross3( vec[ 0 ], vec[ 1 ], normal );
	CbxMath::Normalize3( normal, CbxMath::zero3, normal );

	glBegin( GL_TRIANGLES );
		glNormal3f( ( GLfloat )normal[ 0 ], ( GLfloat )normal[ 1 ], ( GLfloat )normal[ 2 ] );

		glVertex3f( ( GLfloat )p0[ 0 ], ( GLfloat )p0[ 1 ], ( GLfloat )p0[ 2 ] );
		glVertex3f( ( GLfloat )p1[ 0 ], ( GLfloat )p1[ 1 ], ( GLfloat )p1[ 2 ] );
		glVertex3f( ( GLfloat )p2[ 0 ], ( GLfloat )p2[ 1 ], ( GLfloat )p2[ 2 ] );
	glEnd();
}


bool  CbvGLelement::
DrawHandle( void )
{
	double  vertexPos[ 3 ], handlePos[ 3 ], mid[ 3 ];
	bool    dispChk;

	GetVertexOrg( vertexPos );
	if( GetHandleOrg( &dispChk, handlePos ) == true )
		return( true );
	if( dispChk == false )
		return( false );

	glColor3f( 1.0f, 0.0f, 0.0f );

	CbxMath::Ratio3( vertexPos, handlePos, 0.9, mid );

	glLineWidth( 2.0 );
	DrawLine( vertexPos, mid );

	glLineWidth( 4.0 );
	DrawLine( mid, handlePos );

	return( false );
}


bool  CbvGLelement::
DrawCtlVec( void )
{
	double  handlePos[ 3 ], ctlPos[ 3 ], mid[ 3 ];
	bool    dispChk;
	
	if( GetHandleOrg( &dispChk, handlePos ) == true )
		return( true );
	if( dispChk == false )
		return( false );

	if( GetCtlVecOrg( &dispChk, ctlPos ) == true )
		return( true );
	if( dispChk == false )
		return( false );

	glColor3f( 0.0f, 0.0f, 1.0f );

	CbxMath::Ratio3( handlePos, ctlPos, 0.9, mid );

	glLineWidth( 2.0 );
	DrawLine( handlePos, mid );

	glLineWidth( 4.0 );
	DrawLine( mid, ctlPos );

	return( false );
}


void  CbvGLelement::
DrawLine( const double p0[ 3 ], const double p1[ 3 ] )
{
	glBegin( GL_LINES );
		glVertex3f( ( GLfloat )p0[ 0 ], ( GLfloat )p0[ 1 ], ( GLfloat )p0[ 2 ] );
		glVertex3f( ( GLfloat )p1[ 0 ], ( GLfloat )p1[ 1 ], ( GLfloat )p1[ 2 ] );
	glEnd();
}


void  CbvGLelement::
GetVertexOrg( double vertexPos[ 3 ] )
{
	CbxDB::VertexRec  vertex;

	CbxDB::LoadVertexRec( CbvDlogSelVer::Data(), &vertex );
	CbxMath::Copy3( vertex.vertexPos, vertexPos );
}


bool  CbvGLelement::
GetVertexRadial( double vertexPos[ 3 ] )
{
	bool   result;
	ulong  edgeNo, vertexNo;
	short  beginEndNo;
	CbxDB::VertexRec  vertex;
	CbxDB::EdgeRec    edge;

	result = CbxDB::VertexRadialToEdge( CbvDlogSelVer::Data(), CbvDlogSelHand::Data(), &edgeNo, &beginEndNo );
	if( result == true )
		return( true );

	CbxDB::LoadEdgeRec( edgeNo, &edge );
	vertexNo = edge.vertex[ 1 - beginEndNo ].vertexNo;

	CbxDB::LoadVertexRec( vertexNo, &vertex );
	CbxMath::Copy3( vertex.vertexPos, vertexPos );

	return( false );
}


bool  CbvGLelement::
GetHandleOrg( bool* dispChk, double handlePos[ 3 ] )
{
	double  handleVec[ 3 ], vertexPos[ 3 ];

	if( GetHandleSub( false, dispChk, vertexPos, handleVec ) == true )
		return( true );
	if( *dispChk == false )
		return( false );

	CbxMath::Add3( vertexPos, handleVec, handlePos );

	return( false );
}


bool  CbvGLelement::
GetHandleRadial( bool* dispChk, double handlePos[ 3 ] )
{
	double  handleVec[ 3 ], vertexPos[ 3 ];

	if( GetHandleSub( true, dispChk, vertexPos, handleVec ) == true )
		return( true );
	if( *dispChk == false )
		return( false );

	CbxMath::Add3( vertexPos, handleVec, handlePos );

	return( false );
}


bool  CbvGLelement::
GetHandleSub( const bool radial, bool* dispChk, double vertexPos[ 3 ], double handleVec[ 3 ] )
{
	bool   result;
	ulong  edgeNo;
	short  beginEndNo;

	if( radial == false )
		GetVertexOrg( vertexPos );
	else {
		result = GetVertexRadial( vertexPos );
		if( result == true )
			return( true );
	}

	result = CbxDB::VertexRadialToEdge( CbvDlogSelVer::Data(), CbvDlogSelHand::Data(), &edgeNo, &beginEndNo );
	if( result == true )
		return( true );

	if( radial == true )
		beginEndNo = 1 - beginEndNo;

	result = CbxProc::GetHandle( edgeNo, beginEndNo, dispChk, handleVec );
	if( result == true )
		return( true );

	return( false );
}


bool  CbvGLelement::
GetCtlVecOrg( bool* dispChk, double ctlPos[ 3 ] )
{
	return( GetCtlVecSub( false, dispChk, ctlPos ) );
}


bool  CbvGLelement::
GetCtlVecRadial( bool* dispChk, double ctlPos[ 3 ] )
{
	return( GetCtlVecSub( true, dispChk, ctlPos ) );
}


bool  CbvGLelement::
GetCtlVecSub( const bool radial, bool* dispChk, double ctlPos[ 3 ] )
{
	bool    result;
	ulong   edgeNo;
	short   beginEndNo, minusPlusNo;
	double  handlePos[ 3 ], ctlVec[ 3 ];
	
	if( radial == false )
		result = GetHandleOrg( dispChk, handlePos );
	else
		result = GetHandleRadial( dispChk, handlePos );
	if( result == true )
		return( true );
	else if( *dispChk == false )
		return( false );
	
	result = CbxDB::VertexRadialToEdge( CbvDlogSelVer::Data(), CbvDlogSelHand::Data(), &edgeNo, &beginEndNo );
	if( result == true )
		return( true );

	if( radial == false )
		minusPlusNo = CbvDlogSelCtlBar::Data();
	else {
		beginEndNo  = 1 - beginEndNo;
		minusPlusNo = 1 - CbvDlogSelCtlBar::Data();
	}

	result = CbxProc::GetCtlInCtlVec( edgeNo, beginEndNo, minusPlusNo, dispChk, ctlVec );
	if( result == true )
		return( true );
	else if( *dispChk == false )
		return( false );

	CbxMath::Add3( handlePos, ctlVec, ctlPos );

	return( false );
}


bool  CbvGLelement::
GetWingHandleOrg( bool* dispChk, WingHandleInfo* dt )
{
	return( GetWingHandleSub( false, dispChk, dt ) );
}


bool  CbvGLelement::
GetWingHandleRadial( bool* dispChk, WingHandleInfo* dt )
{
	return( GetWingHandleSub( true, dispChk, dt ) );
}


bool  CbvGLelement::
GetWingHandleSub( const bool radial, bool* dispChk, WingHandleInfo* dt )
{
	bool               result;
	ulong              edgeNo;
	short              beginEndNo, minusPlusNo;
	double             vertexPos[ 3 ];
	WingHandleInfoSub  sub;

	result = CbxDB::VertexRadialToEdge( CbvDlogSelVer::Data(), CbvDlogSelHand::Data(), &edgeNo, &beginEndNo );
	if( result == true )
		return( true );

	if( radial == false )
		minusPlusNo = CbvDlogSelCtlBar::Data();
	else {
		beginEndNo  = 1 - beginEndNo;
		minusPlusNo = 1 - CbvDlogSelCtlBar::Data();
	}

	if( minusPlusNo == 0 )
		result = GetWingHandleDec( edgeNo, beginEndNo, dispChk, &sub );
	else
		result = GetWingHandleInc( edgeNo, beginEndNo, dispChk, &sub );
	if( result == true )
		return( true );
	else if( *dispChk == false )
		return( false );

	if( radial == false )
		GetVertexOrg( vertexPos );
	else {
		result = GetVertexRadial( vertexPos );
		if( result == true )
			return( true );
	}
	CbxMath::Add3( vertexPos, sub.handleVecOrg, dt->handlePosOrg );

	dt->extFlag = sub.extFlag;
	if( sub.extFlag == true ) {
		CbxMath::Copy3( sub.vertexPosRadial, dt->vertexPosRadial );
		CbxMath::Add3( sub.vertexPosRadial, sub.handleVecRadial, dt->handlePosRadial );
	}

	return( false );
}


bool  CbvGLelement::
GetWingHandleDec( const ulong edgeNo, const short beginEndNo, bool* dispChk, WingHandleInfoSub* dt )
{
	ulong   vertexNo, edgeNo2;
	ushort  radialNo, radialDec, radialInc, numRadial;
	short   beginEndNo2;
	bool    result, loopFlag;
	CbxDB::VertexRec  vertex;
	CbxDB::EdgeRec    edge;

	CbxDB::LoadEdgeRec( edgeNo, &edge );

	vertexNo  = edge.vertex[ beginEndNo ].vertexNo;

	CbxDB::LoadVertexRec( vertexNo, &vertex );
	numRadial = vertex.numRadial;
	loopFlag  = vertex.loopFlag;

	radialNo  = edge.vertex[ beginEndNo ].radialNo;
	radialDec = ( radialNo + numRadial - 1 ) % numRadial;
	radialInc = ( radialNo + 1 ) % numRadial;

	if( radialNo == 0 && loopFlag == false ) {
		result = CbxDB::VertexRadialToEdge( vertexNo, radialInc, &edgeNo2, &beginEndNo2 );
		if( result == true )
			return( true );
		dt->extFlag = false;
	}
	else {
		result = CbxDB::VertexRadialToEdge( vertexNo, radialDec, &edgeNo2, &beginEndNo2 );
		if( result == true )
			return( true );
		dt->extFlag = true;
		result = GetWingHandleExt( edgeNo2, 1-beginEndNo2, dispChk, dt->vertexPosRadial, dt->handleVecRadial );
		if( result == true )
			return( true );
		else if( *dispChk == false )
			return( false );
	}

	result = CbxProc::GetHandle( edgeNo2, beginEndNo2, dispChk, dt->handleVecOrg );
	if( result == true )
		return( true );
	else if( *dispChk == false )
		return( false );

	if( radialNo == 0 && loopFlag == false )
		CbxMath::Negate2( dt->handleVecOrg, dt->handleVecOrg );

	return( false );
}


bool  CbvGLelement::
GetWingHandleInc( const ulong edgeNo, const short beginEndNo, bool* dispChk, WingHandleInfoSub* dt )
{
	ulong   vertexNo, edgeNo2;
	ushort  radialNo, radialDec, radialInc, numRadial;
	short   beginEndNo2;
	bool    result, loopFlag;
	CbxDB::VertexRec  vertex;
	CbxDB::EdgeRec    edge;

	CbxDB::LoadEdgeRec( edgeNo, &edge );

	vertexNo  = edge.vertex[ beginEndNo ].vertexNo;

	CbxDB::LoadVertexRec( vertexNo, &vertex );
	numRadial = vertex.numRadial;
	loopFlag  = vertex.loopFlag;

	radialNo  = edge.vertex[ beginEndNo ].radialNo;
	radialDec = ( radialNo + numRadial - 1 ) % numRadial;
	radialInc = ( radialNo + 1 ) % numRadial;

	if( radialNo == ( numRadial-1 ) && loopFlag == false ) {
		result = CbxDB::VertexRadialToEdge( vertexNo, radialDec, &edgeNo2, &beginEndNo2 );
		if( result == true )
			return( true );
		dt->extFlag = false;
	}
	else {
		result = CbxDB::VertexRadialToEdge( vertexNo, radialInc, &edgeNo2, &beginEndNo2 );
		if( result == true )
			return( true );
		dt->extFlag = true;
		result = GetWingHandleExt( edgeNo2, 1-beginEndNo2, dispChk, dt->vertexPosRadial, dt->handleVecRadial );
		if( result == true )
			return( true );
		else if( *dispChk == false )
			return( false );
	}

	result = CbxProc::GetHandle( edgeNo2, beginEndNo2, dispChk, dt->handleVecOrg );
	if( result == true )
		return( true );
	else if( *dispChk == false )
		return( false );

	if( radialNo == ( numRadial-1 ) && loopFlag == false )
		CbxMath::Negate2( dt->handleVecOrg, dt->handleVecOrg );

	return( false );
}


bool  CbvGLelement::
GetWingHandleExt( const ulong edgeNo, const short beginEndNo, bool* dispChk, double vertexPos[ 3 ], double handleVec[ 3 ] )
{
	bool   result;
	ulong  vertexNo;
	CbxDB::VertexRec  vertex;
	CbxDB::EdgeRec    edge;

	CbxDB::LoadEdgeRec( edgeNo, &edge );
	vertexNo = edge.vertex[ beginEndNo ].vertexNo;

	CbxDB::LoadVertexRec( vertexNo, &vertex );
	CbxMath::Copy3( vertex.vertexPos, vertexPos );

	result = CbxProc::GetHandle( edgeNo, beginEndNo, dispChk, handleVec );
	if( result == true )
		return( true );
	else if( *dispChk == false )
		return( false );

	return( false );
}
	