/* bvGLmesh.cpp */

#include "../BeziexView.h"

const double  CbvGLmesh::meshSize            =  4.0;
const double  CbvGLmesh::meshDepth           = -1.2;
const double  CbvGLmesh::lenMeshZ            =  0.6;



void  CbvGLmesh::
Draw( void )
{
	glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
	glEnable( GL_DEPTH_TEST );

	glDisable( GL_LIGHTING );
	glLineWidth( 1.0 );

	DrawMesh();
}


void  CbvGLmesh::
DrawMesh( void )
{
	double  org[ 3 ], depth, currentMat[ 4 ][ 4 ];

	CbvGL::GetCurrentMat( currentMat );
	CbxMath::Set3( 0.0, 0.0, meshDepth, org );
	CbxMath::Transform3( org, currentMat, org );

	depth = meshDepth;
	if( org[ 2 ] > 0.0 )
		depth = -meshDepth;

	DrawMeshX( depth );
	DrawMeshY( depth );
	DrawMeshZ( depth );
}


void  CbvGLmesh::
DrawMeshX( const double depth )
{
	double  inc, y[ 2 ];

	inc = meshSize / ( numMeshLine - 1 );

	glColor3f( 0.5, 0.5, 0.5 );

	y[ 0 ] = -meshSize / 2;			y[ 1 ] = meshSize / 2;
	DrawMeshXsub( -inc, -inc, y, depth, numMeshLine/2 );

	y[ 0 ] = -meshSize / 2;			y[ 1 ] = 0.0;
	DrawMeshXsub( 0.0, inc, y, depth, (numMeshLine+1)/2 );

	glColor3f( 0.0, 0.0, 1.0 );
	y[ 0 ] = 0.0;					y[ 1 ] = meshSize / 2;
	DrawMeshXsub( 0.0, inc, y, depth, (numMeshLine+1)/2 );
}


void  CbvGLmesh::
DrawMeshXsub( const double sx, const double inc, const double y[ 2 ], const double depth, const short num )
{
	double  x;
	short   i;

	x = sx;
	for( i=0; i < num; i++ ) {
		glBegin( GL_LINES );
			glVertex3f( ( GLfloat )x, ( GLfloat )y[ 0 ], ( GLfloat )depth );
			glVertex3f( ( GLfloat )x, ( GLfloat )y[ 1 ], ( GLfloat )depth );
		glEnd();

		x += inc;
	}
}


void  CbvGLmesh::
DrawMeshY( const double depth )
{
	double  inc, x[ 2 ];

	inc = meshSize / ( numMeshLine - 1 );

	glColor3f( 0.5, 0.5, 0.5 );

	x[ 0 ] = -meshSize / 2;			x[ 1 ] = meshSize / 2;
	DrawMeshYsub( -inc, -inc, x, depth, numMeshLine/2 );

	x[ 0 ] = -meshSize / 2;			x[ 1 ] = 0.0;
	DrawMeshYsub( 0.0, inc, x, depth, (numMeshLine+1)/2 );

	glColor3f( 1.0, 0.0, 0.0 );
	x[ 0 ] = 0.0;					x[ 1 ] = meshSize / 2;
	DrawMeshYsub( 0.0, inc, x, depth, (numMeshLine+1)/2 );
}


void  CbvGLmesh::
DrawMeshYsub( const double sy, const double inc, const double x[ 2 ], const double depth, const short num )
{
	double  y;
	short   i;

	y = sy;
	for( i=0; i < num; i++ ) {
		glBegin( GL_LINES );
			glVertex3f( ( GLfloat )x[ 0 ], ( GLfloat )y, ( GLfloat )depth );
			glVertex3f( ( GLfloat )x[ 1 ], ( GLfloat )y, ( GLfloat )depth );
		glEnd();

		y += inc;
	}
}


void  CbvGLmesh::
DrawMeshZ( const double depth )
{
	glColor3f( 0.0, 1.0, 0.0 );

	DrawMeshZsub( -meshSize/2, -meshSize/2, depth );
	DrawMeshZsub( -meshSize/2,  meshSize/2, depth );
	DrawMeshZsub(  meshSize/2, -meshSize/2, depth );
	DrawMeshZsub(  meshSize/2,  meshSize/2, depth );
}


void  CbvGLmesh::
DrawMeshZsub( const double x, const double y, const double depth )
{
	glBegin( GL_LINES );
		glVertex3f( ( GLfloat )x, ( GLfloat )y, ( GLfloat )depth );
		glVertex3f( ( GLfloat )x, ( GLfloat )y, ( GLfloat )( depth+lenMeshZ ) );
	glEnd();
}
