/* bvPolygon.cpp */

#include "../BeziexView.h"



ulong                    CbvPolygon::numPolygonAry		= 0;
CbvPolygon::PolygonRec*  CbvPolygon::polygonAry			= 0;



bool  CbvPolygon::
Update( void )
{
	bool  result;

	FreeData();

	result = Alloc();
	if( result == true ) {
		CbvMessage::Error( IDS_CANNOTCONTINUE );
		return( true );
	}

	SetPolygon();

	return( false );
}


bool  CbvPolygon::
Alloc( void )
{
	bool  result, surfaceErr;

	result = Proc( FuncDispChk, &surfaceErr );
	if( result == true || surfaceErr == true )
		return( true );

	if( AllocData() == true )
		return( true );

	return( false );
}


bool  CbvPolygon::
Proc( bool ( *Exec )( const ulong edgeNo, const short beginEndNo, const ulong surfaceNo, bool* dispChk ),
	bool* surfaceErr )
{
	bool  result;

	result = CbxDB::AllocEdgeForPickSurface();
	if( result == true )
		return( true );

	result = ProcMain( Exec, surfaceErr );
	if( result == true ) {
		CbxDB::FreeEdgeForPickSurface();
		return( true );
	}

	CbxDB::FreeEdgeForPickSurface();

	return( false );
}


bool  CbvPolygon::
ProcMain( bool ( *Exec )( const ulong edgeNo, const short beginEndNo, const ulong surfaceNo, bool* dispChk ),
	bool* surfaceErr )
{
	ulong  i;
	short  j, chkSurfExist;
	bool   result, execOp;

	numPolygonAry = 0;

	*surfaceErr = false;
	for( i=0; i<CbxDB::NumEdgeAry(); i++ ) {
		for( j=0; j<2; j++ ) {
			result = ProcVertex( i, j, Exec, &execOp, &chkSurfExist );
			if( result == true )
				return( true );
			else if( execOp == true && chkSurfExist == CbxDB::surfNonExist )
				*surfaceErr = true;
		}
	}

	return( false );
}


bool  CbvPolygon::
ProcVertex( const ulong edgeNo, const short beginEndNo,
	bool ( *Exec )( const ulong edgeNo, const short beginEndNo, const ulong surfaceNo, bool* dispChk ),
	bool* execOp, short* chkSurfExist )
{
	bool                     result, dispChk;
	ulong                    edgeNo2;
	short                    numAngle, i, beginEndNo2;
	CbxDB::EdgeVertexRadial  surfaceVertex[ 4 ];

	if( CbxDB::GetEdgeForPickSurface( edgeNo, beginEndNo ) == true ) {
		*execOp = false;
		return( false );
	}

	*execOp = true;

	result = CbxDB::GetSurfaceVertex( edgeNo, beginEndNo, chkSurfExist, &numAngle, surfaceVertex );
	if( result == true )
		return( true );
	else if( *chkSurfExist != CbxDB::surfExist )
		return( false );

	for( i=0; i<numAngle; i++ ) {
		edgeNo2     = surfaceVertex[ i ].edgeNo;
		beginEndNo2 = surfaceVertex[ i ].beginEndNo;

		CbxDB::SetEdgeForPickSurface( edgeNo2, beginEndNo2 );
	}

	result = ( *Exec )( edgeNo, beginEndNo, numPolygonAry, &dispChk );
	if( result == true )
		return( true );

	if( dispChk == true ) {
		switch( numAngle ) {
			case 3:
				numPolygonAry += numTriInSurface;
				break;
			case 4:
				numPolygonAry += numRectInSurface;
				break;
			default:
				return( true );
		}
	}

	return( false );
}


bool  CbvPolygon::
FuncDispChk( const ulong edgeNo, const short beginEndNo, const ulong surfaceNo, bool* dispChk )
{
	bool                     result;
	CbxCore::SurfaceForPick  surface;

	result = CbxProc::LoadSurfaceForPick( edgeNo, beginEndNo, &surface, dispChk );
	if( result == true )
		return( true );
	else if( *dispChk == false )
		return( false );

	return( false );
}


bool  CbvPolygon::
AllocData( void )
{
	if( numPolygonAry == 0 )
		return( false );

	polygonAry = new PolygonRec[ numPolygonAry ];
	if( polygonAry == 0 )
		return( true );

	return( false );
}


void  CbvPolygon::
FreeData( void )
{
	if( numPolygonAry == 0 )
		return;

	delete [] polygonAry;

	numPolygonAry = 0;
	polygonAry    = 0;
}


void  CbvPolygon::
SetPolygon( void )
{
	bool  result, surfaceErr;

	result = Proc( FuncSetPolygon, &surfaceErr );
	if( result == true || surfaceErr == true ) {
		CbvMessage::Error( IDS_CANNOTCONTINUE );
		return;
	}
}


bool  CbvPolygon::
FuncSetPolygon( const ulong edgeNo, const short beginEndNo, const ulong surfaceNo, bool* dispChk )
{
	bool                     result;
	CbxCore::SurfaceForPick  surface;
	CbxCore::PatchRec        patch;

	result = CbxProc::LoadSurfaceForPick( edgeNo, beginEndNo, &surface, dispChk );
	if( result == true )
		return( true );
	else if( *dispChk == false )
		return( false );

	result = CbxCore::ConvSurface( &surface, dispChk, &patch );
	if( result == true )
		return( true );
	else if( *dispChk == false )
		return( false );

	switch( surface.numVertex ) {
		case  3:
			result = FuncSetTri( &patch, surfaceNo );
			break;
		case  4:
			result = FuncSetRect( &patch, surfaceNo );
			break;
		default:
			result = true;
	}
	if( result == true )
		return( true );

	return( false );
}


bool  CbvPolygon::
FuncSetTri( const CbxCore::PatchRec* patch, const ulong surfaceNo )
{
	short      numLineOnEdgeM, iu, iv, u0, u1, v0, v1, param[ 3 ][ 2 ];
	ulong      sno;
	double     uv[ 2 ];
	VertexRec  pnt[ numLineOnEdge+1 ][ numLineOnEdge+1 ];
	bool       result;

	for( iv=0; iv<( numLineOnEdge+1 ); iv++ ) {
		uv[ 1 ] = ( double )iv / numLineOnEdge;
		for( iu=0; iu<( numLineOnEdge-iv+1 ); iu++ ) {
			if( ( numLineOnEdge-iv ) == 0 )
				uv[ 0 ] = 0.0;
			else
				uv[ 0 ] = ( double )iu / ( numLineOnEdge-iv );
			result = CbxCore::PosNormalSurface( patch, uv, pnt[ iu ][ iv ].pos, pnt[ iu ][ iv ].normal );
			if( result == true )
				return( true );
		}
	}

	sno = surfaceNo;
	numLineOnEdgeM = numLineOnEdge - 1;

	for( iv=0; iv<numLineOnEdge; iv++ ) {
		v0 = iv;
		v1 = iv + 1;

		for( iu=0; iu<( numLineOnEdge-iv ); iu++ ) {
			u0 = iu;
			u1 = iu + 1;

			param[ 0 ][ 0 ] = u0;	param[ 0 ][ 1 ] = v0;
			param[ 1 ][ 0 ] = u1;	param[ 1 ][ 1 ] = v0;
			param[ 2 ][ 0 ] = u0;	param[ 2 ][ 1 ] = v1;
			FuncSetPolygonMain( pnt, sno, param );
			sno++;
		}
	}

	for( iv=0; iv<numLineOnEdgeM; iv++ ) {
		v0 = iv;
		v1 = iv + 1;

		for( iu=0; iu<( numLineOnEdgeM-iv ); iu++ ) {
			u0 = iu;
			u1 = iu + 1;

			param[ 0 ][ 0 ] = u0;	param[ 0 ][ 1 ] = v1;
			param[ 1 ][ 0 ] = u1;	param[ 1 ][ 1 ] = v0;
			param[ 2 ][ 0 ] = u1;	param[ 2 ][ 1 ] = v1;
			FuncSetPolygonMain( pnt, sno, param );
			sno++;
		}
	}

	return( false );
}


bool  CbvPolygon::
FuncSetRect( const CbxCore::PatchRec* patch, const ulong surfaceNo )
{
	short      iu, iv, u0, u1, v0, v1, param[ 3 ][ 2 ];
	ulong      sno;
	double     uv[ 2 ];
	VertexRec  pnt[ numLineOnEdge+1 ][ numLineOnEdge+1 ];
	bool       result;

	for( iv=0; iv<( numLineOnEdge+1 ); iv++ ) {
		uv[ 1 ] = ( double )iv / numLineOnEdge;
		for( iu=0; iu<( numLineOnEdge+1 ); iu++ ) {
			uv[ 0 ] = ( double )iu / numLineOnEdge;
			result = CbxCore::PosNormalSurface( patch, uv, pnt[ iu ][ iv ].pos, pnt[ iu ][ iv ].normal );
			if( result == true )
				return( true );
		}
	}

	sno = surfaceNo;

	for( iv=0; iv<numLineOnEdge; iv++ ) {
		v0 = iv;
		v1 = iv + 1;

		for( iu=0; iu<numLineOnEdge; iu++ ) {
			u0 = iu;
			u1 = iu + 1;

			param[ 0 ][ 0 ] = u0;	param[ 0 ][ 1 ] = v0;
			param[ 1 ][ 0 ] = u1;	param[ 1 ][ 1 ] = v0;
			param[ 2 ][ 0 ] = u0;	param[ 2 ][ 1 ] = v1;
			FuncSetPolygonMain( pnt, sno, param );
			sno++;

			param[ 0 ][ 0 ] = u0;	param[ 0 ][ 1 ] = v1;
			param[ 1 ][ 0 ] = u1;	param[ 1 ][ 1 ] = v0;
			param[ 2 ][ 0 ] = u1;	param[ 2 ][ 1 ] = v1;
			FuncSetPolygonMain( pnt, sno, param );
			sno++;
		}
	}

	return( false );
}


void  CbvPolygon::
FuncSetPolygonMain( const VertexRec pnt[ numLineOnEdge+1 ][ numLineOnEdge+1 ], const ulong surfaceNo,
	const short param[ 3 ][ 2 ] )
{
	short  i;

	for( i=0; i<3; i++ )
		polygonAry[ surfaceNo ].vertex[ i ] = pnt[ param[ i ][ 0 ] ][ param[ i ][ 1 ] ];
}
